Package: HiCDOC
Title: A/B compartment detection and differential analysis
Version: 1.15.2
Date: 2025-07-22
Authors@R: c(
    person("Kurylo", "Cyril", email = "cyril.kurylo@inrae.fr", role = c("aut")),
    person("Zytnicki", "Matthias", email = "matthias.zytnicki@inrae.fr", role = c("aut")),
    person("Foissac", "Sylvain", email = "sylvain.foissac@inrae.fr", role = c("aut")),
    person("Maigné", "Élise", email = "elise.maigne@inrae.fr", role = c("aut", "cre")))
Description: HiCDOC normalizes intrachromosomal Hi-C matrices, uses
        unsupervised learning to predict A/B compartments from multiple
        replicates, and detects significant compartment changes between
        experiment conditions. It provides a collection of functions
        assembled into a pipeline to filter and normalize the data,
        predict the compartments and visualize the results. It accepts
        several type of data: tabular `.tsv` files, Cooler `.cool` or
        `.mcool` files, Juicer `.hic` files or HiC-Pro `.matrix` and
        `.bed` files.
License: LGPL-3
Encoding: UTF-8
Depends: InteractionSet, GenomicRanges, SummarizedExperiment, R (>=
        4.1.0)
Imports: methods, ggplot2, Rcpp (>= 0.12.8), stats, S4Vectors, gtools,
        pbapply, BiocParallel, BiocGenerics, grid, cowplot, gridExtra,
        data.table, multiHiCcompare, Seqinfo, HiCParser
LinkingTo: Rcpp
Suggests: knitr, rmarkdown, testthat, BiocStyle, BiocManager, rhdf5
biocViews: HiC, DNA3DStructure, Normalization, Sequencing, Software,
        Clustering
VignetteBuilder: knitr
URL: https://github.com/mzytnicki/HiCDOC
BugReports: https://github.com/mzytnicki/HiCDOC/issues
Config/roxygen2/version: 8.0.0
Config/pak/sysreqs: libssl-dev zlib1g-dev
Repository: https://bioc.r-universe.dev
Date/Publication: 2026-06-08 14:08:02 UTC
RemoteUrl: https://github.com/bioc/HiCDOC
RemoteRef: HEAD
RemoteSha: 2c0c315182e4170e5604b42608222fb89a7b6a9a
NeedsCompilation: yes
Packaged: 2026-06-08 22:32:04 UTC; root
Author: Kurylo Cyril [aut],
  Zytnicki Matthias [aut],
  Foissac Sylvain [aut],
  Maigné Élise [aut, cre]
Maintainer: Maigné Élise <elise.maigne@inrae.fr>
Built: R 4.6.0; x86_64-apple-darwin20; 2026-06-08 22:35:08 UTC; unix

# Generated by roxygen2: do not edit by hand

export("parameters<-")
export(HiCDOC)
export(HiCDOCDataSetFromCool)
export(HiCDOCDataSetFromHiC)
export(HiCDOCDataSetFromHiCPro)
export(HiCDOCDataSetFromTabular)
export(chromosomes)
export(compartments)
export(concordances)
export(defaultHiCDOCParameters)
export(detectCompartments)
export(differences)
export(filterSmallChromosomes)
export(filterSparseReplicates)
export(filterWeakPositions)
export(normalizeBiologicalBiases)
export(normalizeDistanceEffect)
export(normalizeTechnicalBiases)
export(parameters)
export(plotCentroids)
export(plotCompartmentChanges)
export(plotCompartments)
export(plotConcordanceDifferences)
export(plotConcordances)
export(plotDistanceEffect)
export(plotInteractions)
export(plotSelfInteractionRatios)
export(reduceHiCDOCDataSet)
export(sampleConditions)
export(sampleReplicates)
exportClasses(HiCDOCDataSet)
exportMethods("parameters<-")
exportMethods(chromosomes)
exportMethods(compartments)
exportMethods(concordances)
exportMethods(differences)
exportMethods(parameters)
exportMethods(sampleConditions)
exportMethods(sampleReplicates)
exportMethods(show)
import(HiCParser)
import(InteractionSet)
import(data.table)
import(ggplot2)
import(methods)
importFrom(BiocGenerics,cbind)
importFrom(BiocGenerics,width)
importFrom(BiocParallel,bplapply)
importFrom(BiocParallel,bpmapply)
importFrom(BiocParallel,bpparam)
importFrom(GenomicRanges,GRanges)
importFrom(GenomicRanges,distance)
importFrom(GenomicRanges,end)
importFrom(GenomicRanges,match)
importFrom(GenomicRanges,start)
importFrom(GenomicRanges,union)
importFrom(Rcpp,evalCpp)
importFrom(Rcpp,sourceCpp)
importFrom(S4Vectors,"%in%")
importFrom(S4Vectors,DataFrame)
importFrom(S4Vectors,Rle)
importFrom(S4Vectors,mcols)
importFrom(S4Vectors,runLength)
importFrom(S4Vectors,split)
importFrom(Seqinfo,seqlevels)
importFrom(Seqinfo,seqnames)
importFrom(SummarizedExperiment,assay)
importFrom(SummarizedExperiment,colData)
importFrom(cowplot,get_plot_component)
importFrom(cowplot,ggdraw)
importFrom(grid,gpar)
importFrom(grid,textGrob)
importFrom(gridExtra,arrangeGrob)
importFrom(gtools,mixedsort)
importFrom(multiHiCcompare,cyclic_loess)
importFrom(multiHiCcompare,hic_table)
importFrom(multiHiCcompare,make_hicexp)
importFrom(pbapply,pbmapply)
importFrom(stats,ecdf)
importFrom(stats,loess)
importFrom(stats,loess.control)
importFrom(stats,median)
importFrom(stats,optimize)
importFrom(stats,p.adjust)
importFrom(stats,prcomp)
importFrom(stats,predict)
importFrom(stats,quantile)
importFrom(stats,update)
importFrom(stats,wilcox.test)
useDynLib(HiCDOC)

# HiCDOC: Compartments prediction and differential analysis with multiple replicates

HiCDOC normalizes intrachromosomal Hi-C matrices, uses unsupervised learning to
predict A/B compartments from multiple replicates, and detects significant
compartment changes between experiment conditions.

It provides a collection of functions assembled into a pipeline:

1. [Filter](#filtering-data):
    1. Remove chromosomes which are too small to be useful.
    2. Filter sparse replicates to remove uninformative replicates with few
       interactions.
    3. Filter positions (*bins*) which have too few interactions.
2. [Normalize](#normalizing-biases):
    1. Normalize technical biases using
       [cyclic loess normalization][multihiccompare-publication], so that
       matrices are comparable.
    2. Normalize biological biases using
       [Knight-Ruiz matrix balancing][knight-ruiz-publication], so that
       all the bins are comparable.
    3. Normalize the distance effect, which results from higher interaction
       proportions between closer regions, with a MD loess.
3. [Predict](#predicting-compartments-and-differences):
    1. Predict compartments using
       [constrained K-means][constrained-k-means-publication].
    2. Detect significant differences between experiment conditions.
4. [Visualize](#visualizing-data-and-results):
    1. Plot the interaction matrices of each replicate.
    2. Plot the overall distance effect on the proportion of interactions.
    3. Plot the compartments in each chromosome, along with their concordance
       (confidence measure) in each replicate, and significant changes between
       experiment conditions.
    4. Plot the overall distribution of concordance differences.
    5. Plot the result of the PCA on the compartments' centroids.
    6. Plot the boxplots of self interaction ratios (differences between self
       interactions and the medians of other interactions) of each compartment,
       which is used for the A/B classification.


# Table of contents

* [Installation](#installation)
* [Quick Start](#quick-start)
* [Usage](#quick-start)
    * [Importing Hi-C data](#importing-hi-c-data)
        * [Tabular files](#tabular-files)
        * [Cooler files](#cooler-files)
        * [Juicer files](#juicer-files)
        * [HiC-Pro files](#hic-pro-files)
    * [Running the HiCDOC pipeline](#running-the-hicdoc-pipeline)
        * [Filtering data](#filtering-data)
        * [Normalizing biases](#normalizing-biases)
        * [Predicting compartments and differences](#predicting-compartments-and-differences)
    * [Visualizing data and results](#visualizing-data-and-results)
* [References](#references)


# Installation

To install, execute the following commands in your console:

```bash
Rscript -e 'install.packages("devtools")'
Rscript -e 'devtools::install_github("mzytnicki/HiCDOC")'
```

After installation, the package can be loaded in R >= 4.0:

```r
library("HiCDOC")
```


# Quick Start

To try out HiCDOC, load the simulated toy data set:

```r
data(exampleHiCDOCDataSet)
hic.experiment <- exampleHiCDOCDataSet
```

Then run the default pipeline on the created object:

```r
hic.experiment <- HiCDOC(hic.experiment)
```

And plot some results:

```r
plotCompartmentChanges(hic.experiment, chromosome = 'Y')
```


# Usage

## Importing Hi-C data

HiCDOC can import Hi-C data sets in various different formats:
- Tabular `.tsv` files.
- Cooler `.cool` or `.mcool` files.
- Juicer `.hic` files.
- HiC-Pro `.matrix` and `.bed` files.

### Tabular files

A tabular file is a tab-separated multi-replicate sparse matrix with a header:

    chromosome    position 1    position 2    C1.R1    C1.R2    C2.R1    ...
    3             1500000       7500000       145      184      72       ...
    ...

The interaction proportions between `position 1` and `position 2` of
`chromosome` are reported in each `condition.replicate` column. There is no
limit to the number of conditions and replicates.

To load Hi-C data in this format:

```r
hic.experiment <- HiCDOCDataSetFromTabular('path/to/data.tsv')
```

### Cooler files

To load `.cool` or `.mcool` files generated by [Cooler][cooler-documentation]:

```r
# Path to each file
paths = c(
  'path/to/condition-1.replicate-1.cool',
  'path/to/condition-1.replicate-2.cool',
  'path/to/condition-2.replicate-1.cool',
  'path/to/condition-2.replicate-2.cool',
  'path/to/condition-3.replicate-1.cool'
)

# Replicate and condition of each file. Can be names instead of numbers.
replicates <- c(1, 2, 1, 2, 1)
conditions <- c(1, 1, 2, 2, 3)

# Resolution to select in .mcool files
binSize = 500000

# Instantiation of data set
hic.experiment <- HiCDOCDataSetFromCool(
  paths,
  replicates = replicates,
  conditions = conditions,
  binSize = binSize # Specified for .mcool files.
)
```

### Juicer files

To load `.hic` files generated by [Juicer][juicer-documentation]:

```r
# Path to each file
paths = c(
  'path/to/condition-1.replicate-1.hic',
  'path/to/condition-1.replicate-2.hic',
  'path/to/condition-2.replicate-1.hic',
  'path/to/condition-2.replicate-2.hic',
  'path/to/condition-3.replicate-1.hic'
)

# Replicate and condition of each file. Can be names instead of numbers.
replicates <- c(1, 2, 1, 2, 1)
conditions <- c(1, 1, 2, 2, 3)

# Resolution to select
binSize <- 500000

# Instantiation of data set
hic.experiment <- HiCDOCDataSetFromHiC(
  paths,
  replicates = replicates,
  conditions = conditions,
  binSize = binSize
)
```

### HiC-Pro files

To load `.matrix` and `.bed` files generated by [HiC-Pro][hicpro-documentation]:

```r
# Path to each matrix file
matrixPaths = c(
  'path/to/condition-1.replicate-1.matrix',
  'path/to/condition-1.replicate-2.matrix',
  'path/to/condition-2.replicate-1.matrix',
  'path/to/condition-2.replicate-2.matrix',
  'path/to/condition-3.replicate-1.matrix'
)

# Path to each bed file
bedPaths = c(
  'path/to/condition-1.replicate-1.bed',
  'path/to/condition-1.replicate-2.bed',
  'path/to/condition-2.replicate-1.bed',
  'path/to/condition-2.replicate-2.bed',
  'path/to/condition-3.replicate-1.bed'
)

# Replicate and condition of each file. Can be names instead of numbers.
replicates <- c(1, 2, 1, 2, 1)
conditions <- c(1, 1, 2, 2, 3)

# Instantiation of data set
hic.experiment <- HiCDOCDataSetFromHiCPro(
  matrixPaths = matrixPaths,
  bedPaths = bedPaths,
  replicates = replicates,
  conditions = conditions
)
```

## Running the HiCDOC pipeline

Once your data is loaded, you can run all the filtering, normalization, and
prediction steps with:

```r
hic.experiment <- HiCDOC(hic.experiment)
```

This one-liner runs all the steps detailed below.

### Filtering data

Remove small chromosomes of length smaller than 100 positions:

```r
hic.experiment <- filterSmallChromosomes(hic.experiment, threshold = 100)
```

Remove sparse replicates filled with less than 30% non-zero interactions:

```r
hic.experiment <- filterSparseReplicates(hic.experiment, threshold = 0.3)
```

Remove weak positions with less than 1 interaction in average:

```r
hic.experiment <- filterWeakPositions(hic.experiment, threshold = 1)
```

### Normalizing biases

Normalize technical biases such as sequencing depth:

```r
hic.experiment <- normalizeTechnicalBiases(hic.experiment)
```

Normalize biological biases (such as GC content, number of restriction sites,
etc.):

```r
hic.experiment <- normalizeBiologicalBiases(hic.experiment)
```

Normalize the distance effect resulting from higher interaction proportions
between closer regions:

```r
hic.experiment <- normalizeDistanceEffect(hic.experiment, loessSampleSize = 20000)
```

### Predicting compartments and differences

Predict A and B compartments and detect significant differences:

```r
hic.experiment <- detectCompartments(
  hic.experiment,
  kMeansDelta = 0.0001,
  kMeansIterations = 50,
  kMeansRestarts = 20
)
```


## Visualizing data and results

Plot the interaction matrix of each replicate:

```r
plotInteractions(hic.experiment, chromosome = '3')
```

Plot the overall distance effect on the proportion of interactions:

```r
plotDistanceEffect(hic.experiment)
```

List and plot compartments with their concordance (confidence measure) in each
replicate, and significant changes between experiment conditions:

```r
compartments(hic.experiment)

concordances(hic.experiment)

differences(hic.experiment)

plotCompartmentChanges(hic.experiment, chromosome = '3')
```

Plot the overall distribution of concordance differences:

```r
plotConcordanceDifferences(hic.experiment)
```

Plot the result of the PCA on the compartments' centroids:

```r
plotCentroids(hic.experiment, chromosome = '3')
```

Plot the boxplots of self interaction ratios (differences between self
interactions and the median of other interactions) of each compartment:

```r
plotSelfInteractionRatios(hic.experiment, chromosome = '3')
```


# References

John C Stansfield, Kellen G Cresswell, Mikhail G Dozmorov, multiHiCcompare:
joint normalization and comparative analysis of complex Hi-C experiments,
_Bioinformatics_, 2019, https://doi.org/10.1093/bioinformatics/btz048

Philip A. Knight, Daniel Ruiz, A fast algorithm for matrix balancing, _IMA
Journal of Numerical Analysis_, Volume 33, Issue 3, July 2013, Pages 1029–1047,
https://doi.org/10.1093/imanum/drs019

Kiri Wagstaff, Claire Cardie, Seth Rogers, Stefan Schrödl, Constrained K-means
Clustering with Background Knowledge, _Proceedings of 18th International
Conference on Machine Learning_, 2001, Pages 577-584,
https://pdfs.semanticscholar.org/0bac/ca0993a3f51649a6bb8dbb093fc8d8481ad4.pdf

[multihiccompare-publication]: https://doi.org/10.1093/bioinformatics/btz048
[knight-ruiz-publication]: https://doi.org/10.1093/imanum/drs019
[constrained-k-means-publication]: https://pdfs.semanticscholar.org/0bac/ca0993a3f51649a6bb8dbb093fc8d8481ad4.pdf

[cooler-documentation]: https://cooler.readthedocs.io/en/latest/
[juicer-documentation]: https://github.com/aidenlab/juicer/wiki/Data
[hicpro-documentation]: https://github.com/nservant/HiC-Pro

#' .modeVector Extract the mode of vector.
#'
#' @param x A vector
#'
#' @return The mode of a vector
#'
#' @examples
#' .modeVector(c(1, 2, 2, 2, 4))
#' @noRd
.modeVector <- function(x) {
    ux <- unique(x)
    ux[which.max(tabulate(match(x, ux)))]
}

#' @description
#' Determines the number of bins per chromosome.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#'
#' @return
#' A list of the number of bins per chromosome.
#'
#' @keywords internal
#' @noRd
.determineChromosomeSizes <- function(object) {
    tabChromosomes <- as.data.table(InteractionSet::regions(object))
    tabChromosomes[, minIndex := min(index), by = .(seqnames)]
    # minStart to correct chromosomes not starting at the 0 position
    tabChromosomes[
      index == minIndex,
      minStart := round(start / width),
      by = .(seqnames)
    ]
    # Comuting chromosome entire size
    tabChromosomes <- tabChromosomes[
        ,
        .(binSize = max(index) - min(index) + 1 + max(minStart, na.rm = TRUE)),
        by = .(seqnames)
    ]
    totalBins <- tabChromosomes$binSize
    names(totalBins) <- tabChromosomes$seqnames
    return(totalBins)
}

#' @description
#' Determines the valid conditions and replicates by chromosomes (not empty)
#'
#' @param interactions
#' An InteractionSet object
#'
#' @return
#' A list of length 2, validConditions and validreplicates.
#'
#' @keywords internal
#' @noRd
.determineValids <- function(object) {
    valids <- S4Vectors::split(
        SummarizedExperiment::assay(object),
        S4Vectors::mcols(object)$chromosome,
        drop = FALSE
    )
    valids <- lapply(valids, function(x) 
        apply(x, 2, stats::var, na.rm=TRUE))
    valids <- lapply(valids, function(x) which(x>0 & !is.na(x)))
    return(valids)
}
#' @description
#' Fills parameters and slots describing the data. Called by a
#' \code{\link{HiCDOCDataSet}} constructor.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#'
#' @return
#' A filled \code{\link{HiCDOCDataSet}} ready for analysis.
#'
#' @keywords internal
#' @noRd
.fillHiCDOCDataSet <- function(object) {
    # Reduce the levels in interaction part
    object <- InteractionSet::reduceRegions(object)
    objectRegions <- InteractionSet::regions(object)
    chromosomeNames <- unique(as.character(
        Seqinfo::seqnames(objectRegions)
    ))
    chromosomeNames <- gtools::mixedsort(chromosomeNames)
    Seqinfo::seqlevels(
        InteractionSet::regions(object),
        pruning.mode = "coarse"
    ) <- chromosomeNames

    # Add chromosome column for split purpose
    chromosomes <-
        Seqinfo::seqnames(InteractionSet::anchors(object, "first"))
    chromosomes <- S4Vectors::Rle(factor(chromosomes, levels = chromosomeNames))
    S4Vectors::mcols(object) <- S4Vectors::DataFrame("chromosome" = chromosomes)

    # Sorting interactions and assay
    ids <- InteractionSet::anchors(object, id = TRUE)
    neworder <- order(chromosomes, ids$first, ids$second)
    object <- object[neworder, ]

    # Fill all other slots than interactionSet part
    # Chromosomes and their size (max bin)
    object@chromosomes <- chromosomeNames
    object@totalBins <- .determineChromosomeSizes(object)
    object@parameters <- defaultHiCDOCParameters

    # Valid conditions and replicates by chromosome (==not empty)
    # maybe do a function for valid conditions and replicates ?
    valids <- .determineValids(object)
    object@validAssay <- valids

    # Weakbins
    object@weakBins <- vector("list", length(object@chromosomes))
    names(object@weakBins) <- object@chromosomes

    return(object)
}

#' @description
#' Normalizes the distance effect on the interactions of a given chromosome.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#' @param chromosomeName
#' The name of a chromosome to normalize.
#'
#' @return
#' A data.table of normalized interactions.
#'
#' @keywords internal
#' @noRd
.normalizeDistanceEffectOfChromosome <- function(object) {
    chromosomeName <- as.character(
        SummarizedExperiment::mcols(object)$chromosome[1]
    )
    message(
        "Chromosome ",
        chromosomeName,
        ": normalizing distance effect."
    )

    currentAssay <- SummarizedExperiment::assay(object)

    # Reordering columns in alphabetic order (useful for tests)
    validAssay <- object@validAssay[[chromosomeName]]
    refOrder <- paste(object$condition, object$replicate)
    values <- currentAssay[, validAssay, drop = FALSE]
    values <- values[, order(refOrder[validAssay]), drop = FALSE]

    distances <- InteractionSet::pairdist(object, type = "mid")

    chromosomeValues <- data.table(
        "distance" = rep(distances, length(validAssay)),
        "value" = as.vector(values)
    )
    chromosomeValues <- chromosomeValues[!is.na(value), ]
    setorder(chromosomeValues, distance, value)

    idSample <- sample(
        seq_len(nrow(chromosomeValues)),
        size = min(
            object@parameters$loessSampleSize,
            nrow(chromosomeValues)
        )
    )
    sample <- chromosomeValues[idSample]
    setorder(sample, distance)

    if (nrow(sample) == 0) {
        message("Chromosome ", chromosomeName, " is empty.")
        return(NULL)
    }

    optimizeSpan <- function(
        model,
        criterion = c("aicc", "gcv"),
        spans = c(0.01, 0.9)
    ) {
        criterion <- match.arg(criterion)
        result <- stats::optimize(
            function(span) {
                model <- stats::update(model, span = span)
                span <- model$pars$span
                trace <- model$trace.hat
                sigma2 <- sum(model$residuals ^ 2) / (model$n - 1)
                if (criterion == "aicc") {
                    quality <- (
                        log(sigma2) + 1 + 2 * (2 * (trace + 1)) /
                        (model$n - trace - 2)
                    )
                } else if (criterion == "gcv") {
                    quality <- model$n * sigma2 / (model$n - trace) ^ 2
                }
                return(quality)
            },
            spans
        )
        return(result$minimum)
    }

    traceMethod <- "approximate"
    if (object@parameters$loessSampleSize <= 1000) {
        traceMethod <- "exact"
    }
    sample[,logvalue := log(value)]
    sample[,logdistance := log(distance + 1)]
    loess <- stats::loess(
        logvalue ~ logdistance,
        data = sample,
        control = stats::loess.control(trace.hat = traceMethod)
    )
    span <- optimizeSpan(loess, criterion = "gcv")

    loess <- stats::loess(
        logvalue ~ logdistance,
        span = span,
        data = sample,
        control = stats::loess.control(trace.hat = traceMethod)
    )
    sample[, bias := stats::predict(loess)]
    sample[, value := NULL]
    sample[, logvalue := NULL]
    sample[, logdistance := NULL]
    data.table::setnames(sample, "distance", "sampleDistance")
    sample <- unique(sample)

    uniqueDistances <- unique(sort(chromosomeValues$distance))
    sampleDistance <- unique(sort(sample$sampleDistance))
    sampleDistance <- vapply(
        uniqueDistances,
        function(distance) {
            sampleDistance[which.min(abs(distance - sampleDistance))]
        },
        FUN.VALUE = 0
    )
    valueMap <- data.table(
        "distance" = uniqueDistances,
        "sampleDistance" = sampleDistance
    )
    valueMap <- data.table::merge.data.table(
        valueMap,
        sample,
        by = "sampleDistance"
    )
    # Sort sample distances and bias in the same order than the data
    loessDistances <- data.table::merge.data.table(
        data.table("distance" = distances),
        valueMap,
        by = "distance",
        sort = FALSE,
        all.x = TRUE
    )
    currentAssay <- log(currentAssay) - loessDistances$bias
    return(currentAssay)
}

#' @title
#' Normalize distance effect.
#'
#' @description
#' Normalizes interactions by their "expected" value relative to the distance
#' that separates their positions. The "expected" values are estimated with a
#' loess regression on the proportion of interactions for each distance.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#' @param loessSampleSize
#' The number of positions used as a sample to estimate the effect of distance
#' on proportion of interactions. Defaults to
#' \code{object$loessSampleSize} which is originally set to
#' \code{defaultHiCDOCParameters$loessSampleSize} = 20000.
#' @param parallel
#' Should the normalization be run in parallel mode? Default to FALSE.
#'
#' @return
#' A \code{\link{HiCDOCDataSet}} with normalized interactions.
#'
#' @examples
#' data(exampleHiCDOCDataSet)
#' object <- normalizeDistanceEffect(exampleHiCDOCDataSet)
#'
#' @seealso
#' \code{\link{normalizeTechnicalBiases}},
#' \code{\link{normalizeBiologicalBiases}},
#' \code{\link{HiCDOC}}
#'
#' @export
normalizeDistanceEffect <- function(
    object,
    loessSampleSize = NULL,
    parallel = FALSE
) {
    if (!is.null(loessSampleSize)) {
        object@parameters$loessSampleSize <- loessSampleSize
    }
    object@parameters <- .validateParameters(object@parameters)
    objectChromosomes <- S4Vectors::split(
        object,
        SummarizedExperiment::mcols(object)$chromosome,
        drop = FALSE
    )

    normAssay <- .internalLapply(
        parallel,
        objectChromosomes,
        FUN = .normalizeDistanceEffectOfChromosome
    )

    normAssay <- do.call("rbind", normAssay)
    SummarizedExperiment::assay(object) <- normAssay

    return(object)
}

test_that("normalizeDistanceEffect behaves as expected", {
    data(exampleHiCDOCDataSet)
    object <- reduceHiCDOCDataSet(exampleHiCDOCDataSet, chromosomes = c("X"))
    object <- filterSparseReplicates(object)
    object <- filterWeakPositions(object)
    
    # Apply normalization
    set.seed(123)
    expect_message(norm <- normalizeDistanceEffect(object))
    # Filtering 0 values before normalisation
    expect_equal(length(norm), length(object))
    matAssay <- SummarizedExperiment::assay(norm)
    
    expect_equal(sum(!is.na(matAssay)), 35105)
    expect_equal(round(colSums(matAssay, na.rm=TRUE),3), 
                 c(-3004.957, -1585.685, 0, 0, 1164.882, 1328.834, 2160.781))
})

#' @title
#' Plot A/B compartments.
#'
#' @description
#' Plots the predicted compartments in each experiment condition.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#' @param chromosome
#' A chromosome name or index in \code{chromosomes(object)}.
#' @param xlim
#' A vector of the minimum and maximum positions to display. If NULL, displays
#' all positions. Defaults to NULL.
#' @param colour
#' Border color for the compartments. Default to `gray90`. `NA` means no border.
#'
#' @return
#' A \code{ggplot}.
#'
#' @examples
#' data(exampleHiCDOCDataSetProcessed)
#' plotCompartments(exampleHiCDOCDataSetProcessed, chromosome = 1)
#'
#' @export
plotCompartments <- function(
    object,
    chromosome,
    xlim = NULL,
    colour = "gray90"
) {

    .validateSlots(object, slots = c("compartments"))
    chromosomeName <- .validateNames(object, chromosome, "chromosomes")
    xlim <- .validateXlim(xlim, object, chromosomeName)

    compartments <- object@compartments[
        Seqinfo::seqnames(object@compartments) == chromosomeName
    ]

    if (length(compartments) == 0) {
        message("No compartments for chromosome ", chromosomeName, ".")
        return(NULL)
    }
    compartments <- as.data.table(compartments)
    binSize <- .modeVector(compartments$width)
    compartments[, position := start + 0.5 * binSize]
    # for the last bin if not of the same size
    if(identical(xlim, c(min(compartments[,start]), max(compartments[,end])))){
        xlim <- c(min(xlim[1], compartments[,position]),
                  max(xlim[2], compartments[,start + binSize]))
    }
    plot <- ggplot(
        data = compartments,
        aes(x = position, fill = compartment)
    ) + geom_histogram(
        binwidth = binSize,
        colour = colour,
        linewidth = 0.05
    ) + coord_cartesian(
        xlim=c(xlim[1] - 0.5 * binSize, xlim[2] + 0.5 * binSize)
    ) + facet_grid(
        rows = vars(condition),
        margins = FALSE,
        switch = "y"
    ) + labs(title = paste0("Compartments of chromosome ", 
                            chromosomeName, " by condition")) +
        theme_minimal() + theme(
        axis.title = element_blank(),
        axis.text = element_blank(),
        axis.line = element_blank(),
        axis.ticks = element_blank(),
        panel.grid = element_blank(),
        panel.spacing = unit(2, "pt"),
        legend.position = "bottom",
        legend.title = element_text(size = 8),
        legend.text = element_text(size = 8),
        strip.placement = "outside"
    )

    return(plot)
}

test_that("plotCompartments returns error if no compartments", {
    data(exampleHiCDOCDataSet)
    expect_error(
        pp <- plotCompartments(exampleHiCDOCDataSet),
        "No compartments found."
    )
})

test_that("plotCompartments behaves as expected", {
    data(exampleHiCDOCDataSetProcessed)
    expect_error(plotCompartments(exampleHiCDOCDataSetProcessed),
        "argument \"chromosome\"")
    expect_error(plotCompartments(exampleHiCDOCDataSetProcessed, 5), "Unknown")

    pp <- plotCompartments(exampleHiCDOCDataSetProcessed, 1)
    expect_true(ggplot2::is_ggplot(pp))
    labs <- ggplot2::get_labs(pp)
    expect_equivalent(labs[["title"]], "Compartments of chromosome X by condition")
    expect_equivalent(labs[["x"]], "position")
    expect_equivalent(labs[["fill"]], "compartment")
    expect_equivalent(labs[["y"]], "count")
    expect_equivalent(labs[["weight"]], "weight")
    expect_is(pp$layers[[1]]$geom, "GeomBar")
    # No error when printed
    expect_error(print(pp), NA)
})

#' @title
#' Plot centroids.
#'
#' @description
#' Plots the result of the PCA on the compartments' centroids.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#' @param chromosome
#' A chromosome name or index in \code{chromosomes(object)}.
#' @param size
#' Size of each point. Defaults to 2.
#' @param checks
#' Whether or not to add sanity checks messages on centroids. Default to TRUE.
#'
#' @return
#' A \code{ggplot}.
#'
#' @examples
#' data(exampleHiCDOCDataSetProcessed)
#' plotCentroids(exampleHiCDOCDataSetProcessed, chromosome = 1)
#'
#' @export
plotCentroids <- function(object, chromosome, size = 2, checks=TRUE) {
    .validateSlots(object, slots = "centroids")
    if (length(chromosome) > 1) {
        warning(
            "`chromosome` should be of length 1, ",
            "taking the first one"
        )
        chromosome < chromosome[1]
    }
    chromosomeName <- .validateNames(object, chromosome, "chromosomes")
    
    pcaData <- .computePca(object, chromosomeName)
    pca     <- pcaData$PCA
    propvar <- pcaData$propvar
    propvar <- paste(round(100 * propvar, 2), "%")

    plot <- ggplot(
        pca,
        aes(
            x = PC1,
            y = PC2,
            color = compartment,
            shape = condition
        )
    ) + geom_point(size = size) + labs(
        title = paste0("PCA on centroids of chromosome ", chromosomeName),
        x = paste("PC1 ", propvar[1]),
        y = paste("PC2 ", propvar[2])
    )
    if(checks){
        messages <- .messageCheck(object, chromosomeName)
        plot <- plot + 
            labs(caption = paste("Quality controls:",
                                 messages$PC1, messages$centroids, sep="\n"))
    }
    return(plot)
}

test_that("plotCentroids returns error if no centroids", {
    data(exampleHiCDOCDataSet)
    expect_error(
        pp <- plotCentroids(exampleHiCDOCDataSet),
        "No compartments found."
    )
})

test_that("plotCentroids behaves as expected", {
    data(exampleHiCDOCDataSetProcessed)
    expect_error(plotCentroids(exampleHiCDOCDataSetProcessed),
        "argument \"chromosome\"")
    expect_error(plotCentroids(exampleHiCDOCDataSetProcessed, 5), "Unknown")

    pp <- plotCentroids(exampleHiCDOCDataSetProcessed, 1)
    expect_true(ggplot2::is_ggplot(pp))
    labs <- ggplot2::get_labs(pp)
    expect_equivalent(labs[["caption"]], 
        "Quality controls:\nCentroid PC1 inertia: OK\nA/B clustering consistency: OK")
    expect_equivalent(labs[["x"]], "PC1  91.19 %")
    expect_equivalent(labs[["y"]], "PC2  6.82 %")
    expect_equivalent(labs[["title"]], 
        "PCA on centroids of chromosome X")
    expect_equivalent(labs[["colour"]], "compartment")
    expect_equivalent(labs[["shape"]], "condition")
    expect_is(pp$layers[[1]]$geom, "GeomPoint")
    # No error when printed
    expect_error(print(pp), NA)
})

#' @title
#' Plot concordances.
#'
#' @description
#' Plots the concordances of each replicate in each experiment condition. A
#' concordance can be understood as a confidence in a genomic position's
#' assigned compartment. Mathematically, it is the log ratio of a genomic
#' position's distance to each compartment's centroid, normalized by the
#' distance between both centroids, and min-maxed to a [-1,1] interval.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#' @param chromosome
#' A chromosome name or index in \code{chromosomes(object)}.
#' @param xlim
#' A vector of the minimum and maximum positions to display. If NULL, displays
#' all positions. Defaults to NULL.
#' @param threshold
#' Significance threshold for the compartment changes. Defaults to 0.05.
#' @param points
#' Whether or not to add points to the concordances. Defaults to FALSE.
#'
#' @return
#' A \code{ggplot}.
#'
#' @examples
#' data(exampleHiCDOCDataSetProcessed)
#' plotConcordances(exampleHiCDOCDataSetProcessed, chromosome = 1)
#'
#' @export
plotConcordances <- function(
    object,
    chromosome,
    xlim = NULL,
    threshold = 0.05,
    points = FALSE
) {
    .validateSlots(object, slots = c("concordances", "differences"))
    chromosomeName <- .validateNames(object, chromosome, "chromosomes")
    xlim <- .validateXlim(xlim, object, chromosomeName)

    concordances <- object@concordances[
        Seqinfo::seqnames(object@concordances) == chromosomeName
    ]
    concordances <- data.table::as.data.table(concordances)
    concordances[, condition := paste0("Concordances\n", condition)]
    concordances <- concordances[start >= xlim[1] & start <= xlim[2]]
    binSize <- .modeVector(concordances$width)

    if (nrow(concordances) == 0) {
        message("No concordances for chromosome ", chromosomeName, ".")
        return(NULL)
    }

    # Significant differences
    differences <- object@differences[
        Seqinfo::seqnames(object@differences) == chromosomeName
    ]
    differences <- as.data.table(differences)
    differences <- differences[pvalue.adjusted <= threshold]
    differences <- differences[start >= xlim[1] & start <= xlim[2]]
    differences <- data.table::melt(
        differences,
        id.vars = c("seqnames", "start", "end"),
        measure.vars = c("condition.1", "condition.2"),
        value.name = "condition"
    )
    differences[, condition := paste0("Concordances\n", condition)]


    caption <- "The grey areas are significant changes"
    if (nrow(differences) == 0) {
        caption <- "No change is significant"
    }
    caption <- paste0(
        caption,
        " (adjusted p-value <= ",
        round(100 * threshold, 2),
        "%)"
    )

    ylim <- c(
        min(concordances$concordance, na.rm = TRUE),
        max(concordances$concordance, na.rm = TRUE)
    )

    plot <- ggplot()
    if (nrow(differences) > 0) {
        plot <- plot + geom_rect(
            data = differences,
            aes(
                xmin = start,
                xmax = end,
                ymin = ylim[1],
                ymax = ylim[2]
            ),
            color = NA,
            fill = "gray80"
        )
    }
    plot <- plot + geom_line(
        data = concordances,
        aes(
            x = start + 0.5 * binSize,
            y = concordance,
            color = replicate
        )
    )
    if (points) {
        plot <- plot + geom_point(
            data = concordances,
            aes(
                x = start + 0.5 * binSize,
                y = concordance,
                color = replicate
            )
        )
    }
    plot <- (
        plot +
        labs(caption = caption) +
        xlab("position") +
        xlim(xlim[1], xlim[2] + binSize) +
        ylim(ylim) +
            geom_hline(yintercept = 0.0, linewidth = 0.1) +
        facet_grid(rows = vars(condition), margins = FALSE, switch = "y") +
        labs(title = paste0("Concordances of chromosome ",
                            chromosomeName, " by condition")) +
        theme_minimal() +
        theme(axis.title.x = element_text(size = 10)) +
        theme(
            axis.title.y = element_blank(),
            axis.line.y = element_blank(),
            axis.ticks = element_blank(),
            legend.position = "bottom",  
            legend.title = element_text(size = 8),
            legend.text = element_text(size = 8),
            strip.placement = "outside"
        )
    )
    return(plot)
}

test_that("plotConcordances returns error if no compartments", {
    data(exampleHiCDOCDataSet)
    expect_error(
        pp <- plotConcordances(exampleHiCDOCDataSet),
        "No compartments found."
    )
})

test_that("plotConcordances behaves as expected", {
    data(exampleHiCDOCDataSetProcessed)
    expect_error(
        plotConcordances(exampleHiCDOCDataSetProcessed), 
        '"chromosome"'
    )
    expect_error(plotConcordances(exampleHiCDOCDataSetProcessed, 6), "Unknown")

    pp <- plotConcordances(exampleHiCDOCDataSetProcessed, 1)
    expect_true(ggplot2::is_ggplot(pp))
    labs <- ggplot2::get_labs(pp)
    expect_equivalent(labs[["title"]], 
        "Concordances of chromosome X by condition")
    expect_equivalent(labs[["caption"]], 
        "The grey areas are significant changes (adjusted p-value <= 5%)")
    expect_equivalent(labs[["x"]], "position")
    expect_equivalent(labs[["y"]], "concordance")
    expect_equivalent(labs[["colour"]], "replicate")
    
    expect_is(pp$layers[[1]]$geom, "GeomRect")
    expect_is(pp$layers[[2]]$geom, "GeomLine")
    # No error when printed
    expect_error(print(pp), NA)
})

#' @description
#' Returns the provided limit if valid, or the minimium and maximium of the
#' chromosome's positions.
#'
#' @param xlim
#' A numeric vector of a minimum and a maximum limit for the x axis.
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#' @param chromosomeName
#' The name of a chromosome.
#'
#' @return
#' A numeric vector of a minimum and a maximum limit for the x axis.
#'
#' @keywords internal
#' @noRd
.validateXlim <- function(xlim, object, chromosomeName) {
    if (!is.null(xlim)) {
        if (length(xlim) != 2) {
            message(
                "Expected a vector of two numbers but received '",
                paste(xlim, collapse = " "),
                "'. Setting xlim to NULL."
            )
            xlim <- NULL
        } else {
            xlim <- sort(xlim)
        }
    }
    if (is.null(xlim)) {
        regions <- InteractionSet::regions(object)
        regions <- regions[Seqinfo::seqnames(regions) == chromosomeName]
        xlim <- c(
            min(GenomicRanges::start(regions), na.rm = TRUE),
            max(GenomicRanges::end(regions), na.rm = TRUE)
        )
    }
    return(xlim)
}

#' @description
#' Extract messages for the user after sanity checks
#'
#' @param object
#' A HiCDOCDataSet object.
#' @param chromosome
#' A chromosome name or index in chromosomes(object).
#'
#' @return
#' A character vector.
#'
#' @keywords internal
#' @noRd
.messageCheck <- function(object, chomosomeName) {
    checks <- object@checks[chromosome == chomosomeName]
    messagesChecks <- list(
        "PC1" = "Centroid PC1 inertia: OK",
        "centroids" = "A/B clustering consistency: OK",
        "assignment" = "A/B assignment reliability: OK")
    if(!checks$centroid.check){
        messagesChecks$centroids <- "A/B clustering consistency: WARNING"
    }
    if(!checks$PC1.check){
        messagesChecks$PC1 <- "Centroid PC1 inertia: WARNING"
    }
    if(!checks$assignment.check){
        messagesChecks$assignment <- "A/B assignment reliability: FAIL"
    }
    return(messagesChecks)
}


#' @description
#' Complete the levels of replicates to get balanced condition x replicate
#' @param replicates
#' Vector of replicates for one conditino
#' @param expectedLength
#' Expected length of replicates levels
#' @param condition
#' Name of the condition
#'
#' @return
#' A vector with fictif levels if some missing
#'
#' @keywords internal
#' @noRd
.completeLevels <- function(replicates, expectedLength, condition) {
    if (length(replicates) < expectedLength) {
        complete <- paste0("R.", seq(expectedLength))
        complete[seq(length(replicates))] <- replicates
    } else {
        complete <- replicates
    }
    return(complete)
}

#' @title
#' Compute PCA
#'
#' @description
#' Helper function that computes Principal Components of centroids.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#' @param chromosomeName
#' A chromosome name or index in \code{chromosomes(object)}.
#'
#' @return
#' A list, with a \code{data.table}, which contains the PCA,
#' and the variability explained by the 2 first axes.
#'
#' @keywords internal
#' @noRd
.computePca <- function(object, chromosomeName) {
    df <- object@centroids[
        chromosome == chromosomeName,
        .(condition, compartment, centroid)
    ]
    if (nrow(df) == 0) {
        message("No centroids for chromosome ", chromosomeName, ".")
        return(NULL)
    }
    conditions <- df$condition
    compartments <- df$compartment
    
    df <- lapply(df$centroid, unlist)
    df <- do.call("rbind", df)
    
    pca <- stats::prcomp(df)
    varpca <- pca$sdev ^ 2
    propvar <- varpca / sum(varpca)
    
    pca <- as.data.table(pca$x)
    pca[, condition := conditions]
    pca[, compartment := compartments]
    
    return(list(PCA = pca, propvar = propvar))
}

#' Interactions plot using facet_grid
#'
#' @param interactions
#' a data.table
#' @param xylim
#' a length 2 vector : limits of the matrix
#' @param transform
#' character : transformation to be passed to \code{scale_fill_gradient2}
#' @param colours
#' vector of colors of length 3 to be passed to \code{scale_fill_gradient2}
#' @param midpoint
#' midpoint value to be passed to \code{scale_fill_gradient2}
#' @param chromosomeName
#' Character, name of the chromosome
#' @return
#' a ggplot object
#'
#' @keywords internal
#' @noRd
.plotInteractionsGrid <- function(
    dataplot,
    xylim,
    transform,
    colours,
    midpoint,
    chromosomeName
) {
    plot <- ggplot(
        data = dataplot,
        aes(x = start1, y = start2, z = interaction)
    ) + geom_tile(
        aes(fill = interaction),
        na.rm = TRUE
    ) + geom_tile(
        data = dataplot[start1 != start2, ],
        aes(x = start2, y = start1, fill = interaction),
        na.rm = TRUE
    ) + coord_fixed(ratio = 1) + theme_bw() + xlim(xylim) + scale_y_reverse(
        limits = rev(xylim)
    ) + facet_grid(
        condition ~ replicate,
        drop = FALSE
    ) + labs(
        title = paste0("Chromosome ", chromosomeName),
        x = "",
        y = ""
    ) + scale_fill_gradient2(
        low=colours[1],
        mid=colours[2],
        high=colours[3],
        midpoint = midpoint,
        trans = transform,
        name = "Intensity",
        na.value = "transparent"
    )
    return(plot)
}


#' Interactions plot using facet_wrap
#'
#' @param interactions
#' a data.table
#' @param xylim
#' a length 2 vector : limits of the matrix
#' @param transform
#' character : transformation to be passed to \code{scale_fill_gradient2}
#' @param colours
#' vector of colors of length 3 to be passed to \code{scale_fill_gradient2}
#' @param midpoint
#' midpoint value to be passed to \code{scale_fill_gradient2}
#' @param chromosomeName
#' Character, name of the chromosome
#' @param totalRows
#' Integer, number of rows in facet_wrap
#' @param totalCols
#' Integer, number of colums in facet_wrap
#'
#' @return
#' a ggplot object
#'
#' @keywords internal
#' @noRd
.plotInteractionsWrap <- function(
    dataplot,
    xylim,
    transform,
    colours,
    midpoint,
    chromosomeName,
    totalRows,
    totalCols
) {
    plot <- ggplot(
        data = dataplot,
        aes(x = start1, y = start2, z = interaction)
    ) + geom_tile(
        aes(fill = interaction),
        na.rm = TRUE
    ) + geom_tile(
        data = dataplot[start1 != start2, ],
        aes(x = start2, y = start1, fill = interaction),
        na.rm = TRUE
    ) + coord_fixed(ratio = 1) + theme_bw() + xlim(xylim) + scale_y_reverse(
        limits = rev(xylim)
    ) + facet_wrap(
        . ~ variable,
        ncol = totalCols,
        nrow = totalRows,
        drop = FALSE
    ) + labs(
        title = paste("Chromosome ", chromosomeName),
        x = "",
        y = ""
    ) + scale_fill_gradient2(
        low=colours[1],
        mid=colours[2],
        high=colours[3],
        midpoint = midpoint,
        trans = transform,
        name = "Intensity",
        na.value = "transparent"
    )
    return(plot)
}

#' @title
#' Plot interaction matrices.
#'
#' @description
#' Plots the interaction matrices as heatmaps.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#' @param chromosome
#' A chromosome name or index in \code{chromosomes(object)}.
#' @param transform
#' Transformation of the color scale. Default to NULL (no transformation). See
#' \code{\link[ggplot2]{scale_fill_gradient2}} for other accepted values.
#' @param colours
#' A character vector colours of length 3 to use for the gradient. See
#' \code{\link[ggplot2]{scale_fill_gradient2}} for more info. Defaults to
#' \code{c("low"=#2c7bb6", "mid"=#ffffbf", "high"="#d7191c")}.
#' @param midpoint
#' midpoint value to be passed to \code{scale_fill_gradient2}.
#' Default to 0.
#'
#' @return
#' A \code{ggplot}.
#'
#' @examples
#' data(exampleHiCDOCDataSet)
#' plotInteractions(exampleHiCDOCDataSet, chromosome = 1)
#'
#' @export
plotInteractions <- function(
    object,
    chromosome,
    transform = NULL,
    colours = c("low" = "#2c7bb6", "mid" = "#ffffbf", "high" = "#d7191c"),
    midpoint = 0
) {
    chromosomeName <- .validateNames(object, chromosome, "chromosomes")
    if (is.null(transform)) {
        transform <- "identity"
    }

    rowsChromosome <- (S4Vectors::mcols(object)$chromosome == chromosomeName)
    assayChromosome <- SummarizedExperiment::assay(object[rowsChromosome, ])
    assayChromosome <- data.table::as.data.table(assayChromosome)
    data.table::setnames(
        assayChromosome,
        paste(object$condition, object$replicate, sep = "_")
    )

    interactionsChromosome <- InteractionSet::interactions(
        object[rowsChromosome, ]
    )
    interactionsChromosome <- as.data.table(interactionsChromosome)
    dataplot <- base::cbind(
        interactionsChromosome[, .(
            seqnames = seqnames1,
            start1,
            start2
        )],
        assayChromosome
    )
    dataplot <- data.table::melt.data.table(
        dataplot,
        id.vars = c("seqnames", "start1", "start2"),
        value.name = "interaction",
        variable.factor = FALSE
    )
    dataplot <- dataplot[!is.na(interaction)]
    if (nrow(dataplot) == 0) {
        message("No interactions for chromosome ", chromosomeName, ".")
        return(NULL)
    }
    dataplot[, c("condition", "replicate") := data.table::tstrsplit(
        variable,
        "_",
        fixed = TRUE
    )]
    dataplot[, condition := factor(
        condition,
        levels = sort(unique(object$condition))
    )]
    dataplot[, replicate := factor(
        replicate,
        levels = sort(unique(object$replicate))
    )]
    
    regionsChromosome <- as.data.table(InteractionSet::regions(object))
    regionsChromosome <- regionsChromosome[seqnames == chromosomeName]
    xylim <- c(min(regionsChromosome$start), max(regionsChromosome$start))
    
    if(length(colours) != 3){
        stop("`colours` must be a vector of length 3.")
    } else {
        if(is.null(names(colours))) names(colours) <- c("low", "mid", "high")
        if(!identical(sort(names(colours)), c("high", "low", "mid"))){
            stop("`colours` are supposed to be named 'low', 'mid' and 'high")
        }
    }
    
    if (length(unique(object$replicate)) <= max(table(object$condition))) {
        plot <- .plotInteractionsGrid(
            dataplot,
            xylim,
            transform,
            colours,
            midpoint,
            chromosomeName
        )
    } else {
        totalLevels <- table(object$condition)
        totalCols <- max(totalLevels)
        totalRows <- length(unique(object$condition))

        existing <- by(object$replicate, object$condition, unique)
        existing <- lapply(existing, as.character)
        existing <- lapply(existing, .completeLevels, totalCols)
        existing <- mapply(
            paste,
            names(existing),
            existing,
            sep = "_",
            SIMPLIFY = FALSE
        )
        allLevels <- unlist(existing, use.names = FALSE)
        dataplot[, variable := factor(variable, levels = allLevels)]

        plot <- .plotInteractionsWrap(
            dataplot,
            xylim,
            transform,
            colours,
            midpoint,
            chromosomeName,
            totalRows,
            totalCols
        )
    }
    return(plot)
}

test_that("plotInteractions behaves as expected", {
    data(exampleHiCDOCDataSet)
    object <- reduceHiCDOCDataSet(exampleHiCDOCDataSet, 
                                  chromosomes = c("X", "Y"))
    
    expect_error(plotInteractions(object, 3), "Unknown chromosome")
    expect_error(plotInteractions(object), '"chromosome"')
    pp <- plotInteractions(object, 1)
    expect_true(ggplot2::is_ggplot(pp))
    labs <- ggplot2::get_labs(pp)
    expect_equivalent(labs[["title"]], "Chromosome X")
    expect_equivalent(labs[["x"]], "")
    expect_equivalent(labs[["y"]], "")
    expect_equivalent(labs[["z"]], "interaction")
    expect_equivalent(labs[["fill"]], "Intensity")
    # No error when printed
    expect_error(print(pp), NA)
})

#' @title
#' A/B compartment detection and differential analysis
#' @docType package
#' @import methods
#' @import ggplot2
#' @import InteractionSet
#' @importFrom GenomicRanges start end distance GRanges match union
#' @importFrom Rcpp evalCpp sourceCpp
#' @importFrom stats loess loess.control ecdf predict median update optimize prcomp p.adjust quantile wilcox.test
#' @importFrom S4Vectors DataFrame mcols split runLength Rle %in%
#' @importFrom SummarizedExperiment assay colData
#' @importFrom gtools mixedsort
#' @importFrom pbapply pbmapply
#' @importFrom BiocParallel bpparam bpmapply bplapply
#' @importFrom BiocGenerics cbind width
#' @importFrom cowplot ggdraw get_plot_component
#' @importFrom grid textGrob gpar
#' @importFrom gridExtra arrangeGrob
#' @importFrom multiHiCcompare make_hicexp cyclic_loess hic_table
#' @importFrom Seqinfo seqlevels seqnames
#' @import data.table
#' @useDynLib HiCDOC
#' @aliases HiCDOC-package
"_PACKAGE"

#' @description
#' Returns parameters, updated with their default values if invalid.
#'
#' @param objectParameters
#' A list of parameters.
#'
#' @return
#' A list of valid parameters.
#'
#' @keywords internal
#' @noRd
.validateParameters <- function(objectParameters) {
    defaultParameterNames <- names(defaultHiCDOCParameters)
    inputParameterNames <- names(objectParameters)

    numericParameters <- vapply(
        objectParameters,
        function(parameter) 
            is.numeric(parameter) && length(parameter) == 1,
        FUN.VALUE = TRUE
    )

    if (!all(numericParameters)) {
        notNumericParameters <- inputParameterNames[!numericParameters]
        warning(
            "Non-numeric parameter",
            if (length(notNumericParameters) != 1) "s were" else " was",
            " replaced with ",
            if (length(notNumericParameters) != 1) "their" else "its",
            " default",
            if (length(notNumericParameters) != 1) "s",
            ":\n",
            paste0(
                notNumericParameters,
                ":",
                " ",
                objectParameters[notNumericParameters],
                " -> ",
                defaultHiCDOCParameters[notNumericParameters],
                collapse = "\n"
            ),
            call. = FALSE
        )
        objectParameters[notNumericParameters] <- defaultHiCDOCParameters[
            notNumericParameters
        ]
    }

    known <- inputParameterNames %in% defaultParameterNames

    if (!all(known)) {
        unknownParameterNames <- inputParameterNames[!known]
        warning(
            "Unknown parameter",
            if (length(unknownParameterNames) != 1) "s were" else " was",
            " removed:\n",
            paste(unknownParameterNames, collapse = "\n"),
            call. = FALSE
        )
        objectParameters[unknownParameterNames] <- NULL
    }

    present <- defaultParameterNames %in% inputParameterNames

    if (!all(present)) {
        missingParameterNames <- defaultParameterNames[!present]
        warning(
            "Missing parameter",
            if (length(missingParameterNames) != 1) "s were" else " was",
            " replaced with ",
            if (length(missingParameterNames) != 1) "their" else "its",
            " default",
            if (length(missingParameterNames) != 1) "s",
            ":\n",
            paste0(
                missingParameterNames,
                ":",
                defaultHiCDOCParameters[missingParameterNames],
                collapse = "\n"
            ),
            call. = FALSE
        )
        objectParameters[missingParameterNames] <- defaultHiCDOCParameters[
            missingParameterNames
        ]
    }

    return(objectParameters)
}

#' @description
#' Returns valid chromosome, condition, or replicate names, from given names
#' or id.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#'
#' @param names
#' One or several names or ids to look up.
#'
#' @param category
#' The category in which to look up the names or ids. One of "chromosomes",
#' "conditions", "replicates". Defaults to "chromosomes".
#'
#' @return
#' Valid names.
#'
#' @keywords internal
#' @noRd
.validateNames <-
    function(object, names, category = "chromosomes") {
        validNames <- switch(
            category,
            "chromosomes" = unique(object@chromosomes),
            "replicates" = unique(object$replicate),
            "conditions" = unique(object$condition)
        )

        if (all(names %in% validNames)) {
            return(names)
        }

        if (is.numeric(names) && all(names %in% seq_len(length(validNames)))) {
            return(validNames[names])
        }

        unknown <- names[which(!(names %in% validNames))]

        stop(
            "Unknown ",
            substr(category, 1, nchar(category) - 1),
            if (length(unknown) != 1) "s",
            ": ",
            paste(unknown, collapse = ", "),
            call. = FALSE
        )
    }

#' @description
#' Checks that the provided object is a \code{\link{HiCDOCDataSet}}, and that
#' the provided slots are in that object.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#'
#' @param slots
#' The names of slots to verify. Defaults to NULL.
#'
#' @return
#' Raises an error if the object is not a \code{\link{HiCDOCDataSet}}, or if one
#' of the slots is missing.
#'
#' @keywords internal
#' @noRd
.validateSlots <- function(object, slots = NULL) {
    if (!is(object, "HiCDOCDataSet")) {
        stop("The provided object is not a 'HiCDOCDataSet'.", call. = FALSE)
    }

    if (
        "interactions" %in% slots && (
            !.hasSlot(object, "interactions") ||
            is.null(slot(object, "interactions")) ||
            nrow(object) == 0
        )
    ) {
        stop("No interactions found in the 'HiCDOCDataSet'.", call. = FALSE)
    }

    if (!is.null(slots)) {
        allSlots <- slotNames("HiCDOCDataSet")

        presentSlots <- allSlots[vapply(
            allSlots,
            function(x) .hasSlot(object, x) && !is.null(slot(object, x)),
            FUN.VALUE = TRUE
        )]

        missingSlots <- slots[!(slots %in% presentSlots)]

        if ("binSize" %in% missingSlots) {
            stop(
                "Resolution is unknown.\n",
                "This 'HiCDOCDataSet' wasn't built properly.",
                call. = FALSE
            )
        }

        if ("totalBins" %in% missingSlots) {
            stop(
                "Chromosome lengths are unknown.\n",
                "This 'HiCDOCDataSet' wasn't built properly.",
                call. = FALSE
            )
        }

        if ("validAssay" %in% missingSlots) {
            stop(
                "Cannot process potentially sparse replicates.\n",
                "First, run 'filterSparseReplicates()' on the object.",
                call. = FALSE
            )
        }

        if ("weakBins" %in% missingSlots) {
            stop(
                "Cannot process potentially weak positions.\n",
                "First, run 'filterWeakPositions()' on the object.",
                call. = FALSE
            )
        }

        compartmentSlots <- c(
            "compartments",
            "concordances",
            "differences",
            "distances",
            "centroids",
            "selfInteractionRatios"
        )

        if (any(compartmentSlots %in% missingSlots)) {
            stop(
                "No compartments found.\n",
                "First, run 'detectCompartments()' on the object.",
                call. = FALSE
            )
        }

        if (length(missingSlots) > 0) {
            stop(
                "Missing slots: ",
                paste(missingSlots, collapse = ", "),
                call. = FALSE
            )
        }
    }
}

#' @title
#' Plot boxplots of self interaction ratios.
#'
#' @description
#' Plots the boxplots of self interaction ratios, which are the differences
#' between self interaction and median of other interactions for each genomic
#' position. Since the A compartment is open with more interactions overall, it
#' is assumed that self interaction ratios in compartment A are smaller than in
#' compartment B.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#' @param chromosome
#' A chromosome name or index in \code{chromosomes(object)}.
#' A \code{\link{HiCDOCDataSet}}.
#' @param checks
#' Logical. Should sanity checks messages be printed on plot ? Default to TRUE. 
#'
#' @return
#' A \code{ggplot}.
#'
#' @examples
#' data(exampleHiCDOCDataSetProcessed)
#' plotSelfInteractionRatios(exampleHiCDOCDataSetProcessed, chromosome = 1)
#'
#' @export
plotSelfInteractionRatios <- function(object, chromosome, checks=TRUE) {
    .validateSlots(object, slots = c("selfInteractionRatios", "compartments"))
    chromosomeName <- .validateNames(object, chromosome, "chromosomes")
    
    compartements <- as.data.table(
        object@compartments[
            Seqinfo::seqnames(object@compartments) == chromosomeName
        ]
    )
    dataplot <- data.table::merge.data.table(
        object@selfInteractionRatios[chromosome == chromosomeName],
        compartements[, .(
            chromosome = seqnames,
            condition,
            index,
            compartment
        )],
        by = c("chromosome", "condition", "index"),
        all.x = TRUE
    )

    plot <- ggplot(
        dataplot,
        aes(x = compartment, y = ratio)
    ) + geom_jitter(
        aes(color = compartment),
        width = 0.35
    ) + geom_boxplot(
        outlier.colour = NA,
        fill = NA,
        colour = "grey20"
    ) + labs(
        color = "Compartment",
        x = "Compartment",
        y = "Interaction difference",
        title = paste0(
            "Differences between self-interactions ",
            "and other interactions"
        ),
        subtitle = paste0("Chromosome ", chromosomeName)
    )
    
    if(checks){
        messages <- .messageCheck(object, chromosomeName)
        plot <- plot + labs(caption=paste0("Quality control:\n",
                                           messages$assignment))
    }
    return(plot)
}

test_that("plotSelfInteractionRatios returns an error if no compartments", {
    data(exampleHiCDOCDataSet)
    expect_error(
        pp <- plotSelfInteractionRatios(exampleHiCDOCDataSet),
        "No compartments found."
    )
})

test_that("plotSelfInteractionRatios behaves as expected", {
    data(exampleHiCDOCDataSetProcessed)
    expect_error(
        plotSelfInteractionRatios(exampleHiCDOCDataSetProcessed), 
        '"chromosome"')
    expect_error(
        plotSelfInteractionRatios(exampleHiCDOCDataSetProcessed, 4), 
        "Unknown")

    pp <- plotSelfInteractionRatios(exampleHiCDOCDataSetProcessed, 1)
    expect_true(ggplot2::is_ggplot(pp))
    labs <- ggplot2::get_labs(pp)
    expect_equivalent(labs[["caption"]], 
        "Quality control:\nA/B assignment reliability: OK")
    expect_equivalent(labs[["x"]], "Compartment")
    expect_equivalent(labs[["y"]], "Interaction difference")
    expect_equivalent(labs[["title"]], 
        "Differences between self-interactions and other interactions")
    expect_equivalent(labs[["colour"]], "Compartment")
    expect_equivalent(labs[["subtitle"]], "Chromosome X")
    expect_is(pp$layers[[1]]$geom, "GeomPoint")
    expect_is(pp$layers[[2]]$geom, "GeomBoxplot")
    # No error when printed
    expect_error(print(pp), NA)
})

#' @description
#' Removes weak positions of a given chromosome.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#' @param chromosomeName
#' The name of a chromosome.
#' @param threshold
#' The minimum average interaction for a position to be kept.
#'
#' @return
#' A list of:
#' - The weak positions.
#' - The filtered interactions.
#'
#' @keywords internal
#' @noRd
.filterWeakPositionsOfChromosome <- function(
    chromosomeName,
    reducedObject,
    threshold
) {

    validColumns <- reducedObject@validAssay[[chromosomeName]]

    interactions <- as.data.table(InteractionSet::interactions(reducedObject))
    interactions <- interactions[, .(index1, index2)]

    # All known bins
    minBin <- min(interactions$index1, interactions$index2)
    maxBin <- max(interactions$index1, interactions$index2)
    allBins <- seq(minBin, maxBin)

    # Reducing the values of diagonal by 0.5 factor -> only upper matrix.
    diagonal <- (interactions$index1 == interactions$index2)
    matrixAssay <- SummarizedExperiment::assay(reducedObject)[,validColumns]
    matrixAssay <- matrixAssay * (1-0.5*(diagonal))

    interactions <- base::cbind(
        interactions,
        matrixAssay
    )
    interactions <- data.table::melt.data.table(
        interactions,
        id.vars = c("index1", "index2"),
        na.rm = FALSE
    )
    interactions[is.na(value),value := 0]

    totalBins <- reducedObject@totalBins[chromosomeName]
    removedBins <- allBins[
        !(allBins %in% unique(c(interactions$index1, interactions$index1)))
    ]

    totalNewWeakBins <- 1
    totalRemovedBins <- 0

    # Recursive removal of bins - deleting a bin can create a new weak bin.
    while (totalNewWeakBins > 0 && totalRemovedBins <= length(allBins)) {
        sum1 <- interactions[
            ,
            .(sum1 = sum(value, na.rm = TRUE)),
            by = .(index = index1, variable)
        ]
        sum2 <- interactions[
            ,
            .(sum2 = sum(value, na.rm = TRUE)),
            by = .(index = index2, variable)
        ]
        sum12 <- data.table::merge.data.table(
            sum1,
            sum2,
            by = c("index", "variable"),
            all = TRUE
        )
        sum12[is.na(sum1), sum1 := 0]
        sum12[is.na(sum2), sum2 := 0]
        sum12[, mean := (sum1 + sum2) / totalBins]
        weakBins <- unique(sum12[mean < threshold, index])

        totalNewWeakBins <- length(weakBins) - totalRemovedBins
        removedBins <- c(removedBins, weakBins)

        # Remove interactions of weak bins
        if (totalNewWeakBins > 0) {
            interactions <- interactions[
                !(index1 %in% weakBins | index2 %in% weakBins)
            ]
            totalRemovedBins <- totalRemovedBins + totalNewWeakBins
        }
    }

    message(
        "Chromosome ",
        chromosomeName,
        ": ",
        length(removedBins),
        " position",
        if (length(removedBins) != 1) "s",
        " removed, ",
        length(allBins) - length(removedBins),
        " position",
        if (length(allBins) - length(removedBins) != 1) "s",
        " remaining."
    )
    return(removedBins)
}

#' @title
#' Filter weak positions.
#'
#' @description
#' Removes weak genomic positions whose interactions average is lower than the
#' threshold.
#'
#' @details
#' Detects weak genomic positions in each replicate, and removes them from all
#' replicates to guarantee comparability across conditions when calling
#' \code{\link{detectCompartments}}.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#' @param threshold
#' The minimum average interaction for a position to be kept. If a position's
#' average interaction with the entire chromosome is lower than this value in
#' any of the replicates, it is removed from all replicates and conditions.
#' Defaults to \code{object$weakPositionThreshold} which is originally set to
#' \code{defaultHiCDOCParameters$weakPositionThreshold} = 1.
#'
#' @return
#' A filtered \code{\link{HiCDOCDataSet}}.
#'
#' @seealso
#' \code{\link{filterSmallChromosomes}},
#' \code{\link{filterSparseReplicates}},
#' \code{\link{HiCDOC}}
#'
#' @examples
#' data(exampleHiCDOCDataSet)
#' object <- exampleHiCDOCDataSet
#'
#' object <- filterWeakPositions(object)
#'
#' @export
filterWeakPositions <- function(object, threshold = NULL) {
    .validateSlots(
        object,
        slots = c(
            "chromosomes",
            "totalBins",
            "parameters"
        )
    )

    if (!is.null(threshold)) {
        object@parameters$weakPositionThreshold <- threshold
    }
    object@parameters <- .validateParameters(object@parameters)
    threshold <- object@parameters$weakPositionThreshold

    message(
        "Keeping positions with interactions average greater or equal to ",
        threshold,
        "."
    )

    objectChromosomes <- S4Vectors::split(
        object,
        SummarizedExperiment::mcols(object)$chromosome,
        drop = FALSE
    )

    weakBins <- pbapply::pbmapply(
        function(c, m, t) .filterWeakPositionsOfChromosome(c, m, t),
        object@chromosomes,
        objectChromosomes,
        threshold
    )

    object@weakBins <- weakBins

    indices <- as.data.table(InteractionSet::interactions(object))
    toRemove <- (
        indices$index1 %in% unlist(weakBins) |
        indices$index2 %in% unlist(weakBins)
    )
    if (sum(toRemove)>0) {
        object <- object[!toRemove,]
        object <- InteractionSet::reduceRegions(object)
        object@validAssay <- .determineValids(object)
        # Remove empty chromosomes
        leftChromosomes <- 
            names(object@validAssay[vapply(object@validAssay, 
                                           function(x) length(x)>0, 
                                           FUN.VALUE = TRUE)])
        if(!identical(leftChromosomes,object@chromosomes)){
            object <- reduceHiCDOCDataSet(object, chromosomes = leftChromosomes)
        }
    }

    totalWeakBins <- sum(vapply(weakBins, length, FUN.VALUE = 0))

    message(
        "Removed ",
        totalWeakBins,
        " position",
        if (totalWeakBins != 1) "s",
        " in total."
    )

    if (length(toRemove) == sum(toRemove)) {
        warning("No data left!", call. = FALSE)
    }
    return(object)
}

test_that("normalizeBiologicalBiases behaves as expected", {
    data(exampleHiCDOCDataSet)
    object <- filterSparseReplicates(exampleHiCDOCDataSet)
    object <- filterWeakPositions(object)
    
    # Apply normalization
    expect_message(
        norm <- normalizeBiologicalBiases(object),
        "Chromosome Z: normalizing biological biases."
    )
    expect_equal(length(norm), length(object))
    assay <- SummarizedExperiment::assay(norm)
    expect_equal(sum(!is.na(assay)), 181566)
    expect_equal(round(colSums(assay, na.rm=TRUE),2), 
                 c(179.82, 179.82, 79.29, 220.81, 
                   179.81, 179.81, 280.61))
})

data(exampleHiCDOCDataSet)
test_that("filterWeakPositions behaves as expected", {
    # Filter the 0 values in interactions
    expect_message(
        object <- filterWeakPositions(exampleHiCDOCDataSet),
        "Removed 6 positions in total."
    )
    expect_equal(nrow(object), 42646)
    expect_identical(object@weakBins,
                     list("W" = numeric(0),
                          "X" = c(112, 171, 200),
                          "Y" = c(201, 322, 346),
                          "Z" = numeric(0)))
    assay <- SummarizedExperiment::assay(object)
    expect_equal(sum(!is.na(assay)), 181180)
    expect_equal(colSums(assay, na.rm=TRUE), 
                 c(8990463, 22387622, 7564195, 29556499, 
                   17445466, 23544963, 22932825))
})

test_that("filterWeakPositions behaves as expected, custom param", {
    # Filter values with 50 threshold
    expect_message(
        object <- filterWeakPositions(exampleHiCDOCDataSet, threshold = 50),
        "Removed 123 positions in total."
    )
    expect_equal(nrow(object), 35743)
    expect_identical(sapply(object@weakBins, length),
                     c("W" = 0L, "Y" = 3L, "Z" = 0L))
    expect_identical(object@parameters$weakPositionThreshold, 50)
    assay <- SummarizedExperiment::assay(object)
    expect_equal(sum(!is.na(assay)), 146461)
    expect_equal(colSums(assay, na.rm=TRUE), 
                 c(5161309, 17717881, 7564195, 29432906, 
                   10521998, 16470630, 14974644))
})

test_that("normalizeTechnicalBiases behaves as expected", {
    data(exampleHiCDOCDataSet)
    object <- reduceHiCDOCDataSet(exampleHiCDOCDataSet, chromosomes = c("X"))
    object <- filterSparseReplicates(object)
    object <- filterWeakPositions(object)
    
    # Apply normalization
    set.seed(123)
    expect_warning(norm <- normalizeTechnicalBiases(object, parallel = FALSE))
    # Keep object format
    expect_equal(nrow(norm), nrow(object))
    assay <- SummarizedExperiment::assay(norm)
    expect_equal(sum(!is.na(assay)), 35105)
    
    # Two different possible values, 
    # 1rst with 
    # BLAS:   openblas/libblas.so.3
    # LAPACK: libopenblasp-r0.2.20.so
    # 2nd with 
    # Matrix products: default
    # BLAS:   atlas/libblas.so.3.10.3
    # LAPACK: atlas/liblapack.so.3.10.
    expect_equal(any(sapply(list(
        c(751294.6, 1152226.8, 0, 0, 721966.6, 721539.2, 910899.8),
        c(751294.6, 1152227.4, 0.0, 0.0, 721967.5, 721538.9, 910899.4)),
        function(x) identical(x, round(colSums(assay, na.rm=TRUE),1)))), TRUE
    )
})

#### chromosomes ####
#' Retrieves the vector of chromosome names.
#' @rdname HiCDOCDataSet-methods
#' @export 
setMethod("chromosomes", "HiCDOCDataSet", function(object) {
    object@chromosomes
})

#### sampleConditions ####
#' Retrieves the vector of condition names, one for each sample.
#' @rdname HiCDOCDataSet-methods
#' @export
setMethod("sampleConditions", "HiCDOCDataSet", function(object) {
    object$condition
})

#### sampleReplicates ####
#' Retrieves the vector of replicate names, one for each sample.
#' @rdname HiCDOCDataSet-methods
#' @export
setMethod("sampleReplicates", "HiCDOCDataSet", function(object) {
    object$replicate
})

#### compartments ####
#' Retrieves a \code{GenomicRange} of the compartment of every position
#' @rdname HiCDOCDataSet-methods
#' @export
setMethod("compartments", "HiCDOCDataSet", function(object, passChecks = TRUE) {
    if(passChecks == TRUE){
        compartments <- object@compartments
        compartments[compartments$assignment.check == TRUE & 
                         compartments$centroid.check == TRUE &
                         compartments$assignment.check == TRUE]
    } else {
        object@compartments
    }
})

#### differences ####
#' Retrieves a \code{GenomicRange} of the significant compartment differences
#' @rdname HiCDOCDataSet-methods
#' @export
setMethod("differences", "HiCDOCDataSet", function(object, threshold = NULL) {
    if (is.null(object@differences)) {
        return(NULL)
    }

    if (
        !is.null(threshold) && (
            !is.numeric(threshold) ||
            length(threshold) > 1
        )
    ) {
        stop("'threshold' must be a number.", call. = FALSE)
    }
    differences <- object@differences
    if (!is.null(threshold)) {
        differences <- differences[differences$pvalue.adjusted <= threshold]
    }

    if (length(differences) == 0) {
        if (is.null(threshold)) {
            message("No differences found.")
        } else {
            message(
                "No differences found with adjusted p-value <= ",
                threshold,
                "."
            )
        }
        return(NULL)
    }

    return(differences)
})

#### concordances ####
#' Retrieves a \code{GenomicRange} of the concordance (confidence in assigned
#' compartment) of every position in every replicate.
#' @rdname HiCDOCDataSet-methods 
#' @export
setMethod("concordances", "HiCDOCDataSet", function(object, passChecks = TRUE) {
    if(passChecks == TRUE){
        concordances <- object@concordances
        passingChecks <- object@checks[centroid.check == TRUE &
                                           PC1.check == TRUE &
                                           assignment.check == TRUE, chromosome]
        concordances[concordances@seqnames %in% passingChecks]
    } else {
        object@concordances
    }
})

#### show ####
#' @rdname HiCDOCDataSet-methods
#' @export
setMethod("show", "HiCDOCDataSet", function(object) {
    cat(
        "Object of class 'HiCDOCDataSet'\n\n",
        "- Inputs:\n",
        paste0(
            "  ",
            vapply(
                object@input,
                function(x) paste0(x),
                character(ifelse(
                    length(object@input) > 0,
                    length(object@input[[1]]),
                    1
                ))
            ),
            "\n"
        ),
        "\n",
        "- Chromosomes:\n  ",
        if (
            is.null(object@chromosomes) ||
            length(object@chromosomes) == 0
        ) {
            "None"
        } else {
            paste(object@chromosomes, collapse = ", ")
        },
        "\n\n",
        "- Replicates:\n",
        if (
            is.null(sampleReplicates(object)) ||
            length(sampleReplicates(object)) == 0
        ) {
            "  None\n"
        } else {
            paste0(
                "  condition ",
                sampleConditions(object),
                ", replicate ",
                sampleReplicates(object),
                "\n"
            )
        },
        "\n",
        "- Parameters:\n",
        paste0(
            "  ",
            vapply(
                seq_along(parameters(object)),
                function(x) {
                    paste(
                        names(parameters(object))[x],
                        '=',
                        parameters(object)[x]
                    )
                },
                character(1)
            ),
            "\n"
        ),
        "\n",
        "- Methods:\n",
        "  chromosomes(object)\n",
        "  sampleConditions(object)\n",
        "  sampleReplicates(object)\n",
        "  compartments(object)\n",
        "  differences(object)\n",
        "  concordances(object)\n",
        "  parameters(object)\n",
        "  parameters(object) <- list()\n\n",
        sep = ""
    )
})

#### parameters ####
#' Access the parameters of a \code{\link{HiCDOCDataSet}}.
#' @rdname HiCDOCDataSet-parameters
#' @export
setMethod("parameters", "HiCDOCDataSet", function(object) {
    object@parameters
})

#### parameters<- ####
#' Change the parameters of a \code{\link{HiCDOCDataSet}}.
#' @rdname HiCDOCDataSet-parameters
#' @export
setReplaceMethod("parameters", "HiCDOCDataSet", function(object, value) {
    defaultParameterNames <- names(defaultHiCDOCParameters)

    if (!is(value, "list")) {
        stop(
            "'parameters' must be a named list.\n",
            "No parameters were updated. ",
            "See 'help(parameters)' for details.",
            call. = FALSE
        )
    }
    parameterNames <- names(value)

    duplicatedParameterNames <- unique(
        parameterNames[duplicated(parameterNames)]
    )

    if (length(duplicatedParameterNames) > 0) {
        stop(
            "Duplicate parameter",
            if (length(duplicatedParameterNames) != 1) "s",
            " provided: ",
            paste(duplicatedParameterNames, collapse = ", "),
            "\nNo parameters were updated. ",
            "See 'help(parameters)' for details.",
            call. = FALSE
        )
    }

    invalidParameterNames <- parameterNames[
        !(parameterNames %in% defaultParameterNames)
    ]

    if (length(invalidParameterNames) > 0) {
        stop(
            "Invalid parameter",
            if (length(invalidParameterNames) != 1) "s",
            " provided: ",
            paste(invalidParameterNames, collapse = ", "),
            "\nNo parameters were updated. ",
            "See 'help(parameters)' for details.",
            call. = FALSE
        )
    }

    object@parameters[parameterNames] <- value

    return(object)
})

#' @title
#' Plot compartment changes.
#'
#' @description
#' Plots the predicted compartments, along with their concordance in each
#' replicate, and significant changes between experiment conditions.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#' @param chromosome
#' A chromosome name or index in \code{chromosomes(object)}.
#' @param threshold
#' Significance threshold for the compartment changes. Defaults to 0.05.
#' @param xlim
#' A vector of the minimum and maximum positions to display. If NULL, displays
#' all positions. Defaults to NULL.
#' @param points
#' Whether or not to add points to the concordances. Defaults to FALSE.
#' @param checks
#' Whether or not to add sanity checks messages. Default to TRUE.
#' @param colour
#' Border color for the compartments. Default to `gray90`. `NA` means no border.
#'
#'
#' @return
#' A \code{ggplot}.
#'
#' @examples
#' data(exampleHiCDOCDataSetProcessed)
#' plotCompartmentChanges(exampleHiCDOCDataSetProcessed, chromosome = 1)
#'
#' @export
plotCompartmentChanges <- function(
    object,
    chromosome,
    threshold = 0.05,
    xlim = NULL,
    points = FALSE,
    checks = TRUE,
    colour = "gray90"
    
) {

    .validateSlots(
        object,
        slots = c(
            "concordances",
            "compartments",
            "differences"
        )
    )
    chromosomeName <- .validateNames(object, chromosome, "chromosomes")
    concordancesPlot <- plotConcordances(
        object,
        chromosomeName,
        xlim,
        threshold,
        points
    )

    compartmentsPlot <- plotCompartments(
        object,
        chromosomeName,
        xlim,
        colour
    )
    
    if (is.null(compartmentsPlot) || is.null(concordancesPlot)) {
        return(NULL)
    }
    
    # Messages for the user
    captionConcordances <- concordancesPlot$labels$caption
    concordancesPlot$labels$caption <- NULL
    
    # Horizontal alignment of the sub-graphs (change width of the plots)
    plotsGrobs <- lapply(
        list(
            compartmentsPlot + theme(legend.position = "none",
                                     plot.margin = unit(c(1,0,0,0), "lines")) +
                               labs(title=NULL),
            concordancesPlot + theme(legend.position = "none",
                                     plot.margin = unit(c(0,0,0,0), "lines")) +
                               labs(title=NULL)
        ),
        ggplot2::ggplotGrob
    )

    commonWidths <- plotsGrobs[[length(plotsGrobs)]]$widths
    plotsGrobs <- lapply(
        plotsGrobs,
        function(x) {
            x$widths <- commonWidths
            return(x)
        }
    )
    
    if(checks){
        messages <- .messageCheck(object, chromosomeName)
        messages <- paste(messages, collapse = "\n")
        messages <- paste0("Quality controls:\n", messages)
        legendsGrob <- gridExtra::arrangeGrob(
            gridExtra::arrangeGrob(
                cowplot::get_plot_component(compartmentsPlot, 'guide-box-bottom', return_all = TRUE),
                cowplot::get_plot_component(concordancesPlot, 'guide-box-bottom', return_all = TRUE),
                ncol = 1,
                nrow = 2
            ),
            grid::textGrob(label=messages, x=0.9, y=0.1,
                           just=c("right", "bottom"),
                           gp=grid::gpar(fontsize=8)),
            ncol = 2,
            nrow = 1,
            padding = unit(1, "cm")
        )
    } else {
        legendsGrob <- gridExtra::arrangeGrob(
            cowplot::get_plot_component(compartmentsPlot, 'guide-box-bottom', return_all = TRUE),
            cowplot::get_plot_component(concordancesPlot, 'guide-box-bottom', return_all = TRUE),
            ncol = 2,
            nrow = 1,
            padding = unit(1, "cm")
        )
    }
    plot <- gridExtra::arrangeGrob(
                plotsGrobs[[1]],
                plotsGrobs[[2]],
                grid::textGrob(label=captionConcordances, x=0.1, y=1, 
                               just=c("left", "top"),
                               gp=grid::gpar(fontsize=8)),
                legendsGrob,
                heights = c(2, 10, 0.5, 2),
                nrow=4, 
                ncol=1,
                padding = unit(1, "lines"),
                top = paste0(
                    "Compartments and concordances of chromosome ",
                    chromosomeName, " by condition"
            )
    )
    
    return(cowplot::ggdraw(plot))
}

test_that("plotCompartmentChanges returns error if no compartments", {
    data(exampleHiCDOCDataSet)
    expect_error(
        pp <- plotCompartmentChanges(exampleHiCDOCDataSet),
        "No compartments found."
    )
})

test_that("plotCompartmentChanges behaves as expected", {
    data(exampleHiCDOCDataSetProcessed)
    expect_error(
        plotCompartmentChanges(exampleHiCDOCDataSetProcessed),
        "argument \"chromosome\""
    )
    expect_error(
        plotCompartmentChanges(exampleHiCDOCDataSetProcessed, 5), 
        "Unknown"
    )

    pp <- plotCompartmentChanges(exampleHiCDOCDataSetProcessed, 1)
    expect_true(ggplot2::is_ggplot(pp))
    labs <- ggplot2::get_labs(pp)
    expect_null(labs[["x"]])
    expect_null(labs[["y"]])
    expect_is(pp$layers[[1]]$geom, "GeomDrawGrob")
    # No error when printed
    expect_error(print(pp), NA)
})

#' @description
#' Removes sparse replicates of a given chromosome.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#' @param chromosomeName
#' The name of a chromosome.
#' @param threshold
#' The minimum percentage of non-zero interactions for a replicate to be kept.
#'
#' @return
#' A list of:
#' - The sparse condition names repeated along the sparse replicates.
#' - The sparse replicate names repeated along the sparse conditions.
#' - The filtered interactions.
#'
#' @keywords internal
#' @noRd
.filterSparseReplicatesOfChromosome <- function(
    assay,
    diagonal,
    chromosomeName,
    totalBins,
    threshold,
    validAssay,
    conditions,
    replicates
) {

    filledAssay <- diagonal * (!is.na(assay) &  assay > 0)
    filledPercentage <- colSums(filledAssay) / (totalBins * totalBins)
    toRemove <- which(filledPercentage < threshold)
    toRemove <- toRemove[toRemove %in% validAssay]

    if (length(toRemove) > 0) {
        message(
            paste(
                "\nRemoved interactions matrix of chromosome ",
                chromosomeName,
                ", condition ",
                conditions[toRemove],
                ", replicate ",
                replicates[toRemove],
                " filled at ",
                round(filledPercentage[toRemove], digits = 5) * 100,
                "%.",
                collapse = "\n",
                sep = ""
            )
        )
    }

    assay[, toRemove] <- NA
    return(assay)
}

#' @title
#' Filter sparse replicates.
#'
#' @description
#' Removes chromosome replicates whose percentage of non-zero interactions is
#' smaller than the threshold.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#' @param threshold
#' The minimum percentage of non-zero interactions for a chromosome replicate to
#' be kept. If a chromosome replicate's percentage of non-zero interactions is
#' lower than this value, it is removed. Defaults to
#' \code{object$sparseReplicateThreshold} which is originally set to
#' \code{defaultHiCDOCParameters$sparseReplicateThreshold = 30\%}.
#'
#' @return
#' A filtered \code{\link{HiCDOCDataSet}}.
#'
#' @seealso
#' \code{\link{filterSmallChromosomes}},
#' \code{\link{filterWeakPositions}},
#' \code{\link{HiCDOC}}
#'
#' @examples
#' data(exampleHiCDOCDataSet)
#' object <- exampleHiCDOCDataSet
#'
#' object <- filterSparseReplicates(object)
#'
#' @export
filterSparseReplicates <- function(object, threshold = NULL) {

    .validateSlots(
        object,
        slots = c(
            "chromosomes",
            "parameters"
        )
    )

    if (!is.null(threshold)) {
        object@parameters$sparseReplicateThreshold <- threshold
    }
    object@parameters <- .validateParameters(object@parameters)
    threshold <- object@parameters$sparseReplicateThreshold

    message(
        "Keeping replicates filled with at least ",
        threshold * 100,
        "% non-zero interactions."
    )

    diagonal <- InteractionSet::anchors(object)
    diagonal <- diagonal$first == diagonal$second
    diagonal <- 2 - 1 * diagonal
    diagonals <- S4Vectors::split(
        diagonal,
        SummarizedExperiment::mcols(object)$chromosome,
        drop = FALSE
    )

    chromosomeAssays <- S4Vectors::split(
        SummarizedExperiment::assay(object),
        SummarizedExperiment::mcols(object)$chromosome,
        drop = FALSE
    )

    resultAssay <- pbapply::pbmapply(
        function(a, d, c, t, v) {
            .filterSparseReplicatesOfChromosome(
                a, d, c, t, threshold, v, object$condition, object$replicate
            )
        },
        chromosomeAssays,
        diagonals,
        object@chromosomes,
        object@totalBins,
        object@validAssay,
        SIMPLIFY = FALSE
    )

    resultAssay <- do.call("rbind", resultAssay)
    if (nrow(resultAssay) != nrow(object)) {
        stop("Something went wrong")
    }

    SummarizedExperiment::assay(object) <- resultAssay
    newValidAssay <- .determineValids(object)
    badChromosomes <- vapply(
        newValidAssay,
        function(x) length(x) == 0,
        FUN.VALUE = TRUE
    )
    newValidAssay[badChromosomes] <- list(NULL)
    totalRemovedReplicates <- (
        length(unlist(object@validAssay)) - length(unlist(newValidAssay))
    )
    object@validAssay <- newValidAssay

    rowsToSuppress <- (
        rowSums(SummarizedExperiment::assay(object), na.rm = TRUE) == 0
    )
    if (sum(rowsToSuppress) > 0) {
        object <- object[!rowsToSuppress, ]
        object <- InteractionSet::reduceRegions(object)
        # Remove empty chromosomes
        if(sum(badChromosomes)>0){
            leftChromosomes <- object@chromosomes[!badChromosomes]
            print(leftChromosomes)
            object <- reduceHiCDOCDataSet(object, chromosomes = leftChromosomes)
        }
    }
    message(
        "Removed ",
        totalRemovedReplicates,
        " replicate",
        if (totalRemovedReplicates != 1) "s",
        " in total."
    )

    if (nrow(object) == 0) {
        warning("No data left!", call. = FALSE)
    }

    return(object)
}

data(exampleHiCDOCDataSet)
test_that("filterSparseReplicates behaves as expected", {
    # default filtering on the example dataset
    expect_message(
        object <- filterSparseReplicates(exampleHiCDOCDataSet),
        "Removed 1 replicate in total",
        fixed = TRUE
    )
    expect_equal(nrow(object), 43480)
    expect_identical(object@chromosomes, c("W", "X", "Y", "Z"))
    expect_identical(object@parameters$sparseReplicateThreshold, 0.3)
    expect_is(object@validAssay, "list")
    expect_equal(object@validAssay, 
                     list("W" = c(1, 2, 4, 5, 6, 7),
                          "X" = c(1, 2, 5, 6, 7),
                          "Y" = 1:7,
                          "Z" = c(4,7)))
    assay <- SummarizedExperiment::assay(object)
    expect_equal(sum(!is.na(assay)), 185382)
    expect_equal(colSums(assay, na.rm=TRUE), 
                 c(9217842, 23119136 , 7810636, 29607663, 
                   17856784, 24182877, 23488484))
})

test_that("filterSparseReplicates behaves as expected with custom param", {
    # Filter 1 chromosome on the example dataset
    expect_message(
        object <- filterSparseReplicates(exampleHiCDOCDataSet, 0.9995),
        "Removed 4 replicates in total.",
        fixed = TRUE
    )
    expect_equal(nrow(object), 43480)
    expect_identical(object@chromosomes, c("W", "X", "Y", "Z"))
    # expect_identical(object@weakBins, list("18" = NULL))
    expect_identical(object@parameters$sparseReplicateThreshold, 0.9995)
    assay <- SummarizedExperiment::assay(object)
    expect_equal(sum(!is.na(assay)), 153080)
    expect_equal(colSums(assay, na.rm=TRUE), 
                 c(5338026, 23119136 , 7810636, 29607663, 
                   10799881, 9615769, 23488484))
})

#### HiCDOCDataSet class definition ####
#' @title
#' \code{HiCDOCDataSet} S4 class.
#'
#' @name
#' HiCDOCDataSet-class
#' 
#' @aliases
#' HiCDOCDataSet
#' 
#' @description
#' Data structure for a Hi-C experiment.
#'
#' @details
#' An instance of \code{HiCDOCDataSet} describes a Hi-C experiment with slots
#' for path(s) to input file(s), interactions, pipeline parameters defaulting to
#' \code{defaultHiCDOCParameters}, and computation results. It can be
#' constructed from 4 different types of data:
#' - Tabular files: see \code{\link{HiCDOCDataSetFromTabular}}
#' - (m)Cool files: see \code{\link{HiCDOCDataSetFromCool}}
#' - HiC files: see \code{\link{HiCDOCDataSetFromHiC}}
#' - HiC-Pro matrices and bed files: see \code{\link{HiCDOCDataSetFromHiCPro}}
#' An example \code{HiCDOCDataSet} is also available, see
#' \code{\link{exampleHiCDOCDataSet}}.
#' The \code{HiCDOCDataSet} object can be explored using the appropriate
#' accessors.
#' 
#' @details # Accessors
#' The accessors for a HiCDOCDataset object are the following:
#' - \code{\link{chromosomes}} to retrieve the vector of chromosome names.
#' - \code{\link{sampleConditions}} to  retrieve the vector of condition names, 
#' one for each sample.
#' - \code{\link{sampleReplicates}} to retrieve the vector of replicate names, 
#' one for each sample.
#' 
#' After the detection of compartments you can use this accessors:
#' - \code{\link{compartments}} returns a GenomicRange of the compartment 
#' of every position in every condition.
#' - \code{\link{concordances}} returns a GenomicRange of the significant 
#' compartment differences between conditions, and their p-values.
#' - \code{\link{differences}} returns a GenomicRange of the concordance 
#' (confidence in assigned compartment) of every position in every replicate.
#' 
#' 
#' See the \link{HiCDOCDataSet-methods} man page for more details on methods
#' and accessors. 
#'
#' @seealso
#' \code{\link{HiCDOC}},
#' \code{\link{exampleHiCDOCDataSet}},
#' \code{\link{HiCDOCDataSetFromTabular}},
#' \code{\link{HiCDOCDataSetFromCool}},
#' \code{\link{HiCDOCDataSetFromHiC}},
#' \code{\link{HiCDOCDataSetFromHiCPro}}
#'
#' @md
#' @export
setClass(
    "HiCDOCDataSet",
    contains = "InteractionSet",
    slots = c(
        input = "ANY",
        parameters = "ANY",
        chromosomes = "ANY",
        totalBins = "ANY",
        weakBins = "ANY",
        validAssay = "ANY",
        compartments = "ANY",
        concordances = "ANY",
        differences = "ANY",
        comparisons = "ANY",
        distances = "ANY",
        centroids = "ANY",
        selfInteractionRatios = "ANY",
        checks = "ANY"
    )
)

#' @rdname HiCDOCDataSet-parameters
#' @name HiCDOCDataSet-parameters
#' @export
defaultHiCDOCParameters <- list(
    smallChromosomeThreshold = 100,
    sparseReplicateThreshold = 0.3,
    weakPositionThreshold = 1,
    cyclicLoessSpan = NA_real_,
    loessSampleSize = 20000,
    kMeansDelta = 0.0001,
    kMeansIterations = 50,
    kMeansRestarts = 20,
    PC1CheckThreshold = 0.75
)

#' @title
#' \code{\link{HiCDOCDataSet}} constructor from a tabular file.
#'
#' @description
#' Constructs a \code{\link{HiCDOCDataSet}} from a tabular file.
#'
#' @details
#' Accepts a tabular file with \code{chromosome}, \code{position 1},
#' \code{position 2}, and multiple replicate columns listing interaction counts.
#' Null interactions do not have to be listed. Values must be separated by
#' tabulations. The header must be
#' \code{chromosome position 1 position 2 x.y x.y x.y ...} with \code{x}
#' replaced by condition names and \code{y} replaced by replicate names.
#'
#' @param path
#' A path to a tabular file.
#' @param sep
#' The separator of the tabular file. Default to tabulation.
#'
#' @return
#' A \code{\link{HiCDOCDataSet}}.
#'
#' @examples
#' path <- system.file("extdata", "liver_18_10M_500000.tsv", package = "HiCDOC")
#' object <- HiCDOCDataSetFromTabular(path, sep = '\t')
#'
#' @usage
#' HiCDOCDataSetFromTabular(path, sep = '\t')
#'
#' @export
HiCDOCDataSetFromTabular <- function(
    path = NULL,
    sep = "\t"
) {

    if (!is.character(path) || length(path) > 1) {
        stop("'paths' must be a string of characters.", call. = FALSE)
    }
    if (!file.exists(path)) {
        stop("'", path, "' does not exist.", call. = FALSE)
    }

    object <- .parseTabular(path, sep = sep)
    object <- .fillHiCDOCDataSet(object)
    return(invisible(object))
}

#' @title
#' \code{\link{HiCDOCDataSet}} constructor from Cool files.
#'
#' @description
#' Constructs a \code{\link{HiCDOCDataSet}} from a set of \code{.cool} or
#' \code{.mcool} files.
#'
#' @param paths
#' A vector of paths to \code{.cool} or \code{.mcool} files.
#' @param replicates
#' A vector of replicate names repeated along the conditions.
#' @param conditions
#' A vector of condition names repeated along the replicates.
#' @param binSize
#' The resolution (span of each position in number of bases). Optionally
#' provided to select the appropriate resolution in \code{.mcool} files.
#' Defaults to NULL.
#'
#' @return
#' A \code{\link{HiCDOCDataSet}}.
#'
#' @examples
#' \dontrun{
#'     # Path to each file
#'     paths = c(
#'       'path/to/condition-1.replicate-1.cool',
#'       'path/to/condition-1.replicate-2.cool',
#'       'path/to/condition-2.replicate-1.cool',
#'       'path/to/condition-2.replicate-2.cool',
#'       'path/to/condition-3.replicate-1.cool'
#'     )
#'
#'     # Replicate and condition of each file. Can be names instead of numbers.
#'     replicates <- c(1, 2, 1, 2, 1)
#'     conditions <- c(1, 1, 2, 2, 3)
#'
#'     # Resolution to select in .mcool files
#'     binSize = 500000
#'
#'     # Instantiation of data set
#'     object <- HiCDOCDataSetFromCool(
#'       paths,
#'       replicates = replicates,
#'       conditions = conditions,
#'       binSize = binSize # Specified for .mcool files.
#'     )
#' }
#'
#' @export
HiCDOCDataSetFromCool <- function(
    paths,
    replicates,
    conditions,
    binSize = NA
) {
    if(!requireNamespace('rhdf5')) stop("'rhdf5' package is required. Please install it and retry.")
    if (is.factor(paths)) {
        paths <- as.vector(paths)
    }
    if (!is.character(paths)) {
        stop("'paths' must be a vector of characters.", call. = FALSE)
    }
    for (path in paths) {
        if (!file.exists(path)) {
            stop("'", path, "' does not exist.", call. = FALSE)
        }
    }

    if (is.factor(replicates)) {
        conditions <- as.vector(replicates)
    }
    if (is.null(replicates)) {
        stop("'replicates' must be a vector of replicates.", call. = FALSE)
    }

    if (is.factor(conditions)) {
        conditions <- as.vector(conditions)
    }
    if (is.null(conditions)) {
        stop("'conditions' must be a vector of conditions.", call. = FALSE)
    }

    if (!is.na(binSize) && (!is.numeric(binSize) || length(binSize) != 1)) {
        stop("'binSize' must be an integer.", call. = FALSE)
    }

    object <- new("HiCDOCDataSet")
    object@input <- paths
    object <- .parseCool(object, binSize, replicates, conditions)
    object <- .fillHiCDOCDataSet(object)
    return(invisible(object))
}

#' @title
#' \code{\link{HiCDOCDataSet}} constructor from HiC files.
#'
#' @description
#' Constructs a \code{\link{HiCDOCDataSet}} from a set of
#' \code{.hic} files.
#'
#' @param paths
#' A vector of paths to \code{.hic} files.
#' @param replicates
#' A vector of replicate names repeated along the conditions.
#' @param conditions
#' A vector of condition names repeated along the replicates.
#' @param binSize
#' The resolution (span of each position in number of bases) to select within
#' the \code{.hic} files.
#'
#' @return
#' A \code{\link{HiCDOCDataSet}}.
#'
#' @examples
#' \dontrun{
#'     #' # Path to each file
#'     paths = c(
#'       'path/to/condition-1.replicate-1.hic',
#'       'path/to/condition-1.replicate-2.hic',
#'       'path/to/condition-2.replicate-1.hic',
#'       'path/to/condition-2.replicate-2.hic',
#'       'path/to/condition-3.replicate-1.hic'
#'     )
#'
#'     # Replicate and condition of each file. Can be names instead of numbers.
#'     replicates <- c(1, 2, 1, 2, 1)
#'     conditions <- c(1, 1, 2, 2, 3)
#'
#'     # Resolution to select
#'     binSize <- 500000
#'
#'     # Instantiation of data set
#'     hic.experiment <- HiCDOCDataSetFromHiC(
#'       paths,
#'       replicates = replicates,
#'       conditions = conditions,
#'       binSize = binSize
#'     )
#' }
#'
#' @usage
#' HiCDOCDataSetFromHiC(paths, replicates, conditions, binSize)
#'
#' @export
HiCDOCDataSetFromHiC <- function(
    paths = NULL,
    replicates = NULL,
    conditions = NULL,
    binSize = NULL
) {

    if (is.factor(paths)) {
        paths <- as.vector(paths)
    }
    if (!is.character(paths)) {
        stop("'paths' must be a vector of characters.", call. = FALSE)
    }
    for (path in paths) {
        if (!file.exists(path)) {
            stop("'", path, "' does not exist.", call. = FALSE)
        }
    }

    if (is.factor(replicates)) {
        replicates <- as.vector(replicates)
    }
    if (is.null(replicates)) {
        stop("'replicates' must be a vector of replicates.", call. = FALSE)
    }
    if (length(replicates) != length(paths)) {
        stop("'replicates' should have the same length as 'paths'")
    }

    if (is.factor(conditions)) {
        conditions <- as.vector(conditions)
    }
    if (is.null(conditions)) {
        stop("'conditions' must be a vector of conditions.", call. = FALSE)
    }
    if (length(conditions) != length(paths)) {
        stop("'conditions' should have the same length as 'paths'")
    }

    if (!is.numeric(binSize) || length(binSize) != 1) {
        stop("'binSize' must be an integer.", call. = FALSE)
    }
    binSize <- as.integer(binSize)

    object <- new("HiCDOCDataSet")
    object@input <- paths
    object <- .parseHiC(object, binSize, replicates, conditions)
    object <- .fillHiCDOCDataSet(object)
    return(invisible(object))
}

#' @title
#' \code{\link{HiCDOCDataSet}} constructor from HiC-Pro files.
#'
#' @description
#' Constructs a \code{\link{HiCDOCDataSet}} from a set of HiC-Pro generated
#' files.
#'
#' @param matrixPaths
#' A vector of paths to HiC-Pro matrix files.
#' @param bedPaths
#' A vector of paths to HiC-Pro bed files.
#' @param replicates
#' A vector of replicate names repeated along the conditions.
#' @param conditions
#' A vector of condition names repeated along the replicates.
#'
#' @return
#' A \code{\link{HiCDOCDataSet}}.
#'
#' @examples
#' \dontrun{
#'     # Path to each matrix file
#'     matrixPaths = c(
#'       'path/to/condition-1.replicate-1.matrix',
#'       'path/to/condition-1.replicate-2.matrix',
#'       'path/to/condition-2.replicate-1.matrix',
#'       'path/to/condition-2.replicate-2.matrix',
#'       'path/to/condition-3.replicate-1.matrix'
#'     )
#'
#'     # Path to each bed file
#'     bedPaths = c(
#'       'path/to/condition-1.replicate-1.bed',
#'       'path/to/condition-1.replicate-2.bed',
#'       'path/to/condition-2.replicate-1.bed',
#'       'path/to/condition-2.replicate-2.bed',
#'       'path/to/condition-3.replicate-1.bed'
#'     )
#'
#'     # Replicate and condition of each file. Can be names instead of numbers.
#'     replicates <- c(1, 2, 1, 2, 1)
#'     conditions <- c(1, 1, 2, 2, 3)
#'
#'     # Instantiation of data set
#'     hic.experiment <- HiCDOCDataSetFromHiCPro(
#'       matrixPaths = matrixPaths,
#'       bedPaths = bedPaths,
#'       replicates = replicates,
#'       conditions = conditions
#'     )
#' }
#'
#' @usage
#' HiCDOCDataSetFromHiCPro(matrixPaths, bedPaths, replicates, conditions)
#'
#' @export
HiCDOCDataSetFromHiCPro <- function(
    matrixPaths = NULL,
    bedPaths = NULL,
    replicates = NULL,
    conditions = NULL
) {

    if (is.factor(matrixPaths)) {
        matrixPaths <- as.vector(matrixPaths)
    }
    if (!is.character(matrixPaths)) {
        stop("'matrixPaths' must be a vector of characters.", call. = FALSE)
    }

    if (is.factor(bedPaths)) {
        bedPaths <- as.vector(bedPaths)
    }
    if (!is.character(bedPaths)) {
        stop("'bedPaths' must be a vector of characters.", call. = FALSE)
    }

    if (length(matrixPaths) != length(bedPaths)) {
        stop(
            "'matrixPaths' and 'bedPaths' must have the same length.",
             call. = FALSE
        )
    }

    paths <-
        base::split(
            base::cbind(matrixPaths, bedPaths),
            seq(length(matrixPaths))
        )

    for (path in unlist(paths)) {
        if (!file.exists(path)) {
            stop("'", path, "' does not exist.", call. = FALSE)
        }
    }

    if (is.factor(replicates)) {
        replicates <- as.vector(replicates)
    }
    if (is.null(replicates)) {
        stop("'replicates' must be a vector of replicates.", call. = FALSE)
    }

    if (is.factor(conditions))
        conditions <- as.vector(conditions)
    if (is.null(conditions)) {
        stop("'conditions' must be a vector of conditions.", call. = FALSE)
    }

    if (length(conditions) != length(replicates)) {
        stop(
            "'conditions' and 'replicates' must have the same length",
             call. = FALSE
        )
    }

    object <- new("HiCDOCDataSet")
    object@input <- paths
    object <- .parseHiCPro(object, replicates, conditions)
    object <- .fillHiCDOCDataSet(object)
    return(invisible(object))
}

#' @title
#' Default pipeline to run on the HiCDOC analysis.
#'
#' @description
#' Runs the default filtering, normalization, and computational steps on a
#' \code{HiCDOCDataSet}. To learn more about HiCDOC, browse the vignette:
#' \code{browseVignettes(package = "HiCDOC")}.
#'
#' @details
#' \subsection{\code{HiCDOC} pipeline}{
#'     The HiCDOC pipeline has seven steps: \\
#'     \describe{
#'         Three filtering steps:
#'         \itemize{
#'             \item{\code{\link{filterSmallChromosomes}}} to filter out 
#'             small chromosomes
#'             \item{\code{\link{filterWeakPositions}}} to filter out weak 
#'             positions with very few interactions
#'             \item{\code{\link{filterSparseReplicates}}} to filter out 
#'             sparse replicates with many null interactions
#'         }
#'         Three normalization steps:
#'         \itemize{
#'             \item{\code{\link{normalizeTechnicalBiases}}} to normalize 
#'             technical biases in each replicates
#'             \item{\code{\link{normalizeBiologicalBiases}}} to normalize 
#'             biological biases in each replicate
#'             \item{\code{\link{normalizeDistanceEffect}}} to normalize the 
#'             distance effect in each chromosome
#'         }
#'         One computational step:
#'         \itemize{
#'             \item{\code{\link{detectCompartments}}} to detect compartments 
#'             in each condition and find significant changes between 
#'             conditions.
#'         }
#'     }
#' }
#' \subsection{Parallel processing}{
#' The parallel version of HiCDOC uses the
#' \code{\link{BiocParallel}} package. Before to call the
#' function in parallel you should specify the parallel parameters such as:
#'     \itemize{
#'         \item{On Linux:}
#'            \code{multiParam <- BiocParallel::MulticoreParam(workers = 10)}
#'          \item{On Windows:}
#'              \code{multiParam <- BiocParallel::SnowParam(workers = 10)}
#'     }
#'     And then you can register the parameters to be used by BiocParallel: \\
#'     \code{BiocParallel::register(multiParam, default = TRUE)}
#'
#'     You should be aware that using MulticoreParam, reproducibility of the
#'     detectCompartments function using a RNGseed may not work. See this
#'     \href{https://github.com/Bioconductor/BiocParallel/issues/122}{issue}
#'     for more details.
#' }
#'
#' @param object
#' A \code{HiCDOCDataSet}.
#' @param parallel
#' Whether or not to parallelize each step. Defaults to FALSE.
#'
#' @return
#' A HiCDOCDataSet with all slots filled.
#'
#' @seealso
#' \code{\link{HiCDOCDataSet}}, \code{\link{filterSmallChromosomes}},
#' \code{\link{filterWeakPositions}}, \code{\link{filterSparseReplicates}},
#' \code{\link{normalizeTechnicalBiases}},
#' \code{\link{normalizeBiologicalBiases}},
#' \code{\link{normalizeDistanceEffect}},
#' \code{\link{detectCompartments}}
#'
#' @examples
#' data(exampleHiCDOCDataSet)
#' # Default HiCDOC pipeline
#' # Not printing loess warnings for example purpose. 
#' # Results should be inspected if there is any.
#' suppressWarnings(
#' object <- HiCDOC(exampleHiCDOCDataSet)
#' )
#' 
#' # Equivalent to
#' if(FALSE){
#'     object <- filterSmallChromosomes(exampleHiCDOCDataSet)
#'     object <- filterSparseReplicates(object)
#'     object <- filterWeakPositions(object)
#'     object <- normalizeTechnicalBiases(object)
#'     object <- normalizeBiologicalBiases(object)
#'     object <- normalizeDistanceEffect(object)
#'     object <- detectCompartments(object)
#' }
#'
#' @export
HiCDOC <- function(
    object,
    parallel = FALSE
) {

    object <- filterSmallChromosomes(object)
    object <- filterSparseReplicates(object)
    object <- filterWeakPositions(object)
    object <- normalizeTechnicalBiases(object, parallel = parallel)
    object <- normalizeBiologicalBiases(object, parallel = parallel)
    object <- normalizeDistanceEffect(object)
    object <- detectCompartments(object, parallel = parallel)
    return(invisible(object))
}

test_that("HiCDOCDataSetFromHic works as expected", {
    paths <-
        system.file("extdata", "liver_18_10M.hic", package = "HiCDOC")

    # Replicate and condition of each file. Can be names instead of numbers.
    replicates <- 1
    conditions <- 1
    binSize <- 500000

    # Instantiation of data set
    expect_message(
        object <- HiCDOCDataSetFromHiC(
            paths,
            replicates = replicates,
            conditions = conditions,
            binSize = binSize
        ),
        "liver_18_10M.hic")
    expect_equal(length(object), 210)
    expect_identical(object@chromosomes, "18")
    expect_identical(object$condition, 1)
    expect_identical(object$replicate, 1)
})

test_that("HiCDOCDataSetFromPro works as expected", {
    # Path to each matrix file
    matrixPaths <-
        system.file("extdata", "liver_18_10M_500000.matrix", package = "HiCDOC")

    # Path to each bed file
    bedPaths <-
        system.file("extdata", "liver_18_10M_500000.bed", package = "HiCDOC")

    # Replicate and condition of each file. Can be names instead of numbers.
    replicates <- 1
    conditions <- 1

    # Instantiation of data set
    expect_message(
        object <- HiCDOCDataSetFromHiCPro(
            matrixPaths = matrixPaths,
            bedPaths = bedPaths,
            replicates = replicates,
            conditions = conditions
        ),
        "liver_18_10M_500000.matrix")

    expect_equal(length(object), 210)
    expect_identical(object@chromosomes, "18")
    expect_identical(object$condition, 1)
    expect_identical(object$replicate, 1)
})

test_that("HiCDOCDataSetFromTabular produce correct format", {
    path <- system.file(
        "extdata",
        "liver_18_10M_500000.tsv",
        package = "HiCDOC"
    )
    expect_error(object <- HiCDOCDataSetFromTabular(path), NA)
    # Class and slots
    expect_is(object, "HiCDOCDataSet")
    expect_identical(
        slotNames(object),
        c("input", "parameters", "chromosomes", 
          "totalBins", "weakBins", "validAssay",
          "compartments", "concordances", "differences", "comparisons", 
          "distances", "centroids", "selfInteractionRatios", "checks", "interactions", 
          "colData", "assays", "NAMES", "elementMetadata", "metadata"
        )
    )
    # Class of slots
    expect_is(object@input, "character")
    expect_is(object@weakBins, "list")
    expect_is(object@validAssay, "list")
    expect_is(object@chromosomes, "character")
    expect_is(object@totalBins, "numeric")
    expect_is(object@distances, "NULL")
    expect_is(object@selfInteractionRatios, "NULL")
    expect_is(object@compartments, "NULL")
    expect_is(object@concordances, "NULL")
    expect_is(object@differences, "NULL")
    expect_is(object@centroids, "NULL")
    expect_is(object@parameters, "list")
    expect_is(object@checks, "NULL")
    expect_is(object, "InteractionSet")
    
    # Interactions
    expect_is(SummarizedExperiment::assay(object), "matrix")
    expect_is(InteractionSet::regions(object), "GRanges")
    expect_is(InteractionSet::interactions(object), "StrictGInteractions")
    expect_is(S4Vectors::mcols(object), "DataFrame")
    expect_is(SummarizedExperiment::assay(object), "matrix")
    expect_equal(is.numeric(SummarizedExperiment::assay(object)), TRUE)
})

test_that("HiCDOCDalinkToMatrixtaSetFromTabular produce correct values", {
    path <- system.file(
        "extdata",
        "liver_18_10M_500000.tsv",
        package = "HiCDOC"
    )
    expect_error(object <- HiCDOCDataSetFromTabular(path), NA)
    gi <- InteractionSet::interactions(object)
    assays <- SummarizedExperiment::assay(object)
    
    # Interactions
    expect_equal(nrow(assays), 210)
    expect_equal(mean(gi@anchor1), 7.333333, tolerance = 1e-5)
    expect_equal(mean(gi@anchor2), 13.66667, tolerance = 1e-5)
    expect_equal(mean(assays), 484.019, tolerance = 1e-5)
    
    # chromosomes
    expect_identical(object@chromosomes, "18")
    # bins
    expect_equal(object@totalBins, c("18" = 20))
    # Parameters
    expect_identical(object@parameters, defaultHiCDOCParameters)
    # Positions
    regions <- data.frame(InteractionSet::regions(object))
    expect_equal(mean(regions$index), 10.5, tolerance = 1e-5)
    expect_equal(mean(regions$start), 4750001)
    expect_equal(mean(regions$end), 5250000)
})

test_that("HiCDOCDataSetFromCool works as expected", {
    paths <-
        system.file("extdata", "liver_18_10M_500000.cool", package = "HiCDOC")

    # Replicate and condition of each file. Can be names instead of numbers.
    replicates <- "GG"
    conditions <- "AA"

    # Instantiation of data set
    expect_message(
        object <- HiCDOCDataSetFromCool(
            paths,
            replicates = replicates,
            conditions = conditions
        ),
        "liver_18_10M_500000.cool'")
    matAssay <- SummarizedExperiment::assay(object) 
    expect_equal(dim(matAssay), c(210, 1))
    expect_identical(object@chromosomes, "18")
    expect_identical(object$replicate, c("GG"))
    expect_identical(object$condition, c("AA"))
})


test_that("HiCDOCDataSetFromCool works as expected if mcool", {
    paths <-
        system.file("extdata", "liver_18_10M.mcool", package = "HiCDOC")

    # Replicate and condition of each file. Can be names instead of numbers.
    replicates <- "A"
    conditions <- "C"

    # Resolution to select in .mcool files
    binSize <- 500000

    # Instantiation of data set
    expect_message(
        object <- HiCDOCDataSetFromCool(
            paths,
            replicates = replicates,
            conditions = conditions,
            binSize = binSize
        ),
        "liver_18_10M.mcool")
    
    matAssay <- SummarizedExperiment::assay(object) 
    expect_equal(dim(matAssay), c(210, 1))
    expect_identical(object@chromosomes, "18")
    expect_identical(object$replicate, c("A"))
    expect_identical(object$condition, c("C"))
})

#' @title
#' Plot the distribution of concordance differences.
#'
#' @description
#' Plots the distribution of concordance differences, which are the differences
#' between concordances of each pair of replicates from different conditions. A
#' concordance can be understood as a confidence in a genomic position's
#' assigned compartment. Mathematically, it is the log ratio of a genomic
#' position's distance to each compartment's centroid, normalized by the
#' distance between both centroids, and min-maxed to a [-1,1] interval.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#' @return
#' A \code{ggplot}.
#'
#' @examples
#' data(exampleHiCDOCDataSetProcessed)
#' plotConcordanceDifferences(exampleHiCDOCDataSetProcessed)
#'
#' @export
plotConcordanceDifferences <- function(object) {
    .validateSlots(object, slots = c("comparisons"))

    differences <- object@comparisons[, changed := data.table::fifelse(
        compartment.1 == compartment.2,
        "FALSE",
        "TRUE"
    )]

    plot <- ggplot(
        differences,
        aes(x = difference, fill = changed)
    ) + geom_histogram() + labs(
        x = "Concordance",
        title = "Distribution of concordance differences",
        fill = "Change\nof\ncompartment"
    )
    return(plot)
}

test_that("plotConcordanceDifferences returns an error of no compartments", {
    data(exampleHiCDOCDataSet)
    expect_error(
        pp <- plotConcordanceDifferences(exampleHiCDOCDataSet),
        "Missing slots: comparisons"
    )
})

test_that("plotConcordanceDifferences behaves as expected", {
    data(exampleHiCDOCDataSetProcessed)
    expect_error(
        pp <- plotConcordanceDifferences(exampleHiCDOCDataSetProcessed),
        NA
    )
    labs <- ggplot2::get_labs(pp)
    expect_equivalent(labs[["x"]], "Concordance")
    expect_equivalent(labs[["fill"]], "Change\nof\ncompartment")
    expect_equivalent(labs[["title"]], 
                      "Distribution of concordance differences")
    expect_equivalent(labs[["y"]], "count")
    expect_equivalent(labs[["weight"]], "weight")
    
    expect_is(pp$layers[[1]]$geom, "GeomBar")
    # No error when printed
    expect_error(print(pp), NA)
})

#" Based on
#" https://github.com/dozmorovlab/HiCcompare/blob/master/R/KRnormalization.R
#"
#' @description
#' Applies the Knight-Ruiz balancing algorithm to transform the provided
#' matrix into a doubly stochastic matrix, with sum of each row and sum of each
#' column equal to 1.
#'
#' @param m
#' A matrix.
#'
#' @return
#' The transformed matrix.
#'
#' @keywords internal
#' @noRd
.normalizeKnightRuiz <- function(cm) {
    m <- cm@matrix
    m[is.na(m)] <- 0
    tol <- 1e-6
    minDelta <- 0.1
    maxDelta <- 3
    n <- nrow(m)
    e <- matrix(1, nrow = n, ncol = 1)
    x0 <- e
    g <- 0.9
    etamax <- 0.1
    eta <- etamax
    stop_tol <- tol * .5
    x <- x0
    rt <- tol^2
    v <- x * (m %*% x)
    rk <- 1 - v
    rho_km1 <- drop(t(rk) %*% rk)
    rout <- rho_km1
    rold <- rout
    while (rout > rt) {
        k <- 0
        y <- e
        innertol <- max(c(eta^2 * rout, rt))
        while (rho_km1 > innertol) {
            k <- k + 1
            if (k == 1) {
                Z <- rk / v
                p <- Z
                rho_km1 <- drop(t(rk) %*% Z)
            } else {
                beta <- rho_km1 / rho_km2
                p <- Z + beta * p
            }
            w <- x * (m %*% (x * p)) + v * p
            if (max(w) == Inf) {
                warning("KR algorithm diverges.", call. = FALSE)
                return(t(t(x[, 1] * m) * x[, 1]))
            }
            alpha <- rho_km1 / drop(t(p) %*% w)
            ap <- alpha * p
            ynew <- y + ap
            if (min(ynew) <= minDelta) {
                if (minDelta == 0) break()
                ind <- which(ap < 0)
                gamma <- min((minDelta - y[ind]) / ap[ind])
                y <- y + gamma * ap
                break()
            }
            if (max(ynew) >= maxDelta) {
                ind <- which(ynew > maxDelta)
                gamma <- min((maxDelta - y[ind]) / ap[ind])
                y <- y + gamma * ap
                break()
            }
            y <- ynew
            rk <- rk - alpha * w
            rho_km2 <- rho_km1
            Z <- rk / v
            rho_km1 <- drop(t(rk) %*% Z)
        }
        x <- x * y
        v <- x * (m %*% x)
        rk <- 1 - v
        rho_km1 <- drop(t(rk) %*% rk)
        rout <- rho_km1
        rat <- rout / rold
        rold <- rout
        res_norm <- sqrt(rout)
        eta_o <- eta
        eta <- g * rat
        if (g * eta_o^2 > 0.1) eta <- max(c(eta, g * eta_o^2))
        eta <- max(c(min(c(eta, etamax)), stop_tol / res_norm))
    }
    result <- t(t(x[, 1] * m) * x[, 1])
    cm@matrix <- result
    return(cm)
}

#' @description
#' Normalizes biological biases in the interactions of a given chromosome. Calls
#' \code{.normalizeKnightRuiz} internally.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#' @param chromosomeName
#' The name of a chromosome to normalize.
#'
#' @return
#' A data.table of normalized interactions.
#'
#' @keywords internal
#' @noRd
.normalizeBiologicalBiasesOfChromosome <- function(object) {

    chromosomeName <- as.character(
      SummarizedExperiment::mcols(object)$chromosome[1]
    )
    message("Chromosome ", chromosomeName, ": normalizing biological biases.")
    if (object@totalBins[[chromosomeName]] <= 0) {
        return(NULL)
    }
    currentOrder <- InteractionSet::anchorIds(object)
    currentAssay <- SummarizedExperiment::assay(object)

    # Pass by InteractionSet so we can use inflate/deflate
    chromosomeInteractionSet <- InteractionSet::InteractionSet(
        currentAssay,
        InteractionSet::interactions(object)
    )
    validAssay <- object@validAssay[[chromosomeName]]

    matrices <- lapply(
        validAssay,
        FUN = function(x) {
            InteractionSet::inflate(
                chromosomeInteractionSet,
                rows = chromosomeName,
                columns = chromosomeName,
                sample = x,
                sparse = FALSE
            )
        }
    )
    matrices <- lapply(
        matrices,
        function(m) {
            m@matrix[is.na(m@matrix)] <- 0
            return(m)
        }
    )

    matrices <- lapply(matrices, .normalizeKnightRuiz)
    matrices <- lapply(
        matrices,
        function(m) {
            m@matrix[m@matrix == 0] <- NA
            return(m)
        }
    )
    matrices <- lapply(matrices, InteractionSet::deflate, use.na = TRUE)

    ids <- InteractionSet::anchorIds(matrices[[1]])
    ids <- paste(ids$first, ids$second)
    correctIds <- paste(currentOrder$first, currentOrder$second)
    orderids <- match(correctIds, ids)
    matrices <- lapply(matrices, function(x) SummarizedExperiment::assay(x))
    matrices <- lapply(matrices, function(x) x[orderids, ])
    matrices <- do.call(base::"cbind", matrices)
    currentAssay[, validAssay] <- matrices
    return(currentAssay)
}

#' @title
#' Normalize biological biases.
#'
#' @description
#' Normalizes biological biases such as GC content and repeated regions. Uses
#' the Knight-Ruiz balancing algorithm to transform interaction matrices into
#' doubly stochastic matrices, with sum of each row and sum of each column equal
#' to 1.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#' @param parallel
#' Should the normalization be run in parallel mode? Default to FALSE.
#'
#' @return
#' A \code{\link{HiCDOCDataSet}} with normalized interactions.
#'
#' @examples
#' data(exampleHiCDOCDataSet)
#' object <- exampleHiCDOCDataSet
#' object <- filterSparseReplicates(object)
#' object <- filterWeakPositions(object)
#' object <- normalizeBiologicalBiases(object)
#'
#' @seealso
#' \code{\link{filterSparseReplicates}},
#' \code{\link{filterWeakPositions}},
#' \code{\link{normalizeTechnicalBiases}},
#' \code{\link{normalizeDistanceEffect}},
#' \code{\link{HiCDOC}}
#'
#' @export
normalizeBiologicalBiases <- function(object, parallel = FALSE) {

    .validateSlots(
        object,
        slots = c(
            "totalBins",
            "validAssay"
        )
    )
    objectChromosomes <- S4Vectors::split(
        object,
        SummarizedExperiment::mcols(object)$chromosome,
        drop = FALSE
    )

    normAssay <- .internalLapply(
        parallel,
        objectChromosomes,
        FUN = .normalizeBiologicalBiasesOfChromosome
    )
    normAssay <- do.call("rbind", normAssay)

    SummarizedExperiment::assay(object) <- normAssay

    return(object)
}

#' @description
#' Fill \code{\link{InteractionSet}} with possibly missing values
#'
#' @param interactionSet
#' An \code{\link{InteractionSet}}.
#' @param interactionSetUnion
#' The full \code{\link{InteractionSet}}.
#' @param fill
#' Fill missing values with this.
#'
#' @return
#' The full \code{\link{InteractionSet}}.
#'
#' @keywords internal
#' @noRd
.fillInteractionSet <- function(
    interactionSet,
    interactionSetUnion,
    fill = NA
) {
    over <- GenomicRanges::match(interactionSet, interactionSetUnion)
    totalColumns <- ncol(interactionSet)
    newAssays <- matrix(
        rep(fill, length(interactionSetUnion) * totalColumns),
        ncol = totalColumns
    )
    newAssays[over, ] <- SummarizedExperiment::assay(interactionSet)
    return(
        InteractionSet::InteractionSet(
            newAssays,
            interactionSetUnion,
            colData = SummarizedExperiment::colData(interactionSet)
        )
    )
}

#' @description
#' Format the outputs produced by \code{detectCompartements}.
#' @param object
#' a HiCDOCDataSet object
#' @return
#' a HiCDOCDataSet object
#'
#' @keywords internal
#' @noRd
.formatDetectCompartment <- function(object) {
    chromosomeNames <- object@chromosomes
    conditionNames <- sort(unique(object$condition))
    replicateNames <- sort(unique(object$replicate))

    all.regions <- InteractionSet::regions(object)

    # Concordances
    object@concordances[, `:=`(
        chromosome = factor(chromosome, levels = chromosomeNames),
        replicate = factor(replicate, levels = replicateNames)
    )]
    concordances <- object@concordances
    object@concordances <- all.regions[
        match(concordances$index, S4Vectors::mcols(all.regions)$index)
    ]
    S4Vectors::mcols(object@concordances) <- S4Vectors::DataFrame(
        concordances[, .(
            index,
            condition,
            replicate,
            compartment,
            concordance
        )]
    )

    # Centroids
    object@centroids[, `:=`(
        chromosome = factor(chromosome, levels = chromosomeNames),
        condition = factor(condition, levels = conditionNames)
    )]

    # Differences
    object@differences <- data.table::merge.data.table(object@differences,
        object@checks, by = "chromosome")
    object@differences[, chromosome := factor(
        chromosome,
        levels = chromosomeNames
    )]
    object@differences[, significance := ""]
    object@differences[pvalue.adjusted <= 0.05, significance := "*"]
    object@differences[pvalue.adjusted <= 0.01, significance := "**"]
    object@differences[pvalue.adjusted <= 0.001, significance := "***"]
    object@differences[pvalue.adjusted <= 0.0001, significance := "****"]

    differences <- object@differences
    object@differences <- all.regions[
        match(differences$index, S4Vectors::mcols(all.regions)$index)
    ]
    S4Vectors::mcols(object@differences) <- S4Vectors::DataFrame(
        differences[, .(
            index,
            condition.1,
            condition.2,
            pvalue,
            pvalue.adjusted,
            direction,
            significance,
            centroid.check,
            PC1.check,
            assignment.check
        )]
    )

    # Compartments
    object@compartments[, chromosome := factor(
        chromosome,
        levels = chromosomeNames
    )]
    object@compartments <- data.table::merge.data.table(object@compartments,
        object@checks, by = "chromosome")
    compartments <- object@compartments
    object@compartments <- all.regions[
        match(compartments$index, S4Vectors::mcols(all.regions)$index)
    ]
    S4Vectors::mcols(object@compartments) <- S4Vectors::DataFrame(
        compartments[, .(
            index,
            condition,
            compartment,
            centroid.check,
            PC1.check,
            assignment.check
        )]
    )

    # Distances
    object@distances[, `:=`(
        chromosome = factor(chromosome, levels = chromosomeNames),
        condition = factor(condition, levels = conditionNames),
        replicate = factor(replicate, levels = replicateNames)
    )]

    # Comparisons
    object@comparisons[, chromosome := factor(
        chromosome,
        levels = chromosomeNames
    )]

    # selfInteractionRatios
    object@selfInteractionRatios[, `:=`(
        chromosome = factor(chromosome, levels =  chromosomeNames),
        condition = factor(condition, levels = conditionNames),
        replicate = factor(replicate, levels = replicateNames)
    )]

    return(object)
}

#' @description
#' Parses interactions in tabular format and fills the conditions, replicates,
#' and interactions slots of the provided \code{\link{InteractionSet}}.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#' @param input
#' The name of the input file
#' @param conditions
#' The names of the conditions
#' (infered from the table if not specified).
#' @param replicates
#' The names of the replicates
#' (infered from the table if not specified).
#'
#' @return
#' An \code{\link{InteractionSet}}.
#'
#' @keywords internal
#' @noRd
.setFromTabular <- function(tabular, conditions = NULL, replicates = NULL) {

    if (colnames(tabular)[1] != "chromosome") {
        stop(
            "First column of the input file must be named 'chromosome'.",
            call. = FALSE
        )
    }
    if (colnames(tabular)[2] == "position.1") {
        data.table::setnames(tabular, "position.1", "position 1")
    }
    if (colnames(tabular)[2] != "position 1") {
        stop(
            "Second column of the input file must be named 'position 1'.",
            call. = FALSE
        )
    }
    if (colnames(tabular)[3] == "position.2") {
        data.table::setnames(tabular, "position.2", "position 2")
    }
    if (colnames(tabular)[3] != "position 2") {
        stop(
            "Third column of the input file must be named 'position 2'.",
            call. = FALSE
        )
    }
    if (is.null(conditions) != is.null(replicates)) {
        stop(
            "Conditions and replicates should be both NULL, or none.",
            call. = FALSE
        )
    }
    tabular[, chromosome := as.character(chromosome)]
    tabular[, chromosome := factor(
        chromosome,
        levels = gtools::mixedsort(unique(chromosome)))
    ]
    setorder(tabular, chromosome, `position 1`, `position 2`)
    # Assays part, fill with NA
    assays <- as.matrix(tabular[,4:ncol(tabular), drop = FALSE])

    if (!is.null(conditions) | !is.null(replicates)) {
        if (
            (length(conditions) != ncol(assays)) |
            (length(replicates) != ncol(assays))
        ) {
            stop(
                "Number of conditions and replicates should match the number ",
                "of counts in the matrix.",
                call. = FALSE
            )
        }
    } else {
        if (!all(grepl("^.+?\\..+$", colnames(assays)))) {
            stop(
                "Fourth to last column of the input file must be named 'C.R', ",
                "with C the condition number/name and R the replicate ",
                "number/name.",
                call. = FALSE
            )
        }
    }
    assays[assays == 0] <- NA

    # GInteraction part
    tabular <- tabular[, .(chromosome, `position 1`, `position 2`)]
    data.table::setnames(tabular, "position 1", "bin.1")
    data.table::setnames(tabular, "position 2", "bin.2")

    diagonal <- (tabular$bin.1 == tabular$bin.2)
    binSize <- .modeVector(abs(
        tabular[!diagonal,]$bin.1 - tabular[!diagonal,]$bin.2
    ))

    tabular[, bin.1 := bin.1/binSize]
    tabular[, bin.2 := bin.2/binSize]

    allRegions <- data.table::melt(
        tabular[, .(chromosome, bin.1, bin.2)],
        id.vars = "chromosome",
        value.name = "indexC"
    )
    allRegions[, variable := NULL]
    allRegions <- unique(allRegions)
    setorder(allRegions, chromosome, indexC)

    # Constructing unique index for all chromosomes,
    # taking into account the difference in bins.
    allRegions[
        ,
        index := indexC - data.table::shift(indexC, fill = 0),
        by = .(chromosome)
    ]
    allRegions[index==0, index:=1]
    allRegions[, index := cumsum(index)]
    allRegions[, end := (indexC+1) * binSize]
    allRegions[, start := (indexC) * binSize + 1]
    data.table::setcolorder(
        allRegions,
        c("chromosome", "start", "end", "index", "indexC")
    )

    tabular <- data.table::merge.data.table(
        tabular,
        allRegions[, .(chromosome, startIndex = index, bin.1 = indexC)],
        all.x = TRUE,
        sort = FALSE,
        by = c("chromosome", "bin.1")
    )
    tabular <- data.table::merge.data.table(
        tabular,
        allRegions[, .(chromosome, stopIndex = index, bin.2 = indexC)],
        all.x = TRUE,
        sort = FALSE,
        by = c("chromosome", "bin.2")
    )
    tabular[, bin.1 := NULL]
    tabular[, bin.2 := NULL]

    allRegions[, indexC := NULL]
    order1 <- match(tabular$startIndex, allRegions$index)
    order2 <- match(tabular$stopIndex, allRegions$index)
    allRegions <- GenomicRanges::GRanges(allRegions)

    gi <- InteractionSet::GInteractions(
        allRegions[order1],
        allRegions[order2],
        regions = allRegions,
        mode="strict"
    )
    
    if (is.null(conditions)) {
        conditions <- gsub("^(.+?)\\..+$", "\\1", colnames(assays))
    } 
    if(is.null(replicates)) {
        replicates <- gsub("^.+?\\.(.+)$", "\\1", colnames(assays))
    }
  
    interactionSet <- .createInteractionSet(assays, gi, allRegions, conditions, replicates)
    return(interactionSet)
}

#' @description
#' Read the file, and fills it using \code{\link{.setFromTabular}}.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#' @param sep
#' The separator of the tabular file.
#'
#' @return
#' A filled \code{\link{HiCDOCDataSet}}.
#'
#' @keywords internal
#' @noRd
.parseTabular <- function(input, sep = "\t") {

    message("Parsing '", input, "'.")

    interactions <- data.table::fread(
        file = input,
        sep = sep,
        header = TRUE,
        # comment.char = "#",
        check.names = FALSE,
        data.table = TRUE,
        stringsAsFactors = FALSE
    )

    interactionSet <- .setFromTabular(interactions)
    object <- new("HiCDOCDataSet", interactionSet, input = input)

    return(object)
}

#' @description
#' Parses a single interactions file in \code{.cool} or \code{.mcool} format.
#'
#' @param path
#' The path to the interactions file.
#' @param binSize
#' The resolution (span of each position in number of bases). Optionally
#' provided to select the appropriate resolution in \code{.mcool} files.
#' Defaults to NULL.
#'
#' @return
#' A data.table of interactions.
#'
#' @keywords internal
#' @noRd
.parseOneCool <- function(path, binSize = NA, replicate, condition) {

    message("\nParsing '", path, "'.")

    uri <- function(path) {
        if (!is.numeric(binSize)) return(path)
        return(
            paste(
                "resolutions",
                format(binSize, scientific = FALSE),
                path,
                sep = "/"
            )
        )
    }

    bins <- data.table::data.table(
        chromosome = factor(
            rhdf5::h5read(file = path, name = uri("bins/chrom"))
        ),
        start = rhdf5::h5read(file = path, name = uri("bins/start")),
        end = rhdf5::h5read(file = path, name = uri("bins/end"))
    )
    bins[, start := as.integer(start)]
    bins[, start := start+1]
    bins[, end := as.integer(end)]

    setorder(bins, chromosome, start, end)
    bins[, index := seq_len(nrow(bins))]

    interactions <- data.table::data.table(
        id1 = rhdf5::h5read(file = path, name = uri("pixels/bin1_id")),
        id2 = rhdf5::h5read(file = path, name = uri("pixels/bin2_id")),
        interaction = rhdf5::h5read(file = path, name = uri("pixels/count"))
    )
    interactions[, id1 := as.integer(id1) + 1]
    interactions[, id2 := as.integer(id2) + 1]
    interactions[, interaction := as.numeric(interaction)]

    order1 <- match(interactions$id1, bins$index)
    order2 <- match(interactions$id2, bins$index)
    allRegions <- GenomicRanges::GRanges(bins)

    # GInteractions part
    gi <- InteractionSet::GInteractions(
        allRegions[order1],
        allRegions[order2],
        regions = allRegions,
        mode="strict"
    )
    assay <- as.matrix(interactions$interaction, ncol = 1)
    
    interactionSet <- .createInteractionSet(assay, gi, allRegions, condition, replicate)
    return(interactionSet)
}

#' @description
#' Parses interactions in \code{.cool} or \code{.mcool} format and fills the
#' interactions slot of the provided \code{\link{HiCDOCDataSet}}.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#' @param binSize
#' The resolution (span of each position in number of bases). Optionally
#' provided to select the appropriate resolution in \code{.mcool} files.
#' Defaults to NULL.
#'
#' @return
#' A filled \code{\link{HiCDOCDataSet}}.
#'
#' @keywords internal
#' @noRd
.parseCool <- function(object, binSize = NA, replicates, conditions) {

    interactionSetCool <- pbapply::pbmapply(
        .parseOneCool,
        path = object@input,
        binSize = binSize,
        condition = conditions,
        replicate = replicates
    )

    mergedinteractionSetCool <- Reduce(
        f = .mergeInteractionSet,
        x = interactionSetCool
    )

    new(
        "HiCDOCDataSet",
        mergedinteractionSetCool,
        input = object@input
    )
}

#' @description
#' Parses a single interactions file in \code{.hic} format. Calls the C++
#' \code{parseHiCFile} parser.
#'
#' @param path
#' The path to the interactions file.
#' @param binSize
#' The resolution (span of each position in number of bases) to select within
#' the \code{.hic} file.
#'
#' @return
#' An \code{\link{InteractionSet}}.
#'
#' @keywords internal
#' @noRd
.parseOneHiC <- function(path, binSize, condition, replicate) {
    message("\nParsing '", path, "'.")
    interactions <- parseHiCFile(path, binSize)
    # Automagical stuff to transform Rcpp DataFrame to data.table
    interactions <- data.table::setalloccol(interactions)
    # Create InteractionSet object
    interactions <- .setFromTabular(interactions, condition, replicate)
    return(interactions)
}

#' @description
#' Parses interactions in \code{.hic} format and fills the interactions slots of
#' the provided \code{\link{HiCDOCDataSet}}.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#' @param binSize
#' The resolution (span of each position in number of bases) to select within
#' the \code{.hic} files.
#'
#' @return
#' A filled \code{\link{HiCDOCDataSet}}.
#'
#' @keywords internal
#' @noRd
.parseHiC <- function(object, binSize, replicates, conditions) {

    interactionSet <- pbapply::pbmapply(
        .parseOneHiC,
        path      = object@input,
        binSize   = binSize,
        condition = conditions,
        replicate = replicates
    )

    mergedinteractionSet <- Reduce(
        f = .mergeInteractionSet,
        x = interactionSet
    )

    new(
        "HiCDOCDataSet",
        mergedinteractionSet,
        input = object@input
    )
}

#' @description
#' Parses a single pair of \code{.matrix} and \code{.bed} files.
#'
#' @param matrixPath
#' The path to the interactions matrix file.
#' @param bedPath
#' The path to the bed file.
#'
#' @return
#' A data.table of interactions.
#'
#' @keywords internal
#' @noRd
.parseOneHiCPro <- function(matrixPath, bedPath, replicate, condition) {

    message("\nParsing '", matrixPath, "' and '", bedPath, "'.")

    interactions <- data.table::fread(
        matrixPath,
        header = FALSE,
        stringsAsFactors = FALSE,
        col.names = c("startIndex", "stopIndex", "interaction"),
        data.table = TRUE
    )

    bed <- data.table::fread(
        bedPath,
        header = FALSE,
        stringsAsFactors = FALSE,
        col.names = c("chromosome", "start", "end", "index"),
        data.table = TRUE
    )

    setorder(bed, chromosome, start, end)
    # Adding 1 to follow Bioconductor GRanges recommended format
    bed[,start := start+1]
    # Keeping only intra-chromosomal interactions
    # Add 1 if BED index start with 0
    allChromosomes <- vector("character", length = max(bed$index) + 1)
    allChromosomes[bed[,index]+1] <- bed[,chromosome]
    interactions <- interactions[
        allChromosomes[startIndex + 1] == allChromosomes[stopIndex + 1]]
    
    order1 <- match(interactions$startIndex, bed$index)
    order2 <- match(interactions$stopIndex, bed$index)
    allRegions <- GenomicRanges::GRanges(bed)

    gi <- InteractionSet::GInteractions(
        allRegions[order1],
        allRegions[order2],
        regions = allRegions,
        mode="strict"
    )
    assay <- as.matrix(interactions$interaction, ncol=1)
    interactionSet <- .createInteractionSet(assay, gi, allRegions, condition, replicate)
    
    return(interactionSet)
}

#' @description
#' Parses interactions in pairs of \code{.matrix} and \code{.bed} files and
#' fills the interactions slots of the provided \code{\link{HiCDOCDataSet}}.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#'
#' @return
#' A filled \code{\link{HiCDOCDataSet}}.
#'
#' @keywords internal
#' @noRd
.parseHiCPro <- function(object, replicates, conditions) {

    matrixPaths <- lapply(object@input, `[[`, 1)
    bedPaths <- lapply(object@input, `[[`, 2)

    interactionSet <- pbapply::pbmapply(
        .parseOneHiCPro,
        matrixPaths,
        bedPaths,
        replicates,
        conditions
    )

    mergedinteractionSet <- Reduce(f = .mergeInteractionSet, x = interactionSet)

    object <- new(
        "HiCDOCDataSet",
        mergedinteractionSet,
        input = object@input
    )

    return(object)
}


#' Create the object interactionSet to use in HiCDOCDataSet
#'
#' @param assay matrix of assays
#' @param gi GInteractions object
#' @param allRegions regions object
#' @param condition condition (for colData), vector of length 1 or more
#' @param replicate replicate (for colData), vector of length 1 or more
#'
#' @return an interactionSet object
#' @keywords internal
#' @noRd
.createInteractionSet <- function(assay, gi, allRegions, condition, replicate){
    # Add 0 on the diagonal if there is only off diagonal interaction 
    ids <- InteractionSet::anchors(gi, id = TRUE)
    idsDiagonals <- ids$first[ids$first == ids$second]
    notPresent <- setdiff(unique(c(ids$first, ids$second)), idsDiagonals)
    nb <- length(notPresent)
    colnames(assay) <- NULL
    if(nb>0){
        notPresentRegions <- allRegions[match(notPresent, allRegions$index)]
        gi <- c(gi,
                InteractionSet::GInteractions(
                    notPresentRegions,
                    notPresentRegions,
                    regions = allRegions,
                    mode="strict"))
        assay <- rbind(assay, as.matrix(rep(0, nb*ncol(assay)),
                                        ncol=ncol(assay),
                                        nrow=nb))
    }
    
    interactionSet <- InteractionSet::InteractionSet(
        assays = assay,
        interactions = gi,
        colData=S4Vectors::DataFrame(
            "condition" = condition,
            "replicate" = replicate
        )
        
    )
    
    # Keep only intra-chromosomal interactions
    interactionSet <- interactionSet[InteractionSet::intrachr(interactionSet),]
    return(interactionSet)
}

#' @title
#' Check centroid PCA
#'
#' @description
#' Check whether centroids are correctly placed on a PCA.
#'
#' @param chromosomeName
#' A chromosome name or index in \code{chromosomes(object)}.
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#'
#' @return
#' A \code{data.table}, with 3 elements: the name of the chromosome,
#' whether the centroids of each compartment have the same sign 
#' only works when the number of conditions is 2), and whether the
#' variance explained by the first axis is greater than a threshold
#' (here, 75%).
#'
#' @examples
#' data(exampleHiCDOCDataSetProcessed)
#' .checkPca("X", exampleHiCDOCDataSetProcessed)
#'
#' @noRd
.checkPca <- function(chromosomeName, object) {
    compartments <- as.character(unique(object@centroids$compartment))
    pcaData <- .computePca(object, chromosomeName)
    pca     <- pcaData$PCA
    propvar <- pcaData$propvar
    if(is.null(pca)){
        pc1 <- FALSE
        centroid <- FALSE
    } else {
        f <- function (c) {
            return(length(unique(sign(pca[compartment == c, ]$PC1))) == 1)
        }
        centroid <- all(vapply(compartments, f, FUN.VALUE = TRUE))
        pc1 <- (propvar[[1]] >= object@parameters$PC1CheckThreshold)
    }
    return(data.table(chromosome     = chromosomeName,
                      centroid.check = centroid,
                      PC1.check      = pc1))
}


#' @title
#' Check compartment assignment
#'
#' @description
#' Check "A" compartments are different from "B" compartments.
#'
#' @param chromosome
#' A chromosome name or index in \code{chromosomes(object)}.
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#'
#' @return
#' A \code{data.table}, with 2 elements: the name of the chromosome,
#' whether the check passed (a Wilcoxon test, with p-value less than 5%).
#'
#' @noRd
.checkCompartmentAssignment <- function(chromosomeName, object) {
    compartments <- object@compartments[chromosome == chromosomeName, ]
    selfInteractionRatios <-
        object@selfInteractionRatios[chromosome == chromosomeName, ]
    compartments <- data.table::merge.data.table(compartments,
                                                 selfInteractionRatios,
                                                 by = c("index", "condition"))
    t            <- wilcox.test(ratio ~ compartment, data = compartments)
    return(data.table(chromosome       = chromosomeName,
                      assignment.check = (t$p.value <= 0.05)))
}


#' @title
#' Check whether compartments satisfy sanity checks.
#'
#' @description
#' Check centroid PCA and compartment assignments.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#'
#' @return
#' A \code{data.table}, with 3 elements: the name of the chromosome,
#' whether the centroids of each compartment have the same sign 
#' only works when the number of conditions is 2), and whether the
#' variance explained by the first axis is greater than a threshold
#' (here, 75%), and whether the check passed (a Wilcoxon test, with p-value
#' less than 5%).
#'
#' @noRd
.checkResults <- function(object) {
    pcaChecks         <- lapply(chromosomes(object), .checkPca, object)
    pcaChecks         <- data.table::rbindlist(pcaChecks)
    compartmentChecks <- lapply(chromosomes(object),
                                .checkCompartmentAssignment, object)
    compartmentChecks <- data.table::rbindlist(compartmentChecks)
    checks            <- pcaChecks
    checks[, assignment.check := compartmentChecks$assignment.check]
    object@checks <- checks
    return(object)
}

#' @title
#' Normalize technical biases.
#'
#' @description
#' Normalizes technical biases such as sequencing depth by using a cyclic loess
#' to recursively normalize each pair of interaction matrices. Depends on
#' \code{multiHiCcompare}.
#'
#' @details
#' \subsection{Parallel processing}{
#' If \code{parallel = TRUE}, the function
#' \code{\link[multiHiCcompare]{cyclic_loess}}
#' is launched in parallel mode, using \code{\link[BiocParallel]{bplapply}}
#' function. Before to call the function in parallel you should specify
#' the parallel parameters such as:
#'     \itemize{
#'         \item{On Linux:
#'
#'              \code{multiParam <- BiocParallel::MulticoreParam(workers = 10)}
#'          }
#'          \item{On Windows:
#'
#'              \code{multiParam <- BiocParallel::SnowParam(workers = 10)}
#'         }
#'     }
#'     And then you can register the parameters to be used by BiocParallel:
#'
#'     \code{BiocParallel::register(multiParam, default = TRUE)}
#' }
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#' @param parallel 
#' Logical. Whether or not to parallelize the processing. Defaults to FALSE
#' @param cyclicLoessSpan 
#' A numeric value in between 0 and 1. The span for cyclic loess normalization. 
#' This value is passed to \code{multiHiCcompare::cyclic_loess}. 
#' Defaults to NULL, NULL indicates that the value of 
#' parameters(object)$cyclicLoessSpan will be used. 
#' If this value is NA, the span will be automatically calculated using 
#' generalized cross validation. **For large dataset, it is highly recommended 
#' to set this value to reduce computing time and necessary memory.**
#'
#' @return
#' A \code{\link{HiCDOCDataSet}} with normalized interactions.
#'
#' @examples
#' data(exampleHiCDOCDataSet)
#' object <- filterSmallChromosomes(exampleHiCDOCDataSet)
#' object <- filterSparseReplicates(object)
#' object <- filterWeakPositions(object)
#' # Not printing loess warnings for example purpose. 
#' # Results should be inspected if there is any.
#' suppressWarnings(
#'     object <- normalizeTechnicalBiases(object)
#' )
#' 
#'
#' @seealso
#' \code{\link{filterSparseReplicates}},
#' \code{\link{filterWeakPositions}},
#' \code{\link{normalizeBiologicalBiases}},
#' \code{\link{normalizeDistanceEffect}},
#' \code{\link{HiCDOC}}
#'
#' @export
normalizeTechnicalBiases <- 
    function(object, parallel = FALSE, cyclicLoessSpan = NULL) {
    message("Normalizing technical biases.")

    if (!is.null(cyclicLoessSpan)) {
        object@parameters$cyclicLoessSpan <- cyclicLoessSpan
    }
        
    hic_table <- as.data.table(InteractionSet::interactions(object))
    hic_table <- hic_table[, .(
        chromosome = seqnames1,
        region1 = start1,
        region2 = start2
    )]
    if (!is.factor(hic_table$chromosome)) {
        hic_table[, chromosome := as.factor(chromosome)]
    }
    hic_table[, chromosome := as.numeric(chromosome)]

    currentAssay <- SummarizedExperiment::assay(object)
    currentAssay[is.na(currentAssay)] <- 0
    # Reordering columns in condition order
    refOrder <- paste(object$condition, object$replicate, sep = ".")
    currentAssay <- currentAssay[, order(refOrder), drop=FALSE]
    
    table_list <- lapply(
        seq_len(ncol(currentAssay)),
        function(x) cbind(hic_table, currentAssay[, x])
    )
    
    experiment <- multiHiCcompare::make_hicexp(
        data_list = table_list,
        groups = sort(object$condition),
        remove_zeros = FALSE,
        filter = TRUE,
        zero.p = 1,
        A.min = 0,
        remove.regions = NULL
    )
    
    normalized <- 
        multiHiCcompare::cyclic_loess(
            experiment, 
            parallel = parallel,
            span = object@parameters$cyclicLoessSpan
        )
    normalized <- multiHiCcompare::hic_table(normalized)
    data.table::setnames(normalized, "chr", "chromosome")

    # Re-sorting the rows in the same order as original
    data.table::setindexv(normalized, c("chromosome", "region1", "region2"))
    data.table::setindexv(hic_table, c("chromosome", "region1", "region2"))
    hic_table <- data.table::merge.data.table(
        hic_table,
        normalized,
        sort = FALSE
    )

    currentAssay <- as.matrix(hic_table[, 5:ncol(hic_table)])
    # Reordering columns in original order
    currentAssay <- currentAssay[, match(refOrder, sort(refOrder))]
    colnames(currentAssay) <- NULL
    currentAssay[currentAssay == 0] <- NA
    SummarizedExperiment::assay(object, withDimnames = FALSE) <- currentAssay
    return(object)
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

constrainedClustering <- function(rMatrix, rLinks, maxDelta = 0.0001, maxIterations = 50L, totalRestarts = 20L, totalClusters = 2L) {
    .Call('_HiCDOC_constrainedClustering', PACKAGE = 'HiCDOC', rMatrix, rLinks, maxDelta, maxIterations, totalRestarts, totalClusters)
}

#' @title
#' Example HiCDOCDataSet.
#'
#' @description
#' A S4 HiCDOCDataSet object with 4 chromosomes, 3 conditions and 3 replicates.
#'
#' @format
#' S4 HiCDOCDataSet object with the following characteristics:
#' \describe{
#'   \item{chromosomes}{4 chromosomes: W, X, Y, Z}
#'   \item{conditions}{3 conditions: 1, 2, 3}
#'   \item{replicates}{3 replicates: R1, R2, R3}
#'   \item{binSize}{A resolution of 137 bases}
#' }
#'
#' @usage data(exampleHiCDOCDataSet)
#'
#' @return
#' A \code{\link{HiCDOCDataSet}}.
#'
#' @examples
#' data(exampleHiCDOCDataSet)
#' exampleHiCDOCDataSet
"exampleHiCDOCDataSet"


#' @title
#' Example HiCDOCDataSet, filtered, normalized and with compartements detected.
#'
#' @description
#' A S4 HiCDOCDataSet object with 3 chromosomes, 3 conditions and 3 replicates.
#' Can be retrieved by running :
#' \code{data(exampleHiCDOCDataSet); 
#' set.seed(123); 
#' exampleHiCDOCDataSetProcessed <- HiCDOC(exampleHiCDOCDataSet)}
#'
#' @format
#' S4 HiCDOCDataSet object with the following characteristics:
#' \describe{
#'   \item{chromosomes}{4 chromosomes: X, Y, Z}
#'   \item{conditions}{3 conditions: 1, 2, 3}
#'   \item{replicates}{3 replicates: R1, R2, R3}
#'   \item{binSize}{A resolution of 137 bases}
#' }
#'
#' @usage data(exampleHiCDOCDataSetProcessed)
#'
#' @return
#' A \code{\link{HiCDOCDataSet}}, already filtered and normalized.
#'
#' @examples
#' data(exampleHiCDOCDataSetProcessed)
#' exampleHiCDOCDataSetProcessed
"exampleHiCDOCDataSetProcessed"

#' @description
#' Computes the euclidean distance between two vectors.
#'
#' @param x
#' A vector.
#' @param y
#' A vector.
#'
#' @return
#' A float number.
#'
#' @keywords internal
#' @noRd
.euclideanDistance <- function(x, y) {
    sqrt(sum((x - y) ^ 2))
}

#' @description
#' Computes the log ratio of the distance of a position to each centroid.
#'
#' @param x
#' The vector of a genomic position.
#' @param centroids
#' A list of two vectors.
#' @param eps
#' A small float number to avoid log(0).
#'
#' @return
#' A float number.
#'
#' @keywords internal
#' @noRd
.distanceRatio <- function(x, centroids) {
    epsilon = .euclideanDistance(centroids[[1]], centroids[[2]]) * 1e-10
    return(
        log(
            (
                .euclideanDistance(x, centroids[[1]]) + epsilon
            ) / (
                .euclideanDistance(x, centroids[[2]]) + epsilon
            )
        )
    )
}

#' @description
#' Assigns correct cluster labels by comparing centroids across conditions.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#'
#' @return
#' A \code{\link{HiCDOCDataSet}} with corrected cluster labels in compartments,
#' concordances, distances and centroids.
#'
#' @keywords internal
#' @noRd
.tieCentroids <- function(object) {

    validChromosomeNames <- names(
        base::Filter(
            function(x) !is.null(x),
            object@validAssay
        )
    )

    referenceConditionNames <- vapply(
        validChromosomeNames,
        FUN = function(x) sort(object$condition[object@validAssay[[x]]])[1],
        FUN.VALUE = ""
    )

    referenceCentroids <- data.table::merge.data.table(
        object@centroids[
            cluster == 1 &
            condition == referenceConditionNames[chromosome],
            .(chromosome, reference.1 = centroid)
        ],
        object@centroids[
            cluster == 2 &
            condition == referenceConditionNames[chromosome],
            .(chromosome, reference.2 = centroid)
        ],
        all = TRUE
    )

    clusters <- data.table::merge.data.table(
        object@centroids[
            cluster == 1,
            .(chromosome, condition, centroid.1 = centroid)
        ],
        object@centroids[
            cluster == 2,
            .(chromosome, condition, centroid.2 = centroid)
        ],
        all = TRUE
    )

    clusters <- data.table::merge.data.table(
        clusters,
        referenceCentroids,
        all = TRUE
    )

    c1_r1 <- mapply(
        function(x, y) .euclideanDistance(unlist(x), unlist(y)),
        clusters$centroid.1,
        clusters$reference.1
    )
    c1_r2 <- mapply(
        function(x, y) .euclideanDistance(unlist(x), unlist(y)),
        clusters$centroid.1,
        clusters$reference.2
    )
    c2_r1 <- mapply(
        function(x, y) .euclideanDistance(unlist(x), unlist(y)),
        clusters$centroid.2,
        clusters$reference.1
    )
    c2_r2 <- mapply(
        function(x, y) .euclideanDistance(unlist(x), unlist(y)),
        clusters$centroid.2,
        clusters$reference.2
    )

    clusters[, cluster.1 := 1 * ((c1_r1 * c2_r2) >= (c1_r2 * c2_r1)) + 1]
    clusters[, cluster.2 := 1 + (cluster.1 == 1)]

    clusters <- clusters[, .(chromosome, condition, cluster.1, cluster.2)]

    object@compartments <- data.table::merge.data.table(
        object@compartments,
        clusters,
        by = c("chromosome", "condition"),
        all.x = TRUE,
        sort = FALSE
    )

    object@compartments[
        ,
        cluster := ifelse(
            cluster == 1,
            cluster.1,
            cluster.2
        )
    ]

    object@compartments[, `:=`(cluster.1 = NULL, cluster.2 = NULL)]

    object@concordances <- data.table::merge.data.table(
        object@concordances,
        clusters,
        by = c("chromosome", "condition"),
        all.x = TRUE,
        sort = FALSE
    )

    object@concordances[, change := -1]
    object@concordances[
        cluster == 1 & cluster == cluster.1,
        change := 1
    ]
    object@concordances[
        cluster == 2 & cluster == cluster.2,
        change := 1
    ]
    object@concordances[, concordance := change * concordance]

    object@concordances[, compartment := data.table::fifelse(
        cluster == 1,
        cluster.1,
        cluster.2
    )]

    object@concordances[, `:=`(
        cluster.1 = NULL,
        cluster.2 = NULL,
        change = NULL
    )]

    object@distances <- data.table::merge.data.table(
        object@distances,
        clusters,
        by = c("chromosome", "condition"),
        all.x = TRUE,
        sort = FALSE
    )
    object@distances[, cluster := data.table::fifelse(
        cluster == 1,
        cluster.1,
        cluster.2
    )]
    object@distances[, `:=`(cluster.1 = NULL, cluster.2 = NULL)]

    object@centroids <- data.table::merge.data.table(
        object@centroids,
        clusters,
        by = c("chromosome", "condition"),
        all.x = TRUE,
        sort = FALSE
    )

    object@centroids[, cluster := data.table::fifelse(
        cluster == 1,
        cluster.1,
        cluster.2
    )]
    object@centroids[, `:=`(cluster.1 = NULL, cluster.2 = NULL)]

    return(object)
}

#' @description
#' Constructs a link matrix of interaction rows to be clustered together.
#'
#' @param totalReplicates
#' The number of replicates.
#' @param totalBins
#' The number of bins.
#'
#' @return
#' A matrix, each row holding the row indices of interactions to be clustered
#' together.
#'
#' @keywords internal
#' @noRd
.constructLinkMatrix <- function(totalReplicates, totalBins) {
    return(
        matrix(
            rep(0:(totalReplicates - 1), totalBins) * totalBins +
            rep(0:(totalBins - 1), each = totalReplicates),
            nrow = totalBins,
            byrow = TRUE
        )
    )
}

#' @description
#' Segregates positions of a chromosome into two clusters using constrained
#' K-means.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#' @param chromosomeName
#' The name of a chromosome.
#' @param conditionName
#' The name of a condition.
#'
#' @return
#' A list of:
#' - The compartment (cluster number) of each position.
#' - The concordance (float) of each genomic position in each replicate.
#' - The distances to centroids (float) of each position in each replicate.
#' - The centroid (vector) of each cluster.
#'
#' @md
#' @keywords internal
#' @noRd
.clusterizeChromosomeCondition <- function(object) {
    chromosomeName <- object@chromosomes
    conditionName <- object$condition[1]

    totalBins <- length(InteractionSet::regions(object))
    validAssay <- object@validAssay[[chromosomeName]]
    if (length(validAssay) == 0) {
        return(NULL)
    }
    replicateNames <- object$replicate[validAssay]
    orderedAssay <- validAssay[order(replicateNames)]

    chromosomeInteractionSet <- InteractionSet::InteractionSet(
        SummarizedExperiment::assay(object),
        InteractionSet::interactions(object)
    )
    matrixAssay <- lapply(
        orderedAssay,
        FUN = function(x) {
            InteractionSet::inflate(
                chromosomeInteractionSet,
                rows = chromosomeName,
                columns = chromosomeName,
                sample = x
            )
        }
    )
    matrixAssay <- lapply(
        matrixAssay,
        function(x) x@matrix
    )
    matrixAssay <- do.call("rbind", matrixAssay)
    matrixAssay[is.na(matrixAssay)] <- 0

    mustLink <- .constructLinkMatrix(length(replicateNames), totalBins)
    clusteringOutput <- constrainedClustering(
        matrixAssay,
        mustLink,
        object@parameters$kMeansDelta,
        object@parameters$kMeansIterations,
        object@parameters$kMeansRestarts
    )
    # TODO : question : pourquoi on ne prend que les premiers ?
    # Quel est l'intérêt de retourner 2 fois ?
    clusters <- as.integer(clusteringOutput[["clusters"]][0:totalBins] + 1)
    centroids <- clusteringOutput[["centroids"]]

    min <- .distanceRatio(centroids[[1]], centroids)
    max <- .distanceRatio(centroids[[2]], centroids)

    concordances <- apply(
        matrixAssay,
        1,
        function(row) {
            2 * (.distanceRatio(row, centroids) - min) / (max - min) - 1
        }
    )

    distances <- apply(
        matrixAssay,
        1,
        function(row) {
            c(
                .euclideanDistance(row, centroids[[1]]),
                .euclideanDistance(row, centroids[[2]])
            )
        }
    )

    indices <- S4Vectors::mcols(InteractionSet::regions(object))$index

    dfCompartments <- data.table::data.table(
        "chromosome" = chromosomeName,
        "index" = indices,
        "condition" = conditionName,
        "cluster" = clusters
    )

    dfConcordances <- data.table::data.table(
        "chromosome" = chromosomeName,
        "index" = rep(indices, length(replicateNames)),
        "condition" = conditionName,
        "replicate" = rep(sort(replicateNames), each = totalBins),
        "cluster" = rep(clusters, length(replicateNames)),
        "concordance" = concordances
    )

    dfDistances <- data.table::data.table(
        "chromosome" = chromosomeName,
        "index" = rep(indices, 2 * length(replicateNames)),
        "condition" = conditionName,
        "replicate" = rep(rep(sort(replicateNames), each = totalBins), 2),
        "cluster" = rep(c(1, 2), each = length(replicateNames) * totalBins),
        "distance" = c(t(distances))
    )

    dfCentroids <- data.table::data.table(
        "chromosome" = chromosomeName,
        "condition" = conditionName,
        "cluster" = c(1, 2),
        "centroid" = centroids
    )

    return(
        list(
            "compartments" = dfCompartments,
            "concordances" = dfConcordances,
            "distances" = dfDistances,
            "centroids" = dfCentroids
        )
    )
}

#' @description
#' Runs the clustering to detect compartments in each chromosome and condition.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#' @param parallel
#' Whether or not to parallelize the processing. Defaults to FALSE.
#'
#' @return
#' A \code{\link{HiCDOCDataSet}} with compartments, concordances, distances
#' and centroids.
#'
#' @keywords internal
#' @noRd
.clusterize <- function(object, parallel = FALSE) {
    conditionsPerChromosome <- lapply(
        object@chromosomes,
        FUN = function(x) {
            data.frame(
                "chromosome" = x,
                "condition" = sort(unique(
                    object$condition[object@validAssay[[x]]]
                ))
            )
        }
    )
    conditionsPerChromosome <- Reduce(rbind, conditionsPerChromosome)

    reducedObjects <- mapply(
        function(x, y) {
            reduceHiCDOCDataSet(
                object,
                chromosomes = x,
                conditions = y,
                dropLevels = TRUE
            )
        },
        conditionsPerChromosome$chromosome,
        conditionsPerChromosome$condition
    )

    result <- .internalLapply(
        parallel,
        reducedObjects,
        .clusterizeChromosomeCondition
    )

    compartments <- lapply(
        result,
        function(x) x[["compartments"]]
    )
    concordances <- lapply(
        result,
        function(x) x[["concordances"]]
    )
    distances <- lapply(
        result,
        function(x) x[["distances"]]
    )
    centroids <- lapply(
        result,
        function(x) x[["centroids"]]
    )

    object@compartments <- data.table::rbindlist(compartments)
    object@concordances <- data.table::rbindlist(concordances)
    object@distances <- data.table::rbindlist(distances)
    object@centroids <- data.table::rbindlist(centroids)

    return(object)
}

#' @description
#' Computes the ratio of self interaction vs median of other interactions for
#' each position.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#' @param chromosomeName
#' The name of a chromosome.
#' @param conditionName
#' The name of a condition.
#' @param replicateName
#' The name of a replicate.
#'
#' @return
#' A data.table
#'
#' @keywords internal
#' @noRd
.computeSelfInteractionRatios <- function(object) {
    ids <- InteractionSet::anchors(object, id = FALSE)
    diagonal <- (ids$first == ids$second)
    ids <- lapply(ids, as.data.table)
    columnNames <- paste(object$condition, object$replicate)

    matrixAssay <- SummarizedExperiment::assay(object)
    colnames(matrixAssay) <- columnNames

    # Values on diagonal
    onDiagonal <- data.table::data.table(
        ids$first[diagonal,.(chromosome = seqnames, index)],
        matrixAssay[diagonal, , drop=FALSE]
    )
    onDiagonal <- data.table::melt.data.table(
        onDiagonal,
        id.vars = c("chromosome", "index"),
        value.name = "ratio",
        na.rm = TRUE
    )
    # Compute median by bin, out of diagonal
    offDiagonal <- data.table::data.table(
        "chromosome" = c(ids$first$seqnames[!diagonal], 
                         ids$second$seqnames[!diagonal]),
        "index" = c(ids$first$index[!diagonal], 
                    ids$second$index[!diagonal]),
        matrixAssay[!diagonal, , drop=FALSE]
    )
    offDiagonal <- data.table::melt.data.table(
        offDiagonal,
        id.vars = c("chromosome", "index"),
        value.name = "offDiagonal",
        variable.name = "variable",
        na.rm = TRUE
    )
    offDiagonal <- offDiagonal[!is.na(offDiagonal)]
    offDiagonal <- offDiagonal[
        ,
        .(offDiagonal = sum(offDiagonal, na.rm=TRUE)),
        by = c("chromosome", "index", "variable")
    ]
    # Ratio is value on diagonal - median (off diagonal), by bin
    onDiagonal <- data.table::merge.data.table(
        onDiagonal,
        offDiagonal,
        all = TRUE,
        by = c("chromosome", "index", "variable"),
        sort = FALSE
    )
    #Shoudn't happen after normalizations
    onDiagonal[is.na(ratio) & !is.na(offDiagonal),ratio := 0] 
    onDiagonal[, c("condition", "replicate") := data.table::tstrsplit(
        variable,
        " ",
        fixed = TRUE
    )]
    onDiagonal <- onDiagonal[, .(
        chromosome,
        index,
        condition,
        replicate,
        ratio,
        offDiagonal
    )]

    return(onDiagonal)
}

#' @description
#' Uses ratio between self interactions and other interactions to determine
#' which clusters correspond to compartments A and B.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#' @param parallel
#' Whether or not to parallelize the processing. Defaults to FALSE.
#'
#' @return
#' A \code{\link{HiCDOCDataSet}}, with selfInteractionRatios, and with A and B
#' labels replacing cluster numbers in compartments, concordances, distances,
#' and centroids.
#'
#' @keywords internal
#' @noRd
.predictCompartmentsAB <- function(object, parallel = FALSE) {
    ratios <- .computeSelfInteractionRatios(object)
    object@selfInteractionRatios <- ratios

    compartments <- data.table::merge.data.table(
        object@compartments,
        object@selfInteractionRatios,
        by = c("chromosome", "index", "condition"),
        all.x = TRUE,
        sort = FALSE
    )
    compartments[, offDiagonal := NULL]
    compartments[, ratio := as.numeric(ratio)]
    compartments <- compartments[
        ,
        .(ratio = stats::median(ratio, na.rm = TRUE)),
        by = .(chromosome, cluster)
    ]
    compartments <- data.table::dcast(
        compartments,
        chromosome ~ cluster,
        value.var = "ratio",
        fill = 0
    )

    compartments[, A := data.table::fifelse(`1` >= `2`, 1, 2)]
    compartments <- compartments[, .(chromosome, A)]

    object@compartments <- data.table::merge.data.table(
        object@compartments,
        compartments,
        by = "chromosome",
        all.x = TRUE
    )

    object@compartments[, compartment := factor(
        data.table::fifelse(cluster == A, "A", "B"), 
        levels=c("A", "B"))]
    object@compartments[, A := NULL]

    object@concordances <- data.table::merge.data.table(
        object@concordances,
        compartments,
        by = "chromosome",
        all.x = TRUE
    )
    object@concordances[, change := data.table::fifelse(A == 1, 1,-1)]
    object@concordances[, concordance :=  change * concordance]
    object@concordances[, compartment := factor(
        data.table::fifelse(cluster == A, "A", "B"),
        levels = c("A", "B")
    )]
    object@concordances[, change := NULL]
    object@concordances[, A := NULL]

    object@distances <- data.table::merge.data.table(
        object@distances,
        compartments,
        by = "chromosome",
        all.x = TRUE
    )
    object@distances[, compartment := factor(
        data.table::fifelse(cluster == A, "A", "B"),
        levels = c("A", "B")
    )]
    object@distances[, A := NULL]

    object@centroids <- data.table::merge.data.table(
        object@centroids,
        compartments,
        by = "chromosome",
        all.x = TRUE
    )
    object@centroids[, compartment := factor(
        data.table::fifelse(cluster == A, "A", "B"),
        levels = c("A", "B")
    )]
    object@centroids[, A := NULL]

    return(object)
}

#' @description
#' Computes p-values for genomic positions whose assigned compartment switches
#' between two conditions.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#'
#' @return
#' A \code{\link{HiCDOCDataSet}}, with differences and their p-values.
#'
#' @keywords internal
#' @noRd
.computePValues <- function(object) {
    # Compute median of differences between pairs of concordances
    # N.b. median of differences != difference of medians
    totalReplicates <- length(object$replicate)
    concordances <- object@concordances
    concordances[, condition := factor(
        condition,
        levels = sort(unique(object$condition))
    )]
    data.table::setorder(concordances, chromosome, index, condition, replicate)
    concordances1 <- concordances[
        rep(seq_len(nrow(concordances)), each = totalReplicates),
        .(
            chromosome,
            index,
            condition.1 = condition,
            concordance.1 = concordance
        )
    ]
    concordances2 <- concordances[
        rep(seq_len(nrow(concordances)), totalReplicates),
        .(
             chromosome = chromosome,
             index = index,
             condition.2 = condition,
             concordance.2 = concordance
        )
    ]
    data.table::setorder(concordances2, chromosome, index)
    concordances2[, `:=`(chromosome = NULL, index = NULL)]
    concordanceDifferences <- base::cbind(concordances1, concordances2)
    rm(concordances1, concordances2)
    concordanceDifferences <- concordanceDifferences[
        as.numeric(condition.1) < as.numeric(condition.2)
    ]
    concordanceDifferences <- concordanceDifferences[,
        .(difference = stats::median(abs(concordance.1 - concordance.2))),
        by = .(chromosome, index, condition.1, condition.2)
    ]

    # Format compartments per pair of conditions
    # Join medians of differences and pairs of conditions
    totalConditions <- length(unique(object$condition))
    compartments <- object@compartments
    compartments[, condition := factor(
        condition,
        levels = sort(unique(object$condition))
    )]
    data.table::setorder(compartments, chromosome, index, condition)
    compartments1 <- compartments[
        rep(seq_len(nrow(compartments)), each = totalConditions),
        .(
            chromosome,
            index,
            condition.1 = condition,
            compartment.1 = compartment
        )
    ]
    compartments2 <- compartments[
        rep(seq_len(nrow(compartments)), totalConditions),
        .(
            chromosome,
            index,
            condition.2 = condition,
            compartment.2 = compartment
        )
    ]
    data.table::setorder(compartments2, chromosome, index)
    compartments2[, `:=`(chromosome = NULL, index = NULL)]
    comparisons <- base::cbind(compartments1, compartments2)
    rm(compartments1, compartments2)
    comparisons <- comparisons[
        as.numeric(condition.1) < as.numeric(condition.2)
    ]
    if(nrow(comparisons) == 0){
        object@comparisons <- comparisons[,.(chromosome,
                                             index, 
                                             condition.1, 
                                             condition.2,
                                             compartment.1,
                                             compartment.2,
                                             difference = index)]
        object@differences <- comparisons[,.(chromosome, 
                                             index, 
                                             condition.1,
                                             condition.2,
                                             pvalue = index,
                                             pvalue.adjusted = index,
                                             direction = compartment.1)]
        return(object)
    }
    comparisons <- data.table::merge.data.table(
        comparisons,
        concordanceDifferences,
        by = c("chromosome", "index", "condition.1", "condition.2")
    )
    data.table::setcolorder(
        comparisons,
        c(
            "chromosome",
            "index",
            "condition.1",
            "condition.2",
            "compartment.1",
            "compartment.2",
            "difference"
        )
    )
    object@comparisons <- comparisons

    # Compute p-values for switching positions
    # P-values for a condition pair computed from the whole genome distribution
    differences <- copy(comparisons)
    differences[compartment.1 == compartment.2 , H0_value := difference]
    data.table::setorder(differences, condition.1, condition.2)
    quantiles <- split(
        differences,
        by = c("condition.1", "condition.2")
    )
    quantiles <- lapply(
        quantiles,
        function(x) x[difference > 0]
    )
    quantiles <- lapply(
        quantiles,
        function(x) {
            if (nrow(x) > 0) return(stats::ecdf(x$H0_value)(x$difference))
            return(NULL)
        }
    )
    quantiles <- do.call("c", quantiles)
    differences[difference > 0, quantile := quantiles]

    # Pvalues
    differences <- differences[compartment.1 != compartment.2]
    differences[, pvalue := 1 - quantile]
    differences[pvalue < 0, pvalue := 0]
    differences[pvalue > 1, pvalue := 1]
    pvalueAdjusted <- split(
        differences,
        by = c("condition.1", "condition.2")
    )
    pvalueAdjusted <- lapply(
        pvalueAdjusted,
        function(x) {
            if (nrow(x) > 0) return(stats::p.adjust(x$pvalue, method = "BH"))
            return(NULL)
        }
    )
    pvalueAdjusted <- do.call("c", pvalueAdjusted)
    differences[, pvalue.adjusted := pvalueAdjusted]

    # Changes
    differences[, direction := data.table::fifelse(
        compartment.1 == "A", "A->B", "B->A"
    )]
    differences[, direction := factor(direction, levels = c("A->B", "B->A"))]
    differences <- differences[, .(
        chromosome,
        index,
        condition.1,
        condition.2,
        pvalue,
        pvalue.adjusted,
        direction
    )]
    data.table::setorder(
        differences,
        chromosome,
        index,
        condition.1,
        condition.2
    )
    object@differences <- differences
    return(object)
}

#' @title
#' A and B compartments detection and differences across conditions.
#'
#' @description
#' Detects compartments for each genomic position in each condition, and
#' computes p-values for compartment differences between conditions.
#'
#' @details
#' \subsection{Genomic positions clustering}{
#' To clusterize genomic positions, the algorithm follows these steps:
#'     \enumerate{
#'         \item{
#'             For each chromosome and condition, get the interaction vectors of
#'             each genomic position. Each genomic position can have multiple
#'             interaction vectors, corresponding to the multiple replicates in
#'             that condition.
#'         }
#'         \item{
#'             For each chromosome and condition, use constrained K-means to
#'             clusterize the interaction vectors, forcing replicate interaction
#'             vectors into the same cluster. The euclidean distance between
#'             interaction vectors determines their similarity.
#'         }
#'         \item{
#'             For each interaction vector, compute its concordance, which is
#'             the confidence in its assigned cluster. Mathematically, it is the
#'             log ratio of its distance to each centroid, normalized by the
#'             distance between both centroids, and min-maxed to a [-1,1]
#'             interval.
#'         }
#'         \item{
#'             For each chromosome, compute the distance between all centroids
#'             and the centroids of the first condition. The cross-condition
#'             clusters whose centroids are closest are given the same cluster
#'             label. This results in two clusters per chromosome, spanning all
#'             conditions.
#'         }
#'     }
#' }
#' \subsection{A/B compartments prediction}{
#' To match each cluster with an A or B compartment, the algorithm follows these
#' steps:
#'     \enumerate{
#'         \item{
#'             For each genomic position, compute its self interaction ratio,
#'             which is the difference between its self interaction and the
#'             median of its other interactions.
#'         }
#'         \item{
#'             For each chromosome, for each cluster, get the median self
#'             interaction ratio of the genomic positions in that cluster.
#'         }
#'         \item{
#'             For each chromosome, the cluster with the smallest median self
#'             interaction ratio is matched with compartment A, and the cluster
#'             with the greatest median self interaction ratio is matched with
#'             compartment B. Compartment A being open, there are more overall
#'             interactions between distant genomic positions, so it is assumed
#'             that the difference between self interactions and other
#'             interactions is lower than in compartment B.
#'         }
#'     }
#' }
#' \subsection{Significant differences detection}{
#' To find significant compartment differences across conditions, and compute
#' their p-values, the algorithm follows three steps:
#'     \enumerate{
#'         \item{
#'             For each pair of replicates in different conditions, for each
#'             genomic position, compute the absolute difference between its
#'             concordances.
#'         }
#'         \item{
#'             For each pair of conditions, for each genomic position, compute
#'             the median of its concordance differences.
#'         }
#'         \item{
#'             For each pair of conditions, for each genomic position whose
#'             assigned compartment switches, rank its median against the
#'             empirical cumulative distribution of medians of all non-switching
#'             positions in that condition pair. Adjust the resulting p-value
#'             with the Benjamini–Hochberg procedure.
#'         }
#'     }
#' }
#' \subsection{Parallel processing}{
#' The parallel version of detectCompartments uses the
#' \code{\link[BiocParallel]{bpmapply}} function. Before to call the
#' function in parallel you should specify the parallel parameters such as:
#'     \itemize{
#'         \item{On Linux:
#'
#'              \code{multiParam <- BiocParallel::MulticoreParam(workers = 10)}
#'          }
#'          \item{On Windows:
#'
#'              \code{multiParam <- BiocParallel::SnowParam(workers = 10)}
#'         }
#'     }
#'     And then you can register the parameters to be used by BiocParallel:
#'
#'     \code{BiocParallel::register(multiParam, default = TRUE)}
#'
#'     You should be aware that using MulticoreParam, reproducibility of the
#'     detectCompartments function using a RNGseed may not work. See this
#'     \href{https://github.com/Bioconductor/BiocParallel/issues/122}{issue}
#'     for more details.
#' }
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#'
#' @param parallel
#' Whether or not to parallelize the processing. Defaults to FALSE
#' See 'Details'.
#'
#' @param kMeansDelta
#' The convergence stop criterion for the clustering. When the centroids'
#' distances between two iterations is lower than this value, the clustering
#' stops. Defaults to \code{object$kMeansDelta} which is originally set to
#' \code{defaultHiCDOCParameters$kMeansDelta} = 0.0001.
#'
#' @param kMeansIterations
#' The maximum number of iterations during clustering. Defaults to
#' \code{object$kMeansIterations} which is originally set to
#' \code{defaultHiCDOCParameters$kMeansIterations} = 50.
#'
#' @param kMeansRestarts
#' The amount of times the clustering is restarted. For each restart, the
#' clustering iterates until convergence or reaching the maximum number of
#' iterations. The clustering that minimizes inner-cluster variance is selected.
#' Defaults to \code{object$kMeansRestarts} which is originally set to
#' \code{defaultHiCDOCParameters$kMeansRestarts} = 20.
#'
#' @param PC1CheckThreshold
#' The minimum percentage of variance that should be explained by
#' the first principal component of centroids to pass sanity check.
#' Defaults to \code{object$PC1CheckThreshold} which is originally set to
#' \code{defaultHiCDOCParameters$PC1CheckThreshold} = 0.75
#'
#' @return
#' A \code{\link{HiCDOCDataSet}}, with compartments, concordances, distances,
#' centroids, and differences.
#'
#' @examples
#' data(exampleHiCDOCDataSet)

#' ## Run all filtering and normalization steps (not run for timing reasons)
#' # object <- filterSmallChromosomes(exampleHiCDOCDataSet)
#' # object <- filterSparseReplicates(object)
#' # object <- filterWeakPositions(object)
#' # object <- normalizeTechnicalBiases(object)
#' # object <- normalizeBiologicalBiases(object)
#' # object <- normalizeDistanceEffect(object)
#'
#' # Detect compartments and differences across conditions
#' object <- detectCompartments(exampleHiCDOCDataSet)
#'
#' @usage
#' detectCompartments(
#'     object,
#'     parallel = FALSE,
#'     kMeansDelta = NULL,
#'     kMeansIterations = NULL,
#'     kMeansRestarts = NULL,
#'     PC1CheckThreshold = NULL
#' )
#'
#' @export
detectCompartments <- function(
    object,
    parallel = FALSE,
    kMeansDelta = NULL,
    kMeansIterations = NULL,
    kMeansRestarts = NULL,
    PC1CheckThreshold = NULL
) {

    .validateSlots(
        object,
        slots = c(
            "chromosomes",
            "validAssay",
            "parameters"
        )
    )

    if (!is.null(kMeansDelta)) {
        object@parameters$kMeansDelta <- kMeansDelta
    }
    if (!is.null(kMeansIterations)) {
        object@parameters$kMeansIterations <- kMeansIterations
    }
    if (!is.null(kMeansRestarts)) {
        object@parameters$kMeansRestarts <- kMeansRestarts
    }
    if (!is.null(PC1CheckThreshold)) {
        object@parameters$PC1CheckThreshold <- PC1CheckThreshold
    }
    object@parameters <- .validateParameters(object@parameters)

    message("Clustering genomic positions.")
    object <- .clusterize(object, parallel)
    object <- .tieCentroids(object)

    message("Predicting A/B compartments.")
    object <- .predictCompartmentsAB(object, parallel)
    object <- .checkResults(object)
    message("Detecting significant differences.")
    object <- .computePValues(object)

    # Reformating outputs
    object <- .formatDetectCompartment(object)

    return(object)
}

data(exampleHiCDOCDataSet)
object <- reduceHiCDOCDataSet(exampleHiCDOCDataSet, chromosomes=c("X", "Z"), conditions = c(1, 2))

test_that("detectCompartments behaves as expected", {
    # Detect Compartments
    set.seed(3215) # Test with 123 : no significant differences
    expect_message(
        object <- detectCompartments(object, parallel = FALSE),
        "Detecting significant differences."
    )
    # Create new objects in correct format
    expect_is(object@distances, "data.table")
    expect_is(object@selfInteractionRatios, "data.table")
    expect_is(object@compartments, "GRanges")
    expect_is(object@concordances, "GRanges")
    expect_is(object@differences, "GRanges")
    expect_is(object@centroids, "data.table")

    # Differences
    expect_equal(length(object@differences), 1)
    expect_equal(
        length(object@differences[object@differences$pvalue.adjusted <= 0.05]),
        0
    )
    expect_is(object@differences@seqnames, "Rle")
    expect_is(object@differences$condition.1, "factor")
    expect_is(object@differences$condition.2, "factor")

    # Centroids
    expect_equal(nrow(object@centroids), 6)
    expect_equal(
        sapply(object@centroids$centroid, length),
        c(rep(120, 4), rep(200, 2))
    )
    expect_equal(
        sapply(object@centroids$centroid, mean),
        c(572.776, 631.262, 514.537, 571.618, 676.45, 803.454),
        tolerance = 1e-04
    )
    expect_is(object@centroids$chromosome, "factor")
    expect_is(object@centroids$condition, "factor")
    expect_is(object@centroids$compartment, "factor")
    expect_is(object@centroids$centroid, "list")

    # Compartments
    expect_equal(length(object@compartments), 440)
    expect_equal(
        length(object@compartments[object@compartments$compartment == "B"]),
        169
    )
    expect_is(object@compartments@seqnames, "Rle")
    expect_is(object@compartments$condition, "factor")
    expect_is(object@compartments$compartment, "factor")
    expect_is(object@compartments$index, "numeric")

    # Concordance
    expect_is(object@concordances@seqnames, "Rle")
    expect_is(object@concordances$index, "numeric")
    expect_is(object@concordances$condition, "factor")
    expect_is(object@concordances$replicate, "factor")
    expect_is(object@concordances$compartment, "factor")
    expect_is(object@concordances$concordance, "numeric")
    expect_equal(length(object@concordances), 880)
    expect_equal(
      length(object@concordances[object@concordances$compartment == "A"]),
      543
    )
    expect_equal(
      100 * mean(object@concordances$concordance),
      0.03121712,
      tolerance = 1e-05
    )

    # Distances
    expect_is(object@distances$chromosome, "factor")
    expect_is(object@distances$index, "numeric")
    expect_is(object@distances$condition, "factor")
    expect_is(object@distances$replicate, "factor")
    expect_is(object@distances$compartment, "factor")
    expect_is(object@distances$distance, "numeric")
    expect_equal(nrow(object@distances), 1760)
    expect_equal(mean(object@distances$distance), 5713.632, tolerance = 1e-04)

    # SelfInteractionRatios
    expect_equal(nrow(object@selfInteractionRatios), 879)
    expect_equal(
      mean(object@selfInteractionRatios$ratio),
      1087.815,
      tolerance = 1e-04
    )
    expect_is(object@selfInteractionRatios$chromosome, "factor")
    expect_is(object@selfInteractionRatios$index, "numeric")
    expect_is(object@selfInteractionRatios$condition, "factor")
    expect_is(object@selfInteractionRatios$replicate, "factor")
    expect_true(is.numeric(object@selfInteractionRatios$ratio))
})

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>"
)
options(warn=-1)

## ----eval=FALSE---------------------------------------------------------------
# if (!requireNamespace("BiocManager", quietly=TRUE))
#     install.packages("BiocManager")
# BiocManager::install("HiCDOC")

## -----------------------------------------------------------------------------
library(HiCDOC)

## ----tabFormat, eval = FALSE--------------------------------------------------
# hic.experiment <- HiCDOCDataSetFromTabular('path/to/data.tsv')

## ----coolFormat, eval = FALSE-------------------------------------------------
# # Path to each file
# paths = c(
#     'path/to/condition-1.replicate-1.cool',
#     'path/to/condition-1.replicate-2.cool',
#     'path/to/condition-2.replicate-1.cool',
#     'path/to/condition-2.replicate-2.cool',
#     'path/to/condition-3.replicate-1.cool'
# )
# 
# # Replicate and condition of each file. Can be names instead of numbers.
# replicates <- c(1, 2, 1, 2, 1)
# conditions <- c(1, 1, 2, 2, 3)
# 
# # Resolution to select in .mcool files
# binSize = 500000
# 
# # Instantiation of data set
# hic.experiment <- HiCDOCDataSetFromCool(
#     paths,
#     replicates = replicates,
#     conditions = conditions,
#     binSize = binSize # Specified for .mcool files.
# )

## ----hicFormat, eval = FALSE--------------------------------------------------
# # Path to each file
# paths = c(
#     'path/to/condition-1.replicate-1.hic',
#     'path/to/condition-1.replicate-2.hic',
#     'path/to/condition-2.replicate-1.hic',
#     'path/to/condition-2.replicate-2.hic',
#     'path/to/condition-3.replicate-1.hic'
# )
# 
# # Replicate and condition of each file. Can be names instead of numbers.
# replicates <- c(1, 2, 1, 2, 1)
# conditions <- c(1, 1, 2, 2, 3)
# 
# # Resolution to select
# binSize <- 500000
# 
# # Instantiation of data set
# hic.experiment <- HiCDOCDataSetFromHiC(
#     paths,
#     replicates = replicates,
#     conditions = conditions,
#     binSize = binSize
# )

## ----hicproFormat, eval = FALSE-----------------------------------------------
# # Path to each matrix file
# matrixPaths = c(
#     'path/to/condition-1.replicate-1.matrix',
#     'path/to/condition-1.replicate-2.matrix',
#     'path/to/condition-2.replicate-1.matrix',
#     'path/to/condition-2.replicate-2.matrix',
#     'path/to/condition-3.replicate-1.matrix'
# )
# 
# # Path to each bed file
# bedPaths = c(
#     'path/to/condition-1.replicate-1.bed',
#     'path/to/condition-1.replicate-2.bed',
#     'path/to/condition-2.replicate-1.bed',
#     'path/to/condition-2.replicate-2.bed',
#     'path/to/condition-3.replicate-1.bed'
# )
# 
# # Replicate and condition of each file. Can be names instead of numbers.
# replicates <- c(1, 2, 1, 2, 1)
# conditions <- c(1, 1, 2, 2, 3)
# 
# # Instantiation of data set
# hic.experiment <- HiCDOCDataSetFromHiCPro(
#     matrixPaths = matrixPaths,
#     bedPaths = bedPaths,
#     replicates = replicates,
#     conditions = conditions
# )

## ----reloadExample------------------------------------------------------------
data(exampleHiCDOCDataSet)

## ----filterSmallChromosomes---------------------------------------------------
hic.experiment <- filterSmallChromosomes(exampleHiCDOCDataSet, threshold = 100)

## ----filterSparseReplicates---------------------------------------------------
hic.experiment <- filterSparseReplicates(hic.experiment, threshold = 0.3)

## ----filterWeakPositions------------------------------------------------------
hic.experiment <- filterWeakPositions(hic.experiment, threshold = 1)

## ----normalizeTechnicalBiases-------------------------------------------------
suppressWarnings(hic.experiment <- normalizeTechnicalBiases(hic.experiment))

## ----normalizeBiologicalBiases------------------------------------------------
hic.experiment <- normalizeBiologicalBiases(hic.experiment)

## ----normalizeDistanceEffect--------------------------------------------------
hic.experiment <- 
    normalizeDistanceEffect(hic.experiment, loessSampleSize = 20000)

## ----detectCompartments-------------------------------------------------------
hic.experiment <- detectCompartments(hic.experiment)

## ----plotInteractions---------------------------------------------------------
p <- plotInteractions(hic.experiment, chromosome = "Y")
p

## ----plotDistanceEffect-------------------------------------------------------
p <- plotDistanceEffect(hic.experiment)
p

## ----extractCompartments------------------------------------------------------
compartments(hic.experiment)

## ----extractConcordances------------------------------------------------------
concordances(hic.experiment)

## ----extractDifferences-------------------------------------------------------
differences(hic.experiment)

## ----plotCompartmentChanges---------------------------------------------------
p <- plotCompartmentChanges(hic.experiment, chromosome = "Y")
p

## ----plotConcordanceDifferences-----------------------------------------------
p <- plotConcordanceDifferences(hic.experiment)
p

## ----plotCentroids------------------------------------------------------------
p <- plotCentroids(hic.experiment, chromosome = "Y")
p

## ----plotSelfInteractionRatios------------------------------------------------
p <- plotSelfInteractionRatios(hic.experiment, chromosome = "Y")
p

## ----sessionInfo--------------------------------------------------------------
sessionInfo()

#' @title
#' Methods to access a \code{\link{HiCDOCDataSet}} components.
#'
#' @name
#' HiCDOCDataSet-methods
#'
#' @description
#' Retrieve information and results from a \code{\link{HiCDOCDataSet}}.
#' 
#' @examples
#' # Load an example dataset already processed 
#' # (i.e. after the detection of compartments)
#' data(exampleHiCDOCDataSetProcessed)
#' 
#' exampleHiCDOCDataSetProcessed
#' chromosomes(exampleHiCDOCDataSetProcessed)
#' sampleConditions(exampleHiCDOCDataSetProcessed)
#' sampleReplicates(exampleHiCDOCDataSetProcessed)
#' compartments(exampleHiCDOCDataSetProcessed)
#' differences(exampleHiCDOCDataSetProcessed)
#' concordances(exampleHiCDOCDataSetProcessed)
#' 
#' @return
#' A character vector (for \code{chromosomes}, \code{sampleConditions},
#' \code{sampleReplicates}), 
#' or a GRanges object
#' (for \code{compartments}, \code{concordances}, \code{differences}).
NULL

#' @describeIn HiCDOCDataSet-methods
#' Retrieves the vector of chromosome names.
#' @usage
#' NULL
#' @export
setGeneric(
    name = "chromosomes",
    def = function(object) standardGeneric("chromosomes")
)

#### sampleConditions ####
#' @describeIn HiCDOCDataSet-methods
#' Retrieves the vector of condition names, one for each sample.
#' @usage
#' NULL
#' @export
setGeneric(
    name = "sampleConditions",
    def = function(object) standardGeneric("sampleConditions")
)

#### sampleReplicates ####
#' @describeIn HiCDOCDataSet-methods
#' Retrieves the vector of replicate names, one for each sample.
#' @usage
#' NULL
#' @export
setGeneric(
    name = "sampleReplicates",
    def = function(object) standardGeneric("sampleReplicates")
)

#### compartments ####
#' @describeIn HiCDOCDataSet-methods
#' Retrieves a \code{GenomicRange} of the compartment of every position
#' in every condition.
#' @param passChecks logical. Display only the concordances/compartments for 
#' the chromosomes passing sanity checks.
#' @usage
#' NULL
#' @export
setGeneric(
    name = "compartments",
    def = function(object, passChecks = TRUE) standardGeneric("compartments")
)

#### differences ####
#' @describeIn HiCDOCDataSet-methods
#' Retrieves a \code{GenomicRange} of the significant compartment differences
#' between conditions, and their p-values.
#' @usage
#' NULL
#' @param object
#' a HiCDOCDataSet object
#' @param threshold
#' a numeric value between 0 and 1. If no threshold, all the differences will
#' be printed even the non significant ones. Otherwise the differences printed
#' are filtered to show the ones with an adjusted p-value <= \code{threshold}.
#' @export
#' @usage
#' NULL
setGeneric(
    name = "differences",
    def = function(object, threshold = 0.05) standardGeneric("differences")
)

#### concordances ####
#' @describeIn HiCDOCDataSet-methods
#' Retrieves a \code{GenomicRange} of the concordance (confidence in assigned
#' compartment) of every position in every replicate.
#' @usage
#' NULL
#' @export
setGeneric(
    name = "concordances",
    def = function(object, passChecks = TRUE) standardGeneric("concordances")
)

#' @title
#' Access the parameters of a \code{\link{HiCDOCDataSet}}.
#
#' @name
#' HiCDOCDataSet-parameters
#' 
#' @description
#' Retrieves or sets parameters used for filtering, normalization, and
#' prediciton of compartments.
#'
#' @details
#' A \code{\link{HiCDOCDataSet}}'s parameters are automatically set to default
#' values retrieved from \code{\link{defaultHiCDOCParameters}}. They are
#' accessed by filtering, normalization, and compartment detection functions.
#' If those functions are called with custom arguments, the object's
#' parameters are updated to record the actual parameters used. If the
#' object's parameters are customized before calling the functions, the
#' custom parameters will be used.
#'
#' See
#' \code{\link{filterSmallChromosomes}},
#' \code{\link{filterSparseReplicates}},
#' \code{\link{filterWeakPositions}},
#' \code{\link{normalizeDistanceEffect}}, and
#' \code{\link{detectCompartments}},
#' for details on how these parameters are used.

#' \subsection{All parameters are listed here:}{
#'     \describe{
#'         \item{\code{smallChromosomeThreshold}}{
#'             The minimum length (number of positions) for a chromosome to be
#'             kept when filtering with \code{\link{filterSmallChromosomes}}.
#'             Defaults to
#'             \code{defaultHiCDOCParameters$smallChromosomeThreshold} = 100.
#'         }
#'         \item{\code{sparseReplicateThreshold}}{
#'             The minimum percentage of non-zero interactions for a chromosome
#'             replicate to be kept when filtering with
#'             \code{\link{filterSparseReplicates}}. If a chromosome replicate's
#'             percentage of non-zero interactions is lower than this value, it
#'             is removed. Defaults to
#'             \code{defaultHiCDOCParameters$smallChromosomeThreshold} = 30%.
#'         }
#'         \item{\code{weakPositionThreshold}}{
#'             The minimum average interaction for a position to be kept when
#'             filtering with \code{\link{filterWeakPositions}}. If a position's
#'             average interaction with the entire chromosome is lower than this
#'             value in any of the replicates, it is removed from all replicates
#'             and conditions. Defaults to
#'             \code{defaultHiCDOCParameters$smallChromosomeThreshold} = 1.
#'         }
#'         \item{\code{cyclicLoessSpan}}{
#'             The span for cyclic loess normalization used in 
#'             \code{\link{normalizeTechnicalBiases}}. This value is passed to 
#'             \code{multiHiCcompare::cyclic_loess}. 
#'             Defaults to NA indicating that span will be automatically 
#'             calculated using generalized cross validation.
#'             For large dataset, it is highly recommended to set this value
#'             to reduce computing time and necessary memory.
#'         }
#'         \item{\code{loessSampleSize}}{
#'             The number of positions used as a sample to estimate the effect
#'             of distance on proportion of interactions when normalizing with
#'             \code{\link{normalizeDistanceEffect}}. Defaults to
#'             \code{defaultHiCDOCParameters$loessSampleSize} = 20000.
#'         }
#'         \item{\code{kMeansDelta}}{
#'             The convergence stop criterion for the clustering when detecting
#'             compartments with \code{\link{detectCompartments}}. When the
#'             centroids' distances between two iterations is lower than this
#'             value, the clustering stops. Defaults to
#'             \code{defaultHiCDOCParameters$kMeansDelta} = 0.0001.
#'         }
#'         \item{\code{kMeansIterations}}{
#'             The maximum number of iterations during clustering when detecting
#'             compartments with \code{\link{detectCompartments}}. Defaults to
#'             \code{defaultHiCDOCParameters$kMeansIterations} = 50.
#'         }
#'         \item{\code{kMeansRestarts}}{
#'             The amount of times the clustering is restarted when detecting
#'             compartments with \code{\link{detectCompartments}}. For each
#'             restart, the clustering iterates until convergence or reaching
#'             the maximum number of iterations. The clustering that minimizes
#'             inner-cluster variance is selected. Defaults to
#'             \code{defaultHiCDOCParameters$kMeansRestarts} = 20.
#'         }
#'         \item{\code{PC1CheckThreshold}}{
#'             The minimum percentage of variance that should be explained by
#'             the first principal component of centroids to pass sanity check.
#'             Defaults to 
#'             \code{defaultHiCDOCParameters$PC1CheckThreshold} = 0.75
#'         }
#'     }
#' }
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#'
#' @examples
#' data(exampleHiCDOCDataSet)
#'
#' # Retrieve parameters
#' parameters(exampleHiCDOCDataSet)
#'
#' # Set parameters
#' parameters(exampleHiCDOCDataSet) <- list("smallChromosomeThreshold" = 50)
#' parameters(exampleHiCDOCDataSet) <- list(
#'     "weakPositionThreshold" = 10,
#'     "kMeansRestarts" = 30
#' )
NULL

#### parameters ####
#' @rdname HiCDOCDataSet-parameters
#' @usage NULL
#' @export
setGeneric(
    name = "parameters",
    def = function(object) standardGeneric("parameters")
)

#### parameters <- ####
#' @rdname HiCDOCDataSet-parameters
#' @usage NULL
#' @param value a named list containing the names and valued of the
#' parameters to change (see Details).
#' @export
setGeneric(
    name = "parameters<-",
    def = function(object, value) standardGeneric("parameters<-")
)

# code inspired from multiHiCcompare package (fonction smartApply)
.internalLapply <- function(parallel, ...) {
    if (parallel) {
        if (!is.null(BiocParallel::bpparam()$RNGseed)) {
            warning(
                "The use of RNGseed may not be ensured ",
                "See ?detectCompartments for more details",
                call. = FALSE
            )
        }
        BiocParallel::bplapply(..., BPPARAM = BiocParallel::bpparam())
    } else {
        pbapply::pblapply(...)
    }
}

#' @title
#' Filter small chromosomes.
#'
#' @description
#' Removes chromosomes whose length (in number of positions) is smaller than the
#' threshold.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#' @param threshold
#' The minimum length (number of positions) for a chromosome to be kept.
#' Defaults to \code{object$smallChromosomeThreshold} which is originally set to
#' \code{defaultHiCDOCParameters$smallChromosomeThreshold} = 100.
#'
#' @return
#' A filtered \code{\link{HiCDOCDataSet}}.
#'
#' @seealso
#' \code{\link{filterSparseReplicates}},
#' \code{\link{filterWeakPositions}},
#' \code{\link{HiCDOC}}
#'
#' @examples
#' data(exampleHiCDOCDataSet)
#' object <- exampleHiCDOCDataSet
#'
#' chromosomes(object)
#' object <- filterSmallChromosomes(object)
#' chromosomes(object)
#'
#' @export
filterSmallChromosomes <- function(object, threshold = NULL) {
    .validateSlots(
      object,
      slots = c("chromosomes", "totalBins", "parameters")
    )

    if (!is.null(threshold)) {
        object@parameters$smallChromosomeThreshold <- threshold
    }
    object@parameters <- .validateParameters(object@parameters)
    threshold <- object@parameters$smallChromosomeThreshold
    if(threshold<=1) {
        threshold <- 1
        message("Keeping non-empty chromosomes")
    } else {
        message(
            "Keeping chromosomes with at least ",
            threshold,
            " positions."
        )
    }

    bigChromosomes <- vapply(
        object@totalBins,
        function(totalBins) totalBins >= threshold,
        FUN.VALUE = TRUE
    )
    bigChromosomeNames <- names(bigChromosomes)[bigChromosomes]
    bigChromosomeNames <- gtools::mixedsort(bigChromosomeNames)
    smallChromosomeNames <- object@chromosomes[
        !(object@chromosomes %in% bigChromosomeNames)
    ]
    if(length(smallChromosomeNames)>0) {
        object <- reduceHiCDOCDataSet(
            object,
            chromosomes = bigChromosomeNames,
            dropLevels = TRUE
        )
        message(
            "Kept ",
            length(bigChromosomeNames),
            " chromosome",
            if (length(bigChromosomeNames) != 1) "s",
            if (length(bigChromosomeNames) > 0) ": " else ".",
            paste(bigChromosomeNames, collapse = ", ")
        )
        message(
            "Removed ",
            length(smallChromosomeNames),
            " chromosome",
            if (length(smallChromosomeNames) != 1) "s",
            if (length(smallChromosomeNames) > 0) ": " else ".",
            paste(smallChromosomeNames, collapse = ", ")
        )
        
        if (length(bigChromosomeNames) == 0) {
            warning("No data left!", call. = FALSE)
        }
    } else {
        message("All chromosomes kept.")
    }
    return(object)
}

data(exampleHiCDOCDataSet)

test_that("filterSmallChromosomes behave as expected with default filter", {
    # No filter on the example dataset
    expect_message(
        object <- filterSmallChromosomes(exampleHiCDOCDataSet),
        "Keeping chromosomes with at least 100 positions."
    )
    expect_equal(length(object@chromosomes), 3)
    expect_equal(nrow(object), 40240)
    assay <- SummarizedExperiment::assay(object)
    expect_equal(sum(!is.na(assay)), 166150)
    expect_equal(colSums(assay, na.rm=TRUE), 
                 c(7891910, 20681568, 7810636, 28374121, 
                   15745554, 21976428, 21352885))
})

test_that("filterSmallChromosomes behave as expected with custom filter", {
    # Filter on 1 chromosome
    expect_message(
        object <- filterSmallChromosomes(exampleHiCDOCDataSet, 161),
        "Keeping chromosomes with at least 161 positions."
    )
    expect_identical(object@chromosomes, "Z")
    expect_equal(nrow(object), 20100)
    expect_identical(object@parameters$smallChromosomeThreshold, 161)
    assay <- SummarizedExperiment::assay(object)
    expect_equal(sum(!is.na(assay)), 40200)
    expect_equal(colSums(assay, na.rm=TRUE), 
                 c(0, 0, 0, 22706459, 
                   0, 0, 7635309))
})

#' @description
#' Reduces a \code{\link{HiCDOCDataSet}} by keeping only given chromosomes.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#' @param chromosomeNames
#' The names of chromosomes to keep.
#' @param dropLevels
#' Whether or not to also remove unused factor levels after filtering. Should
#' be set to FALSE if the reduced objects are meant to be re-combined later.
#' Defaults to TRUE.
#'
#' @return
#' A reduced \code{\link{HiCDOCDataSet}}.
#'
#' @keywords internal
#' @noRd
.reduceHiCDOCChromosomes <- function(object, chromosomeNames, dropLevels) {
    chromosomeIds <- which(object@chromosomes %in% chromosomeNames)
    object@chromosomes <- object@chromosomes[chromosomeIds]

    object@weakBins <- object@weakBins[chromosomeIds]
    object@totalBins <- object@totalBins[chromosomeIds]
    object@validAssay <- object@validAssay[chromosomeIds]

    toKeep <- S4Vectors::`%in%`(
        S4Vectors::mcols(object)$chromosome,
        chromosomeNames
    )
    object <- object[toKeep, ]

    if (dropLevels) {
        SummarizedExperiment::mcols(object)$chromosome <- droplevels(
            SummarizedExperiment::mcols(object)$chromosome
        )
        object <- InteractionSet::reduceRegions(object)
        Seqinfo::seqlevels(
            InteractionSet::regions(object),
            pruning.mode = "coarse"
        ) <- object@chromosomes

    }
    for (slotName in c(
        "distances",
        "selfInteractionRatios",
        "compartments",
        "concordances",
        "differences",
        "centroids",
        "comparisons"
    )) {
        if (!is.null(slot(object, slotName))) {
            tmp <- slot(object, slotName)
            if (is(tmp, "data.table")) {
                tmp <- tmp[chromosome %in% chromosomeNames]
                if (dropLevels) {
                    tmp[, chromosome := droplevels(chromosome)]
                }
            } else {
                if (!is(tmp, "GRanges")) {
                    stop("malformed HiCDOCDataSet")
                }
                if (dropLevels) {
                    Seqinfo::seqlevels(
                        tmp,
                        pruning.mode = "coarse"
                    ) <- object@chromosomes
                } else {
                    tmp <- tmp[S4Vectors::`%in%`(tmp@seqnames, chromosomeNames)]
                }
            }
            slot(object, slotName)  <- tmp
        }
    }
    return(object)
}

#' @description
#' Reduces a \code{\link{HiCDOCDataSet}} by keeping only given conditions.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#' @param conditionNames
#' The names of conditions to keep.
#' @param dropLevels
#' Whether or not to also remove unused factor levels after filtering. Should
#' be set to FALSE if the reduced objects are meant to be re-combined later.
#' Defaults to TRUE.
#'
#' @return
#' A reduced \code{\link{HiCDOCDataSet}}.
#'
#' @keywords internal
#' @noRd
.reduceHiCDOCConditions <- function(object, conditionNames, dropLevels) {
    conditionIds <- which(object$condition %in% conditionNames)
    object <- object[, conditionIds]

    object@validAssay <- .determineValids(object)
    for (slotName in c(
        "distances",
        "selfInteractionRatios",
        "compartments",
        "concordances",
        "centroids"
    )) {
        if (!is.null(slot(object, slotName))) {
            tmp <- slot(object, slotName)
            if (is(tmp, "data.table")) {
                tmp <- tmp[condition %in% conditionNames]
                if (dropLevels) {
                    tmp[, condition := droplevels(condition)]
                }
            } else {
                if (!is(tmp, "GRanges")) {
                    stop("malformed HiCDOCDataSet")
                }
                tmp <- tmp[tmp$condition %in% conditionNames]
                if (dropLevels) {
                    tmp$condition <- droplevels(tmp$condition)
                }
            }
            slot(object, slotName)  <- tmp
        }
    }
    return(object)
}

#' @description
#' Reduces a \code{\link{HiCDOCDataSet}} by keeping only given replicates.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#' @param replicateNames
#' The names of replicates to keep.
#' @param dropLevels
#' Whether or not to also remove unused factor levels after filtering. Should
#' be set to FALSE if the reduced objects are meant to be re-combined later.
#' Defaults to TRUE.
#'
#' @return
#' A reduced \code{\link{HiCDOCDataSet}}.
#'
#' @keywords internal
#' @noRd
.reduceHiCDOCReplicates <- function(object, replicateNames, dropLevels) {
    replicateIds <- which(object$replicate %in% replicateNames)
    object <- object[, replicateIds]

    object@validAssay <- .determineValids(object)


    for (slotName in c(
        "distances",
        "selfInteractionRatios",
        "concordances"
    )) {
        if (!is.null(slot(object, slotName))) {
            tmp <- slot(object, slotName)
            if (is(tmp, "data.table")) {
                tmp <- tmp[replicate %in% replicateNames]
                if (dropLevels) {
                    tmp[, replicate := droplevels(replicate)]
                }
            } else {
                if (!is(tmp, "GRanges")) {
                    stop("malformed HiCDOCDataSet")
                }
                tmp <- tmp[tmp$replicate %in% replicateNames]
                if (dropLevels) {
                    tmp$replicate <- droplevels(tmp$replicate)
                }
            }
            slot(object, slotName)  <- tmp
        }
    }
    return(object)
}

#' @title
#' Reduce a \code{\link{HiCDOCDataSet}}.
#'
#' @description
#' Reduces a \code{\link{HiCDOCDataSet}} by keeping only given chromosomes,
#' conditions, or replicates.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#' @param chromosomes
#' The chromosome names or indices in \code{chromosomes(object)} to keep.
#' Defaults to NULL.
#' @param conditions
#' The condition names in \code{sampleConditions(object)} to keep. 
#' Defaults to NULL.
#' @param replicates
#' The replicate names in \code{sampleReplicates(object)} to keep. 
#' Defaults to NULL.
#' @param dropLevels
#' Whether or not to also remove unused factor levels after filtering. Should
#' be set to FALSE if the reduced objects are meant to be re-combined later.
#' Defaults to TRUE.
#'
#' @return
#' A reduced \code{\link{HiCDOCDataSet}}.
#'
#' @examples
#' data(exampleHiCDOCDataSet)
#' reduced <- reduceHiCDOCDataSet(exampleHiCDOCDataSet, chromosomes = c(1, 2))
#'
#' @export
reduceHiCDOCDataSet <- function(
    object,
    chromosomes = NULL,
    conditions = NULL,
    replicates = NULL,
    dropLevels = TRUE
) {
    if (!is.null(object@differences)) {
        warning(
            "You should not reduce a HiCDOCDataSet after calling ",
            "'detectCompartments()'. All chromosomes, conditions, and ",
            "replicates have been used in the computations.",
            call. = FALSE
        )
    }

    if (!is.null(chromosomes)) {
        chromosomeNames <- .validateNames(object, chromosomes, "chromosomes")
        object <- .reduceHiCDOCChromosomes(object, chromosomeNames, dropLevels)
    }

    if (!is.null(conditions)) {
        conditionNames <- .validateNames(object, conditions, "conditions")
        object <- .reduceHiCDOCConditions(object, conditions, dropLevels)
    }

    if (!is.null(replicates)) {
        replicateNames <- .validateNames(object, replicates, "replicates")
        object <- .reduceHiCDOCReplicates(object, replicates, dropLevels)
    }

    return(object)
}

test_that("plotDistanceEffect behaves as expected", {
    data(exampleHiCDOCDataSet)
    object <- reduceHiCDOCDataSet(exampleHiCDOCDataSet, chromosomes = "X")
    expect_message(pp <- plotDistanceEffect(exampleHiCDOCDataSet))
    expect_is(pp, "ggplot")
    # No error when printed
    expect_error(print(pp), NA)
})

data("exampleHiCDOCDataSet")
data("exampleHiCDOCDataSetProcessed")

test_chromosome_levels <- function(object, expectedLevels){
    expect_identical(levels(object@distances$chromosome), expectedLevels)
    expect_identical(levels(object@selfInteractionRatios$chromosome), expectedLevels)
    expect_identical(seqlevels(object@compartments), expectedLevels)
    expect_identical(seqlevels(object@concordances), expectedLevels)
    expect_identical(seqlevels(object@differences), expectedLevels)
    expect_identical(levels(object@centroids$chromosome), expectedLevels)
}

test_condition_levels <- function(object, expectedLevels){
    expect_identical(levels(object@distances$condition), expectedLevels)
    expect_identical(levels(object@selfInteractionRatios$condition),
                     expectedLevels)
    expect_identical(levels(object@compartments$condition), expectedLevels)
    expect_identical(levels(object@concordances$condition), expectedLevels)
    expect_identical(levels(object@centroids$condition), expectedLevels)
}

test_replicate_levels <- function(object, expectedLevels){
    expect_identical(levels(object@distances$replicate), expectedLevels)
    expect_identical(levels(object@selfInteractionRatios$replicate),
                     expectedLevels)
    expect_identical(levels(object@concordances$replicate), expectedLevels)
}

test_that("reduceHiCDOCDataSet return correct errors", {
    # On chromosomes
    expect_error(
        reduceHiCDOCDataSet(exampleHiCDOCDataSet, chromosomes = c(5, 6)),
        "Unknown chromosomes"
    )
    expect_error(
        reduceHiCDOCDataSet(exampleHiCDOCDataSet, chromosomes = "chr1"),
        "Unknown chromosome"
    )
    # On conditions
    expect_error(
        reduceHiCDOCDataSet(exampleHiCDOCDataSet, conditions = c(3, 4)),
        "Unknown condition: 4"
    )
    expect_error(
        reduceHiCDOCDataSet(exampleHiCDOCDataSet, conditions = "cond1"),
        "Unknown condition"
    )
    # On replicates
    expect_error(
        reduceHiCDOCDataSet(exampleHiCDOCDataSet, replicates = c(3, 4)),
        "Unknown replicates"
    )
    expect_error(
        reduceHiCDOCDataSet(exampleHiCDOCDataSet, replicates = "rep1"),
        "Unknown replicate"
    )
})

test_that("reduceHiCDOCDataSet works if select chromosome, dropLevels", {
    expect_warning(
        objectRed <- reduceHiCDOCDataSet(exampleHiCDOCDataSetProcessed, 
                                         chromosomes = "X"),
        "You should not reduce a HiCDOCDataSet after"
    )
    # Chromosomes
    expect_identical(objectRed@chromosomes, "X")
    expect_identical(objectRed@totalBins, c("X" = 120))
    expect_identical(objectRed@weakBins, list("X" = c(171, 200)))
    # Doesn't remove replicates & conditions
    expect_identical(objectRed$replicate,
                     c("R2", "R1", "R1", "R2", "R2", "R1", "R3"))
    expect_identical(objectRed$condition, c("2", "1", "2", "1", "3", "3", "1"))
    # Interactions
    expect_equal(dim(SummarizedExperiment::assay(objectRed)), c(7021, 7))
    expect_equal(round(sum(SummarizedExperiment::assay(objectRed), na.rm=TRUE),2), 
                 36.77)
    
    # Objects produced by detectCompartments
    test_chromosome_levels(objectRed, "X")
    test_condition_levels(objectRed, c("1", "2", "3"))
    test_replicate_levels(objectRed, c("R1", "R2", "R3"))
})

test_that("reduceHiCDOCDataSet works if select chromosome, keep levels", {
    expect_warning(
        objectRed <- reduceHiCDOCDataSet(
            exampleHiCDOCDataSetProcessed,
            chromosomes = "X",
            dropLevels = FALSE
        ),
        "You should not reduce a HiCDOCDataSet after"
    )
    # Chromosomes
    expect_identical(objectRed@chromosomes, "X")
    expect_identical(objectRed@totalBins, c("X" = 120))
    expect_identical(objectRed@weakBins, list("X" = c(171, 200)))
    # Doesn't remove replicates & conditions
    expect_identical(objectRed$replicate, exampleHiCDOCDataSetProcessed$replicate)
    expect_identical(objectRed$condition, exampleHiCDOCDataSetProcessed$condition)

    # Interactions
    expect_equal(dim(SummarizedExperiment::assay(objectRed)), c(7021, 7))
    expect_equal(round(sum(SummarizedExperiment::assay(objectRed), na.rm=TRUE),2), 
                 36.77)
    
    # Objects prduced by detectCompartments
    test_chromosome_levels(objectRed, c("X", "Y", "Z"))
    test_condition_levels(objectRed, c("1", "2", "3"))
    test_replicate_levels(objectRed, c("R1", "R2", "R3"))
})

test_that("reduceHiCDOCDataSet works if select condition, drop levels", {
    expect_warning(
        objectRed <- reduceHiCDOCDataSet(
            exampleHiCDOCDataSetProcessed,
            conditions = c(1, 2),
            dropLevels = TRUE
        ),
        "You should not reduce a HiCDOCDataSet after"
    )
    # Chromosomes
    expect_identical(objectRed@chromosomes, exampleHiCDOCDataSetProcessed@chromosomes)
    expect_identical(objectRed@totalBins, exampleHiCDOCDataSetProcessed@totalBins)
    expect_identical(objectRed@weakBins,  exampleHiCDOCDataSetProcessed@weakBins)
    expect_identical(objectRed$replicate, c("R2", "R1", "R1", "R2", "R3"))
    expect_identical(objectRed$condition, c("2", "1", "2", "1", "1"))

    # Interactions
    expect_equal(dim(SummarizedExperiment::assay(objectRed)), c(39524, 5))
    expect_equal(sum(SummarizedExperiment::assay(objectRed), na.rm=TRUE), 
                 115.80, tolerance=1e-2)
    
    # Objects prduced by detectCompartments
    test_chromosome_levels(objectRed, c("X", "Y", "Z"))
    test_condition_levels(objectRed, c("1", "2"))
    test_replicate_levels(objectRed, c("R1", "R2", "R3"))
})

test_that("reduceHiCDOCDataSet works if select replicate, drop levels", {
    expect_warning(
        objectRed <- reduceHiCDOCDataSet(
            exampleHiCDOCDataSetProcessed,
            replicate = "R1",
            dropLevels = TRUE
        ),
        "You should not reduce a HiCDOCDataSet after"
    )
    # Chromosomes
    expect_identical(objectRed@chromosomes, exampleHiCDOCDataSetProcessed@chromosomes)
    expect_identical(objectRed@totalBins, exampleHiCDOCDataSetProcessed@totalBins)
    expect_identical(objectRed@weakBins,  exampleHiCDOCDataSetProcessed@weakBins)
    expect_identical(objectRed$replicate, c("R1", "R1", "R1"))
    expect_identical(objectRed$condition, c("1", "2", "3"))

    # Interactions
    expect_equal(dim(SummarizedExperiment::assay(objectRed)), c(39524, 3))
    expect_equal(sum(SummarizedExperiment::assay(objectRed), na.rm=TRUE), 
                 48.62, tolerance=1e-2)
    
    # Objects prduced by detectCompartments
    test_chromosome_levels(objectRed, c("X", "Y", "Z"))
    test_condition_levels(objectRed, c("1", "2", "3"))
    test_replicate_levels(objectRed, c("R1"))
})

test_that("reduceHiCDOCDataSet works if select chr, cond & rep, keep levels", {
    expect_warning(
        objectRed <- reduceHiCDOCDataSet(
            exampleHiCDOCDataSetProcessed,
            chromosome = "X",
            replicate = "R1",
            condition = "1",
            dropLevels = FALSE
        ),
        "You should not reduce a HiCDOCDataSet after"
    )
    # Chromosomes
    expect_identical(objectRed@chromosomes, "X")
    expect_identical(objectRed@totalBins, c("X" = 120))
    expect_identical(objectRed@weakBins,  list("X" = c(171, 200)))
    expect_identical(objectRed$replicate, "R1")
    expect_identical(objectRed$condition, "1")

    # Interactions
    expect_equal(dim(SummarizedExperiment::assay(objectRed)), c(7021, 1))
    expect_equal(sum(SummarizedExperiment::assay(objectRed), na.rm=TRUE), 
                 -12.19, tolerance=1e-2)

    # Objects prduced by detectCompartments
    test_chromosome_levels(objectRed, c("X", "Y", "Z"))
    test_condition_levels(objectRed, c("1", "2", "3"))
    test_replicate_levels(objectRed, c("R1", "R2", "R3"))
})

#' @title
#' Plot the distance effect.
#'
#' @description
#' Plots the distance effect on proportion of interactions.
#' Each point is a cell in the interaction matrix, such that
#' the x-axis is the distance with respect to the diagonal,
#' the y-axis is number of counts.
#' Dots are binned.
#'
#' @param object
#' A \code{\link{HiCDOCDataSet}}.
#' @param chromosome
#' Name (character) or index of the chromosome, if the plot should be
#' restricted to only one chromosome. Default to NULL.
#' @param transformX
#' Transformation of the X axis. Default to "identity". See
#' \code{\link[ggplot2]{scale_x_continuous}} for other accepted values.
#' @param transformY
#' Transformation of the Y axis. Default to "identity". See
#' \code{\link[ggplot2]{scale_y_continuous}} for other accepted values.
#'
#' @return
#' A \code{ggplot}.
#'
#' @examples
#' data(exampleHiCDOCDataSet)
#' plotDistanceEffect(exampleHiCDOCDataSet)
#'
#' @export
plotDistanceEffect <- function(object, chromosome = NULL, transformX="identity", transformY="identity") {
    .validateSlots(object, slots = c("interactions"))
    if (!is.null(chromosome)) {
        if (length(chromosome) > 1) {
            warning(
                "`chromosome` should be of length 1, taking the first one."
            )
            chromosome < chromosome[1]
        }
        chromosomeName <- .validateNames(object, chromosome, "chromosomes")
        rowsId <- as.logical(
            S4Vectors::mcols(object)$chromosome == chromosomeName
        )
        addTitle <- paste(", chromosome", chromosomeName)
    } else {
        rowsId <- rep(TRUE, length(object))
        addTitle <- ""
    }

    distances <- InteractionSet::pairdist(object, type = "mid")[rowsId]
    matrixAssay <- SummarizedExperiment::assay(object)[rowsId, ]
    dfDistance <- data.table::data.table(
        "distance" = rep(distances, ncol(matrixAssay)),
        "interaction" = as.vector(matrixAssay)
    )
    dfDistance <- dfDistance[!is.na(interaction)]
    
    plot <- ggplot(
        dfDistance,
        aes(x = distance, y = interaction)
    ) + geom_bin2d() + scale_fill_gradient(
        low = "white",
        high = "blue",
        trans = "log2"
    ) + geom_point(col = "transparent") + 
        geom_smooth(col = "red") + 
        scale_y_continuous(trans=transformY)  + 
        scale_x_continuous(trans=transformX) + theme_bw() 
    
    margPlot <-  ggplot(dfDistance, aes(x = distance)) +
        geom_histogram(fill="transparent", col="black") + 
        theme_minimal() +
        theme(panel.grid.major.x = element_blank(),
              axis.title.x = element_blank(),
              axis.text.x = element_blank(),
              axis.ticks.x = element_blank()) 
    
    layoutMatrix <- rbind(c(1,1,1,1,1,1,1,3),
                          c(2,2,2,2,2,2,2,3),
                          c(2,2,2,2,2,2,2,3),
                          c(2,2,2,2,2,2,2,3),
                          c(2,2,2,2,2,2,2,3))
    
    plot <- gridExtra::arrangeGrob(
                margPlot,
                plot + theme(legend.position = "none"),
                cowplot::get_plot_component(plot, 'guide-box-right', return_all = TRUE),
                layout_matrix = layoutMatrix,
                padding = unit(0.2, "lines"),
                top = paste0("Distance effect", addTitle)
            )
    return(cowplot::ggdraw(plot))
}

test_that("defaultHiCDOCParameters has the expected format", {
    expect_is(defaultHiCDOCParameters, "list")
    expect_identical(
        names(defaultHiCDOCParameters),
        c(
            "smallChromosomeThreshold",
            "sparseReplicateThreshold",
            "weakPositionThreshold",
            "cyclicLoessSpan",
            "loessSampleSize",
            "kMeansDelta",
            "kMeansIterations",
            "kMeansRestarts",
            "PC1CheckThreshold"
        )
    )
})

test_that("defaultHiCDOCParameters has the expected values", {
    expect_equal(defaultHiCDOCParameters$smallChromosomeThreshold, 100)
    expect_equal(defaultHiCDOCParameters$weakPositionThreshold, 1)
    expect_equal(defaultHiCDOCParameters$sparseReplicateThreshold, 0.3)
    expect_equal(defaultHiCDOCParameters$cyclicLoessSpan, NA_real_)
    expect_equal(defaultHiCDOCParameters$loessSampleSize, 20000)
    expect_equal(defaultHiCDOCParameters$kMeansDelta, 1e-04)
    expect_equal(defaultHiCDOCParameters$kMeansIterations, 50)
    expect_equal(defaultHiCDOCParameters$kMeansRestarts, 20)
    expect_equal(defaultHiCDOCParameters$PC1CheckThreshold, 0.75)
})

---
title: "HiCDOC"
date: "`r format(Sys.Date(), '%m/%d/%Y')`"
author: "Cyril Kurylo & Matthias Zytnicki & Sylvain Foissac & Elise Maigné"
output:
    BiocStyle::html_document:
       fig_width: 7
       fig_height: 5
       toc_float: true
bibliography: library.bib
vignette: >
  %\VignetteIndexEntry{HiCDOC}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
  %\VignettePackage{HiCDOC}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>"
)
options(warn=-1)
```

# Introduction

The aim of HiCDOC is to detect significant A/B compartment changes, using Hi-C
data with replicates.

HiCDOC normalizes intrachromosomal Hi-C matrices, uses unsupervised learning to
predict A/B compartments from multiple replicates, and detects significant
compartment changes between experiment conditions.

It provides a collection of functions assembled into a pipeline:

1. [Filter](#filtering-data):
    1. Remove chromosomes which are too small to be useful.
    2. Filter sparse replicates to remove uninformative replicates with few
    interactions.
    3. Filter positions (*bins*) which have too few interactions.
2. [Normalize](#normalizing-biases):
    1. Normalize technical biases (inter-matrix normalization) using
    cyclic loess normalization [@multihiccompare], so that
    matrices are comparable.
    2. Normalize biological biases (intra-matrix normalization) using
    Knight-Ruiz matrix balancing [@kr], so that
    all the bins are comparable.
    3. Normalize the distance effect, which results from higher interaction
    proportions between closer regions, with a MD loess.
3. [Predict](#predicting-compartments-and-differences):
    1. Predict compartments using
    constrained K-means [@kmeans].
    2. Detect significant differences between experiment conditions.
4. [Visualize](#visualizing-data-and-results):
    1. Plot the interaction matrices of each replicate.
    2. Plot the overall distance effect on the proportion of interactions.
    3. Plot the compartments in each chromosome, along with their concordance
    (confidence measure) in each replicate, and significant changes between
    experiment conditions.
    4. Plot the overall distribution of concordance differences.
    5. Plot the result of the PCA on the compartments' centroids.
    6. Plot the boxplots of self interaction ratios (differences between self
    interactions and the medians of other interactions) of each compartment,
    which is used for the A/B classification.

# Installation

HiCDOC can be installed from Bioconductor:

```{r, eval=FALSE}
if (!requireNamespace("BiocManager", quietly=TRUE))
    install.packages("BiocManager")
BiocManager::install("HiCDOC")
```

The package can then be loaded:

```{r}
library(HiCDOC)
```

# Importing Hi-C data

HiCDOC can import Hi-C data sets in various different formats:
- Tabular `.tsv` files.
- Cooler `.cool` or `.mcool` files.
- Juicer `.hic` files.
- HiC-Pro `.matrix` and `.bed` files.

## Tabular files

A tabular file is a tab-separated multi-replicate sparse matrix with a header:

chromosome    position 1    position 2    C1.R1    C1.R2    C2.R1    ...
Y             1500000       7500000       145      184      72       ...
...

The number of interactions between `position 1` and `position 2` of
`chromosome` are reported in each `condition.replicate` column. There is no
limit to the number of conditions and replicates.

To load Hi-C data in this format:

```{r tabFormat, eval = FALSE}
hic.experiment <- HiCDOCDataSetFromTabular('path/to/data.tsv')
```

## Cooler files

To load `.cool` or `.mcool` files generated by [Cooler][cooler-documentation]
[@cooler]:

```{r coolFormat, eval = FALSE}
# Path to each file
paths = c(
    'path/to/condition-1.replicate-1.cool',
    'path/to/condition-1.replicate-2.cool',
    'path/to/condition-2.replicate-1.cool',
    'path/to/condition-2.replicate-2.cool',
    'path/to/condition-3.replicate-1.cool'
)

# Replicate and condition of each file. Can be names instead of numbers.
replicates <- c(1, 2, 1, 2, 1)
conditions <- c(1, 1, 2, 2, 3)

# Resolution to select in .mcool files
binSize = 500000

# Instantiation of data set
hic.experiment <- HiCDOCDataSetFromCool(
    paths,
    replicates = replicates,
    conditions = conditions,
    binSize = binSize # Specified for .mcool files.
)
```

## Juicer files

To load `.hic` files generated by [Juicer][juicer-documentation] [@juicer]:

```{r hicFormat, eval = FALSE}
# Path to each file
paths = c(
    'path/to/condition-1.replicate-1.hic',
    'path/to/condition-1.replicate-2.hic',
    'path/to/condition-2.replicate-1.hic',
    'path/to/condition-2.replicate-2.hic',
    'path/to/condition-3.replicate-1.hic'
)

# Replicate and condition of each file. Can be names instead of numbers.
replicates <- c(1, 2, 1, 2, 1)
conditions <- c(1, 1, 2, 2, 3)

# Resolution to select
binSize <- 500000

# Instantiation of data set
hic.experiment <- HiCDOCDataSetFromHiC(
    paths,
    replicates = replicates,
    conditions = conditions,
    binSize = binSize
)
```

## HiC-Pro files

To load `.matrix` and `.bed` files generated by [HiC-Pro][hicpro-documentation]
[@hicpro]:

```{r hicproFormat, eval = FALSE}
# Path to each matrix file
matrixPaths = c(
    'path/to/condition-1.replicate-1.matrix',
    'path/to/condition-1.replicate-2.matrix',
    'path/to/condition-2.replicate-1.matrix',
    'path/to/condition-2.replicate-2.matrix',
    'path/to/condition-3.replicate-1.matrix'
)

# Path to each bed file
bedPaths = c(
    'path/to/condition-1.replicate-1.bed',
    'path/to/condition-1.replicate-2.bed',
    'path/to/condition-2.replicate-1.bed',
    'path/to/condition-2.replicate-2.bed',
    'path/to/condition-3.replicate-1.bed'
)

# Replicate and condition of each file. Can be names instead of numbers.
replicates <- c(1, 2, 1, 2, 1)
conditions <- c(1, 1, 2, 2, 3)

# Instantiation of data set
hic.experiment <- HiCDOCDataSetFromHiCPro(
    matrixPaths = matrixPaths,
    bedPaths = bedPaths,
    replicates = replicates,
    conditions = conditions
)
```

# Running the HiCDOC pipeline

An example dataset can be loaded from the HiCDOC package:

```{r reloadExample}
data(exampleHiCDOCDataSet)
```

Once your data is loaded, you can run all the filtering, normalization, and
prediction steps with the command : **`HiCDOC(exampleHiCDOCDataSet)`**.
This one-liner runs all the steps detailed below.

## Filtering data

Remove small chromosomes of length smaller than 100 positions 
(100 is the default value):

```{r filterSmallChromosomes}
hic.experiment <- filterSmallChromosomes(exampleHiCDOCDataSet, threshold = 100)
```

Remove sparse replicates filled with less than 30% non-zero interactions
(30% is the default value):

```{r filterSparseReplicates}
hic.experiment <- filterSparseReplicates(hic.experiment, threshold = 0.3)
```

Remove weak positions with less than 1 interaction in average 
(1 is the default value):

```{r filterWeakPositions}
hic.experiment <- filterWeakPositions(hic.experiment, threshold = 1)
```

## Normalizing biases

### Technical biases

Normalize technical biases such as sequencing depth (inter-matrix 
normalization) so that  matrices are comparable :

```{r normalizeTechnicalBiases}
suppressWarnings(hic.experiment <- normalizeTechnicalBiases(hic.experiment))
```

This normalization uses uses cyclic loess normalization from [multiHiCcompare package] [@multihiccompare].
**Note** : For large dataset, it is highly recommended to set a value for 
`cycleLoessSpan` parameter to reduce computing time and necessary memory. See
`?HiCDOC::normalizeTechnicalBiases`


### Biological biases
Normalize biological biases, such as GC content, number of restriction sites,
etc. (intra-matrix normalization):

```{r normalizeBiologicalBiases}
hic.experiment <- normalizeBiologicalBiases(hic.experiment)
```

### Distance effect

Normalize the linear distance effect resulting from more interactions between 
closer genomic regions (20000 is the default value for `loessSampleSize`):

```{r normalizeDistanceEffect}
hic.experiment <- 
    normalizeDistanceEffect(hic.experiment, loessSampleSize = 20000)
```

## Predicting compartments and differences

Predict A and B compartments and detect significant differences, here using 
the default values as parameters:

```{r detectCompartments}
hic.experiment <- detectCompartments(hic.experiment)
```


# Visualizing data and results

Plot the interaction matrix of each replicate:

```{r plotInteractions}
p <- plotInteractions(hic.experiment, chromosome = "Y")
p
```

Plot the overall distance effect on the proportion of interactions:

```{r plotDistanceEffect}
p <- plotDistanceEffect(hic.experiment)
p
```

List and plot compartments with their concordance (confidence measure) in each
replicate, and significant changes between experiment conditions:

```{r extractCompartments}
compartments(hic.experiment)
```


```{r extractConcordances}
concordances(hic.experiment)
```


```{r extractDifferences}
differences(hic.experiment)
```


```{r plotCompartmentChanges}
p <- plotCompartmentChanges(hic.experiment, chromosome = "Y")
p
```

Plot the overall distribution of concordance differences:

```{r plotConcordanceDifferences}
p <- plotConcordanceDifferences(hic.experiment)
p
```

Plot the result of the PCA on the compartments' centroids:

```{r plotCentroids}
p <- plotCentroids(hic.experiment, chromosome = "Y")
p
```

Plot the boxplots of self interaction ratios (differences between self
interactions and the median of other interactions) of each compartment:

```{r plotSelfInteractionRatios}
p <- plotSelfInteractionRatios(hic.experiment, chromosome = "Y")
p
```


## Sanity checks

Sometimes, basic assumptions on the data are not met.
We try to detect inconsistencies, and warn the user.


### PCA checks

We perform a principal component analysis on the centroids.
Each centroid represent an ideal bin, located either on compartment A or B,
in each sample, and each chromosome.
Given a chromosome, if all the centroids of the A compartment do not have the
same sign on the first principal component, we issue a warning for this
chromosome.
Likewise for the B compartment.

We also check that the inertia on the first principal component is at least
75%.

These checks make sure that centroids of the same compartments do cluster
together.
If the conditions are too different from each other, they may cluster together.
This case is detected by this check.


### Compartment assignemnt

We use "self-interaction" in order to classify centroids to A and B
compartments.
The self-interaction of a bin is the ratio of the number of pairs that link this
bin with other bins of the same compartment, divided by the number of pairs
The self-interactions should be different from compartments A and B.
We perform a Wilcoxon t-test.
If it is not significant, then a warning is issued.


### Warnings

If at least of the PCA checks fail, the warnings are printed on the PCA plot.
If the compartment assignment check fail, the warning is printed on the
corresponding plot.

When accessing the compartments and the concordances, chromosomes which
fail to pass the checks are removed (unless the appropriate parameter is set).


# Session info

```{r sessionInfo}
sessionInfo()
```


# References

[cooler-documentation]: https://cooler.readthedocs.io/en/latest/
[juicer-documentation]: https://github.com/aidenlab/juicer/wiki/Data
[hicpro-documentation]: https://github.com/nservant/HiC-Pro

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/normalizeBiologicalBiases.R
\name{normalizeBiologicalBiases}
\alias{normalizeBiologicalBiases}
\title{Normalize biological biases.}
\usage{
normalizeBiologicalBiases(object, parallel = FALSE)
}
\arguments{
\item{object}{A \code{\link{HiCDOCDataSet}}.}

\item{parallel}{Should the normalization be run in parallel mode? Default to FALSE.}
}
\value{
A \code{\link{HiCDOCDataSet}} with normalized interactions.
}
\description{
Normalizes biological biases such as GC content and repeated regions. Uses
the Knight-Ruiz balancing algorithm to transform interaction matrices into
doubly stochastic matrices, with sum of each row and sum of each column equal
to 1.
}
\examples{
data(exampleHiCDOCDataSet)
object <- exampleHiCDOCDataSet
object <- filterSparseReplicates(object)
object <- filterWeakPositions(object)
object <- normalizeBiologicalBiases(object)

}
\seealso{
\code{\link{filterSparseReplicates}},
\code{\link{filterWeakPositions}},
\code{\link{normalizeTechnicalBiases}},
\code{\link{normalizeDistanceEffect}},
\code{\link{HiCDOC}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{exampleHiCDOCDataSet}
\alias{exampleHiCDOCDataSet}
\title{Example HiCDOCDataSet.}
\format{
S4 HiCDOCDataSet object with the following characteristics:
\describe{
  \item{chromosomes}{4 chromosomes: W, X, Y, Z}
  \item{conditions}{3 conditions: 1, 2, 3}
  \item{replicates}{3 replicates: R1, R2, R3}
  \item{binSize}{A resolution of 137 bases}
}
}
\usage{
data(exampleHiCDOCDataSet)
}
\value{
A \code{\link{HiCDOCDataSet}}.
}
\description{
A S4 HiCDOCDataSet object with 4 chromosomes, 3 conditions and 3 replicates.
}
\examples{
data(exampleHiCDOCDataSet)
exampleHiCDOCDataSet
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotSelfInteractionRatios.R
\name{plotSelfInteractionRatios}
\alias{plotSelfInteractionRatios}
\title{Plot boxplots of self interaction ratios.}
\usage{
plotSelfInteractionRatios(object, chromosome, checks = TRUE)
}
\arguments{
\item{object}{A \code{\link{HiCDOCDataSet}}.}

\item{chromosome}{A chromosome name or index in \code{chromosomes(object)}.
A \code{\link{HiCDOCDataSet}}.}

\item{checks}{Logical. Should sanity checks messages be printed on plot ? Default to TRUE.}
}
\value{
A \code{ggplot}.
}
\description{
Plots the boxplots of self interaction ratios, which are the differences
between self interaction and median of other interactions for each genomic
position. Since the A compartment is open with more interactions overall, it
is assumed that self interaction ratios in compartment A are smaller than in
compartment B.
}
\examples{
data(exampleHiCDOCDataSetProcessed)
plotSelfInteractionRatios(exampleHiCDOCDataSetProcessed, chromosome = 1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotCentroids.R
\name{plotCentroids}
\alias{plotCentroids}
\title{Plot centroids.}
\usage{
plotCentroids(object, chromosome, size = 2, checks = TRUE)
}
\arguments{
\item{object}{A \code{\link{HiCDOCDataSet}}.}

\item{chromosome}{A chromosome name or index in \code{chromosomes(object)}.}

\item{size}{Size of each point. Defaults to 2.}

\item{checks}{Whether or not to add sanity checks messages on centroids. Default to TRUE.}
}
\value{
A \code{ggplot}.
}
\description{
Plots the result of the PCA on the compartments' centroids.
}
\examples{
data(exampleHiCDOCDataSetProcessed)
plotCentroids(exampleHiCDOCDataSetProcessed, chromosome = 1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotConcordances.R
\name{plotConcordances}
\alias{plotConcordances}
\title{Plot concordances.}
\usage{
plotConcordances(
  object,
  chromosome,
  xlim = NULL,
  threshold = 0.05,
  points = FALSE
)
}
\arguments{
\item{object}{A \code{\link{HiCDOCDataSet}}.}

\item{chromosome}{A chromosome name or index in \code{chromosomes(object)}.}

\item{xlim}{A vector of the minimum and maximum positions to display. If NULL, displays
all positions. Defaults to NULL.}

\item{threshold}{Significance threshold for the compartment changes. Defaults to 0.05.}

\item{points}{Whether or not to add points to the concordances. Defaults to FALSE.}
}
\value{
A \code{ggplot}.
}
\description{
Plots the concordances of each replicate in each experiment condition. A
concordance can be understood as a confidence in a genomic position's
assigned compartment. Mathematically, it is the log ratio of a genomic
position's distance to each compartment's centroid, normalized by the
distance between both centroids, and min-maxed to a [-1,1] interval.
}
\examples{
data(exampleHiCDOCDataSetProcessed)
plotConcordances(exampleHiCDOCDataSetProcessed, chromosome = 1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R
\name{HiCDOCDataSetFromCool}
\alias{HiCDOCDataSetFromCool}
\title{\code{\link{HiCDOCDataSet}} constructor from Cool files.}
\usage{
HiCDOCDataSetFromCool(paths, replicates, conditions, binSize = NA)
}
\arguments{
\item{paths}{A vector of paths to \code{.cool} or \code{.mcool} files.}

\item{replicates}{A vector of replicate names repeated along the conditions.}

\item{conditions}{A vector of condition names repeated along the replicates.}

\item{binSize}{The resolution (span of each position in number of bases). Optionally
provided to select the appropriate resolution in \code{.mcool} files.
Defaults to NULL.}
}
\value{
A \code{\link{HiCDOCDataSet}}.
}
\description{
Constructs a \code{\link{HiCDOCDataSet}} from a set of \code{.cool} or
\code{.mcool} files.
}
\examples{
\dontrun{
    # Path to each file
    paths = c(
      'path/to/condition-1.replicate-1.cool',
      'path/to/condition-1.replicate-2.cool',
      'path/to/condition-2.replicate-1.cool',
      'path/to/condition-2.replicate-2.cool',
      'path/to/condition-3.replicate-1.cool'
    )

    # Replicate and condition of each file. Can be names instead of numbers.
    replicates <- c(1, 2, 1, 2, 1)
    conditions <- c(1, 1, 2, 2, 3)

    # Resolution to select in .mcool files
    binSize = 500000

    # Instantiation of data set
    object <- HiCDOCDataSetFromCool(
      paths,
      replicates = replicates,
      conditions = conditions,
      binSize = binSize # Specified for .mcool files.
    )
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R
\name{HiCDOCDataSetFromTabular}
\alias{HiCDOCDataSetFromTabular}
\title{\code{\link{HiCDOCDataSet}} constructor from a tabular file.}
\usage{
HiCDOCDataSetFromTabular(path, sep = '\t')
}
\arguments{
\item{path}{A path to a tabular file.}

\item{sep}{The separator of the tabular file. Default to tabulation.}
}
\value{
A \code{\link{HiCDOCDataSet}}.
}
\description{
Constructs a \code{\link{HiCDOCDataSet}} from a tabular file.
}
\details{
Accepts a tabular file with \code{chromosome}, \code{position 1},
\code{position 2}, and multiple replicate columns listing interaction counts.
Null interactions do not have to be listed. Values must be separated by
tabulations. The header must be
\code{chromosome position 1 position 2 x.y x.y x.y ...} with \code{x}
replaced by condition names and \code{y} replaced by replicate names.
}
\examples{
path <- system.file("extdata", "liver_18_10M_500000.tsv", package = "HiCDOC")
object <- HiCDOCDataSetFromTabular(path, sep = '\t')

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/detectCompartments.R
\name{detectCompartments}
\alias{detectCompartments}
\title{A and B compartments detection and differences across conditions.}
\usage{
detectCompartments(
    object,
    parallel = FALSE,
    kMeansDelta = NULL,
    kMeansIterations = NULL,
    kMeansRestarts = NULL,
    PC1CheckThreshold = NULL
)
}
\arguments{
\item{object}{A \code{\link{HiCDOCDataSet}}.}

\item{parallel}{Whether or not to parallelize the processing. Defaults to FALSE
See 'Details'.}

\item{kMeansDelta}{The convergence stop criterion for the clustering. When the centroids'
distances between two iterations is lower than this value, the clustering
stops. Defaults to \code{object$kMeansDelta} which is originally set to
\code{defaultHiCDOCParameters$kMeansDelta} = 0.0001.}

\item{kMeansIterations}{The maximum number of iterations during clustering. Defaults to
\code{object$kMeansIterations} which is originally set to
\code{defaultHiCDOCParameters$kMeansIterations} = 50.}

\item{kMeansRestarts}{The amount of times the clustering is restarted. For each restart, the
clustering iterates until convergence or reaching the maximum number of
iterations. The clustering that minimizes inner-cluster variance is selected.
Defaults to \code{object$kMeansRestarts} which is originally set to
\code{defaultHiCDOCParameters$kMeansRestarts} = 20.}

\item{PC1CheckThreshold}{The minimum percentage of variance that should be explained by
the first principal component of centroids to pass sanity check.
Defaults to \code{object$PC1CheckThreshold} which is originally set to
\code{defaultHiCDOCParameters$PC1CheckThreshold} = 0.75}
}
\value{
A \code{\link{HiCDOCDataSet}}, with compartments, concordances, distances,
centroids, and differences.
}
\description{
Detects compartments for each genomic position in each condition, and
computes p-values for compartment differences between conditions.
}
\details{
\subsection{Genomic positions clustering}{
To clusterize genomic positions, the algorithm follows these steps:
    \enumerate{
        \item{
            For each chromosome and condition, get the interaction vectors of
            each genomic position. Each genomic position can have multiple
            interaction vectors, corresponding to the multiple replicates in
            that condition.
        }
        \item{
            For each chromosome and condition, use constrained K-means to
            clusterize the interaction vectors, forcing replicate interaction
            vectors into the same cluster. The euclidean distance between
            interaction vectors determines their similarity.
        }
        \item{
            For each interaction vector, compute its concordance, which is
            the confidence in its assigned cluster. Mathematically, it is the
            log ratio of its distance to each centroid, normalized by the
            distance between both centroids, and min-maxed to a [-1,1]
            interval.
        }
        \item{
            For each chromosome, compute the distance between all centroids
            and the centroids of the first condition. The cross-condition
            clusters whose centroids are closest are given the same cluster
            label. This results in two clusters per chromosome, spanning all
            conditions.
        }
    }
}
\subsection{A/B compartments prediction}{
To match each cluster with an A or B compartment, the algorithm follows these
steps:
    \enumerate{
        \item{
            For each genomic position, compute its self interaction ratio,
            which is the difference between its self interaction and the
            median of its other interactions.
        }
        \item{
            For each chromosome, for each cluster, get the median self
            interaction ratio of the genomic positions in that cluster.
        }
        \item{
            For each chromosome, the cluster with the smallest median self
            interaction ratio is matched with compartment A, and the cluster
            with the greatest median self interaction ratio is matched with
            compartment B. Compartment A being open, there are more overall
            interactions between distant genomic positions, so it is assumed
            that the difference between self interactions and other
            interactions is lower than in compartment B.
        }
    }
}
\subsection{Significant differences detection}{
To find significant compartment differences across conditions, and compute
their p-values, the algorithm follows three steps:
    \enumerate{
        \item{
            For each pair of replicates in different conditions, for each
            genomic position, compute the absolute difference between its
            concordances.
        }
        \item{
            For each pair of conditions, for each genomic position, compute
            the median of its concordance differences.
        }
        \item{
            For each pair of conditions, for each genomic position whose
            assigned compartment switches, rank its median against the
            empirical cumulative distribution of medians of all non-switching
            positions in that condition pair. Adjust the resulting p-value
            with the Benjamini–Hochberg procedure.
        }
    }
}
\subsection{Parallel processing}{
The parallel version of detectCompartments uses the
\code{\link[BiocParallel]{bpmapply}} function. Before to call the
function in parallel you should specify the parallel parameters such as:
    \itemize{
        \item{On Linux:

             \code{multiParam <- BiocParallel::MulticoreParam(workers = 10)}
         }
         \item{On Windows:

             \code{multiParam <- BiocParallel::SnowParam(workers = 10)}
        }
    }
    And then you can register the parameters to be used by BiocParallel:

    \code{BiocParallel::register(multiParam, default = TRUE)}

    You should be aware that using MulticoreParam, reproducibility of the
    detectCompartments function using a RNGseed may not work. See this
    \href{https://github.com/Bioconductor/BiocParallel/issues/122}{issue}
    for more details.
}
}
\examples{
data(exampleHiCDOCDataSet)
## Run all filtering and normalization steps (not run for timing reasons)
# object <- filterSmallChromosomes(exampleHiCDOCDataSet)
# object <- filterSparseReplicates(object)
# object <- filterWeakPositions(object)
# object <- normalizeTechnicalBiases(object)
# object <- normalizeBiologicalBiases(object)
# object <- normalizeDistanceEffect(object)

# Detect compartments and differences across conditions
object <- detectCompartments(exampleHiCDOCDataSet)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotCompartmentChanges.R
\name{plotCompartmentChanges}
\alias{plotCompartmentChanges}
\title{Plot compartment changes.}
\usage{
plotCompartmentChanges(
  object,
  chromosome,
  threshold = 0.05,
  xlim = NULL,
  points = FALSE,
  checks = TRUE,
  colour = "gray90"
)
}
\arguments{
\item{object}{A \code{\link{HiCDOCDataSet}}.}

\item{chromosome}{A chromosome name or index in \code{chromosomes(object)}.}

\item{threshold}{Significance threshold for the compartment changes. Defaults to 0.05.}

\item{xlim}{A vector of the minimum and maximum positions to display. If NULL, displays
all positions. Defaults to NULL.}

\item{points}{Whether or not to add points to the concordances. Defaults to FALSE.}

\item{checks}{Whether or not to add sanity checks messages. Default to TRUE.}

\item{colour}{Border color for the compartments. Default to `gray90`. `NA` means no border.}
}
\value{
A \code{ggplot}.
}
\description{
Plots the predicted compartments, along with their concordance in each
replicate, and significant changes between experiment conditions.
}
\examples{
data(exampleHiCDOCDataSetProcessed)
plotCompartmentChanges(exampleHiCDOCDataSetProcessed, chromosome = 1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filterSmallChromosomes.R
\name{filterSmallChromosomes}
\alias{filterSmallChromosomes}
\title{Filter small chromosomes.}
\usage{
filterSmallChromosomes(object, threshold = NULL)
}
\arguments{
\item{object}{A \code{\link{HiCDOCDataSet}}.}

\item{threshold}{The minimum length (number of positions) for a chromosome to be kept.
Defaults to \code{object$smallChromosomeThreshold} which is originally set to
\code{defaultHiCDOCParameters$smallChromosomeThreshold} = 100.}
}
\value{
A filtered \code{\link{HiCDOCDataSet}}.
}
\description{
Removes chromosomes whose length (in number of positions) is smaller than the
threshold.
}
\examples{
data(exampleHiCDOCDataSet)
object <- exampleHiCDOCDataSet

chromosomes(object)
object <- filterSmallChromosomes(object)
chromosomes(object)

}
\seealso{
\code{\link{filterSparseReplicates}},
\code{\link{filterWeakPositions}},
\code{\link{HiCDOC}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotCompartments.R
\name{plotCompartments}
\alias{plotCompartments}
\title{Plot A/B compartments.}
\usage{
plotCompartments(object, chromosome, xlim = NULL, colour = "gray90")
}
\arguments{
\item{object}{A \code{\link{HiCDOCDataSet}}.}

\item{chromosome}{A chromosome name or index in \code{chromosomes(object)}.}

\item{xlim}{A vector of the minimum and maximum positions to display. If NULL, displays
all positions. Defaults to NULL.}

\item{colour}{Border color for the compartments. Default to `gray90`. `NA` means no border.}
}
\value{
A \code{ggplot}.
}
\description{
Plots the predicted compartments in each experiment condition.
}
\examples{
data(exampleHiCDOCDataSetProcessed)
plotCompartments(exampleHiCDOCDataSetProcessed, chromosome = 1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotDistanceEffect.R
\name{plotDistanceEffect}
\alias{plotDistanceEffect}
\title{Plot the distance effect.}
\usage{
plotDistanceEffect(
  object,
  chromosome = NULL,
  transformX = "identity",
  transformY = "identity"
)
}
\arguments{
\item{object}{A \code{\link{HiCDOCDataSet}}.}

\item{chromosome}{Name (character) or index of the chromosome, if the plot should be
restricted to only one chromosome. Default to NULL.}

\item{transformX}{Transformation of the X axis. Default to "identity". See
\code{\link[ggplot2]{scale_x_continuous}} for other accepted values.}

\item{transformY}{Transformation of the Y axis. Default to "identity". See
\code{\link[ggplot2]{scale_y_continuous}} for other accepted values.}
}
\value{
A \code{ggplot}.
}
\description{
Plots the distance effect on proportion of interactions.
Each point is a cell in the interaction matrix, such that
the x-axis is the distance with respect to the diagonal,
the y-axis is number of counts.
Dots are binned.
}
\examples{
data(exampleHiCDOCDataSet)
plotDistanceEffect(exampleHiCDOCDataSet)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R
\name{HiCDOCDataSetFromHiC}
\alias{HiCDOCDataSetFromHiC}
\title{\code{\link{HiCDOCDataSet}} constructor from HiC files.}
\usage{
HiCDOCDataSetFromHiC(paths, replicates, conditions, binSize)
}
\arguments{
\item{paths}{A vector of paths to \code{.hic} files.}

\item{replicates}{A vector of replicate names repeated along the conditions.}

\item{conditions}{A vector of condition names repeated along the replicates.}

\item{binSize}{The resolution (span of each position in number of bases) to select within
the \code{.hic} files.}
}
\value{
A \code{\link{HiCDOCDataSet}}.
}
\description{
Constructs a \code{\link{HiCDOCDataSet}} from a set of
\code{.hic} files.
}
\examples{
\dontrun{
    #' # Path to each file
    paths = c(
      'path/to/condition-1.replicate-1.hic',
      'path/to/condition-1.replicate-2.hic',
      'path/to/condition-2.replicate-1.hic',
      'path/to/condition-2.replicate-2.hic',
      'path/to/condition-3.replicate-1.hic'
    )

    # Replicate and condition of each file. Can be names instead of numbers.
    replicates <- c(1, 2, 1, 2, 1)
    conditions <- c(1, 1, 2, 2, 3)

    # Resolution to select
    binSize <- 500000

    # Instantiation of data set
    hic.experiment <- HiCDOCDataSetFromHiC(
      paths,
      replicates = replicates,
      conditions = conditions,
      binSize = binSize
    )
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HiCDOC-package.R
\docType{package}
\name{HiCDOC-package}
\alias{HiCDOC-package}
\title{A/B compartment detection and differential analysis}
\description{
HiCDOC normalizes intrachromosomal Hi-C matrices, uses unsupervised learning to predict A/B compartments from multiple replicates, and detects significant compartment changes between experiment conditions. It provides a collection of functions assembled into a pipeline to filter and normalize the data, predict the compartments and visualize the results. It accepts several type of data: tabular `.tsv` files, Cooler `.cool` or `.mcool` files, Juicer `.hic` files or HiC-Pro `.matrix` and `.bed` files.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/mzytnicki/HiCDOC}
  \item Report bugs at \url{https://github.com/mzytnicki/HiCDOC/issues}
}

}
\author{
\strong{Maintainer}: Maigné Élise \email{elise.maigne@inrae.fr}

Authors:
\itemize{
  \item Maigné Élise \email{elise.maigne@inrae.fr}
  \item Kurylo Cyril \email{cyril.kurylo@inrae.fr}
  \item Zytnicki Matthias \email{matthias.zytnicki@inrae.fr}
  \item Foissac Sylvain \email{sylvain.foissac@inrae.fr}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filterSparseReplicates.R
\name{filterSparseReplicates}
\alias{filterSparseReplicates}
\title{Filter sparse replicates.}
\usage{
filterSparseReplicates(object, threshold = NULL)
}
\arguments{
\item{object}{A \code{\link{HiCDOCDataSet}}.}

\item{threshold}{The minimum percentage of non-zero interactions for a chromosome replicate to
be kept. If a chromosome replicate's percentage of non-zero interactions is
lower than this value, it is removed. Defaults to
\code{object$smallChromosomeThreshold} which is originally set to
\code{defaultHiCDOCParameters$smallChromosomeThreshold = 30\%}.}
}
\value{
A filtered \code{\link{HiCDOCDataSet}}.
}
\description{
Removes chromosome replicates whose percentage of non-zero interactions is
smaller than the threshold.
}
\examples{
data(exampleHiCDOCDataSet)
object <- exampleHiCDOCDataSet

object <- filterSparseReplicates(object)

}
\seealso{
\code{\link{filterSmallChromosomes}},
\code{\link{filterWeakPositions}},
\code{\link{HiCDOC}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R, R/methods.R
\name{HiCDOCDataSet-methods}
\alias{HiCDOCDataSet-methods}
\alias{chromosomes}
\alias{sampleConditions}
\alias{sampleReplicates}
\alias{compartments}
\alias{differences}
\alias{concordances}
\alias{chromosomes,HiCDOCDataSet-method}
\alias{sampleConditions,HiCDOCDataSet-method}
\alias{sampleReplicates,HiCDOCDataSet-method}
\alias{compartments,HiCDOCDataSet-method}
\alias{differences,HiCDOCDataSet-method}
\alias{concordances,HiCDOCDataSet-method}
\alias{show,HiCDOCDataSet-method}
\title{Methods to access a \code{\link{HiCDOCDataSet}} components.}
\usage{
\S4method{chromosomes}{HiCDOCDataSet}(object)

\S4method{sampleConditions}{HiCDOCDataSet}(object)

\S4method{sampleReplicates}{HiCDOCDataSet}(object)

\S4method{compartments}{HiCDOCDataSet}(object, passChecks = TRUE)

\S4method{differences}{HiCDOCDataSet}(object, threshold = NULL)

\S4method{concordances}{HiCDOCDataSet}(object, passChecks = TRUE)

\S4method{show}{HiCDOCDataSet}(object)
}
\arguments{
\item{object}{a HiCDOCDataSet object}

\item{passChecks}{logical. Display only the concordances/compartments for 
the chromosomes passing sanity checks.}

\item{threshold}{a numeric value between 0 and 1. If no threshold, all the differences will
be printed even the non significant ones. Otherwise the differences printed
are filtered to show the ones with an adjusted p-value <= \code{threshold}.}
}
\value{
A character vector (for \code{chromosomes}, \code{sampleConditions},
\code{sampleReplicates}), 
or a GRanges object
(for \code{compartments}, \code{concordances}, \code{differences}).
}
\description{
Retrieve information and results from a \code{\link{HiCDOCDataSet}}.
}
\section{Functions}{
\itemize{
\item \code{chromosomes()}: Retrieves the vector of chromosome names.

\item \code{sampleConditions()}: Retrieves the vector of condition names, one for each sample.

\item \code{sampleReplicates()}: Retrieves the vector of replicate names, one for each sample.

\item \code{compartments()}: Retrieves a \code{GenomicRange} of the compartment of every position
in every condition.

\item \code{differences()}: Retrieves a \code{GenomicRange} of the significant compartment differences
between conditions, and their p-values.

\item \code{concordances()}: Retrieves a \code{GenomicRange} of the concordance (confidence in assigned
compartment) of every position in every replicate.

}}
\examples{
# Load an example dataset already processed 
# (i.e. after the detection of compartments)
data(exampleHiCDOCDataSetProcessed)

exampleHiCDOCDataSetProcessed
chromosomes(exampleHiCDOCDataSetProcessed)
sampleConditions(exampleHiCDOCDataSetProcessed)
sampleReplicates(exampleHiCDOCDataSetProcessed)
compartments(exampleHiCDOCDataSetProcessed)
differences(exampleHiCDOCDataSetProcessed)
concordances(exampleHiCDOCDataSetProcessed)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotInteractions.R
\name{plotInteractions}
\alias{plotInteractions}
\title{Plot interaction matrices.}
\usage{
plotInteractions(
  object,
  chromosome,
  transform = NULL,
  colours = c(low = "#2c7bb6", mid = "#ffffbf", high = "#d7191c"),
  midpoint = 0
)
}
\arguments{
\item{object}{A \code{\link{HiCDOCDataSet}}.}

\item{chromosome}{A chromosome name or index in \code{chromosomes(object)}.}

\item{transform}{Transformation of the color scale. Default to NULL (no transformation). See
\code{\link[ggplot2]{scale_fill_gradient2}} for other accepted values.}

\item{colours}{A character vector colours of length 3 to use for the gradient. See
\code{\link[ggplot2]{scale_fill_gradient2}} for more info. Defaults to
\code{c("low"=#2c7bb6", "mid"=#ffffbf", "high"="#d7191c")}.}

\item{midpoint}{midpoint value to be passed to \code{scale_fill_gradient2}.
Default to 0.}
}
\value{
A \code{ggplot}.
}
\description{
Plots the interaction matrices as heatmaps.
}
\examples{
data(exampleHiCDOCDataSet)
plotInteractions(exampleHiCDOCDataSet, chromosome = 1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R
\name{HiCDOC}
\alias{HiCDOC}
\title{Default pipeline to run on the HiCDOC analysis.}
\usage{
HiCDOC(object, parallel = FALSE)
}
\arguments{
\item{object}{A \code{HiCDOCDataSet}.}

\item{parallel}{Whether or not to parallelize each step. Defaults to FALSE.}
}
\value{
A HiCDOCDataSet with all slots filled.
}
\description{
Runs the default filtering, normalization, and computational steps on a
\code{HiCDOCDataSet}. To learn more about HiCDOC, browse the vignette:
\code{browseVignettes(package = "HiCDOC")}.
}
\details{
\subsection{\code{HiCDOC} pipeline}{
    The HiCDOC pipeline has seven steps: \\
    \describe{
        Three filtering steps:
        \itemize{
            \item{\code{\link{filterSmallChromosomes}}} to filter out 
            small chromosomes
            \item{\code{\link{filterWeakPositions}}} to filter out weak 
            positions with very few interactions
            \item{\code{\link{filterSparseReplicates}}} to filter out 
            sparse replicates with many null interactions
        }
        Three normalization steps:
        \itemize{
            \item{\code{\link{normalizeTechnicalBiases}}} to normalize 
            technical biases in each replicates
            \item{\code{\link{normalizeBiologicalBiases}}} to normalize 
            biological biases in each replicate
            \item{\code{\link{normalizeDistanceEffect}}} to normalize the 
            distance effect in each chromosome
        }
        One computational step:
        \itemize{
            \item{\code{\link{detectCompartments}}} to detect compartments 
            in each condition and find significant changes between 
            conditions.
        }
    }
}
\subsection{Parallel processing}{
The parallel version of HiCDOC uses the
\code{\link{BiocParallel}} package. Before to call the
function in parallel you should specify the parallel parameters such as:
    \itemize{
        \item{On Linux:}
           \code{multiParam <- BiocParallel::MulticoreParam(workers = 10)}
         \item{On Windows:}
             \code{multiParam <- BiocParallel::SnowParam(workers = 10)}
    }
    And then you can register the parameters to be used by BiocParallel: \\
    \code{BiocParallel::register(multiParam, default = TRUE)}

    You should be aware that using MulticoreParam, reproducibility of the
    detectCompartments function using a RNGseed may not work. See this
    \href{https://github.com/Bioconductor/BiocParallel/issues/122}{issue}
    for more details.
}
}
\examples{
data(exampleHiCDOCDataSet)
# Default HiCDOC pipeline
# Not printing loess warnings for example purpose. 
# Results should be inspected if there is any.
suppressWarnings(
object <- HiCDOC(exampleHiCDOCDataSet)
)

# Equivalent to
if(FALSE){
    object <- filterSmallChromosomes(exampleHiCDOCDataSet)
    object <- filterSparseReplicates(object)
    object <- filterWeakPositions(object)
    object <- normalizeTechnicalBiases(object)
    object <- normalizeBiologicalBiases(object)
    object <- normalizeDistanceEffect(object)
    object <- detectCompartments(object)
}

}
\seealso{
\code{\link{HiCDOCDataSet}}, \code{\link{filterSmallChromosomes}},
\code{\link{filterWeakPositions}}, \code{\link{filterSparseReplicates}},
\code{\link{normalizeTechnicalBiases}},
\code{\link{normalizeBiologicalBiases}},
\code{\link{normalizeDistanceEffect}},
\code{\link{detectCompartments}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/normalizeTechnicalBiases.R
\name{normalizeTechnicalBiases}
\alias{normalizeTechnicalBiases}
\title{Normalize technical biases.}
\usage{
normalizeTechnicalBiases(object, parallel = FALSE, cyclicLoessSpan = NULL)
}
\arguments{
\item{object}{A \code{\link{HiCDOCDataSet}}.}

\item{parallel}{Logical. Whether or not to parallelize the processing. Defaults to FALSE}

\item{cyclicLoessSpan}{A numeric value in between 0 and 1. The span for cyclic loess normalization. 
This value is passed to \code{multiHiCcompare::cyclic_loess}. 
Defaults to NULL, NULL indicates that the value of 
parameters(object)$cyclicLoessSpan will be used. 
If this value is NA, the span will be automatically calculated using 
generalized cross validation. **For large dataset, it is highly recommended 
to set this value to reduce computing time and necessary memory.**}
}
\value{
A \code{\link{HiCDOCDataSet}} with normalized interactions.
}
\description{
Normalizes technical biases such as sequencing depth by using a cyclic loess
to recursively normalize each pair of interaction matrices. Depends on
\code{multiHiCcompare}.
}
\details{
\subsection{Parallel processing}{
If \code{parallel = TRUE}, the function
\code{\link[multiHiCcompare]{cyclic_loess}}
is launched in parallel mode, using \code{\link[BiocParallel]{bplapply}}
function. Before to call the function in parallel you should specify
the parallel parameters such as:
    \itemize{
        \item{On Linux:

             \code{multiParam <- BiocParallel::MulticoreParam(workers = 10)}
         }
         \item{On Windows:

             \code{multiParam <- BiocParallel::SnowParam(workers = 10)}
        }
    }
    And then you can register the parameters to be used by BiocParallel:

    \code{BiocParallel::register(multiParam, default = TRUE)}
}
}
\examples{
data(exampleHiCDOCDataSet)
object <- filterSmallChromosomes(exampleHiCDOCDataSet)
object <- filterSparseReplicates(object)
object <- filterWeakPositions(object)
# Not printing loess warnings for example purpose. 
# Results should be inspected if there is any.
suppressWarnings(
    object <- normalizeTechnicalBiases(object)
)


}
\seealso{
\code{\link{filterSparseReplicates}},
\code{\link{filterWeakPositions}},
\code{\link{normalizeBiologicalBiases}},
\code{\link{normalizeDistanceEffect}},
\code{\link{HiCDOC}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R, R/AllGenerics.R, R/methods.R
\docType{data}
\name{HiCDOCDataSet-parameters}
\alias{HiCDOCDataSet-parameters}
\alias{defaultHiCDOCParameters}
\alias{parameters}
\alias{parameters<-}
\alias{parameters,HiCDOCDataSet-method}
\alias{parameters<-,HiCDOCDataSet-method}
\title{Access the parameters of a \code{\link{HiCDOCDataSet}}.}
\format{
An object of class \code{list} of length 9.
}
\usage{
defaultHiCDOCParameters

\S4method{parameters}{HiCDOCDataSet}(object)

\S4method{parameters}{HiCDOCDataSet}(object) <- value
}
\arguments{
\item{object}{A \code{\link{HiCDOCDataSet}}.}

\item{value}{a named list containing the names and valued of the
parameters to change (see Details).}
}
\description{
Retrieves or sets parameters used for filtering, normalization, and
prediciton of compartments.
}
\details{
A \code{\link{HiCDOCDataSet}}'s parameters are automatically set to default
values retrieved from \code{\link{defaultHiCDOCParameters}}. They are
accessed by filtering, normalization, and compartment detection functions.
If those functions are called with custom arguments, the object's
parameters are updated to record the actual parameters used. If the
object's parameters are customized before calling the functions, the
custom parameters will be used.

See
\code{\link{filterSmallChromosomes}},
\code{\link{filterSparseReplicates}},
\code{\link{filterWeakPositions}},
\code{\link{normalizeDistanceEffect}}, and
\code{\link{detectCompartments}},
for details on how these parameters are used.
\subsection{All parameters are listed here:}{
    \describe{
        \item{\code{smallChromosomeThreshold}}{
            The minimum length (number of positions) for a chromosome to be
            kept when filtering with \code{\link{filterSmallChromosomes}}.
            Defaults to
            \code{defaultHiCDOCParameters$smallChromosomeThreshold} = 100.
        }
        \item{\code{sparseReplicateThreshold}}{
            The minimum percentage of non-zero interactions for a chromosome
            replicate to be kept when filtering with
            \code{\link{filterSparseReplicates}}. If a chromosome replicate's
            percentage of non-zero interactions is lower than this value, it
            is removed. Defaults to
            \code{defaultHiCDOCParameters$smallChromosomeThreshold} = 30%.
        }
        \item{\code{weakPositionThreshold}}{
            The minimum average interaction for a position to be kept when
            filtering with \code{\link{filterWeakPositions}}. If a position's
            average interaction with the entire chromosome is lower than this
            value in any of the replicates, it is removed from all replicates
            and conditions. Defaults to
            \code{defaultHiCDOCParameters$smallChromosomeThreshold} = 1.
        }
        \item{\code{cyclicLoessSpan}}{
            The span for cyclic loess normalization used in 
            \code{\link{normalizeTechnicalBiases}}. This value is passed to 
            \code{multiHiCcompare::cyclic_loess}. 
            Defaults to NA indicating that span will be automatically 
            calculated using generalized cross validation.
            For large dataset, it is highly recommended to set this value
            to reduce computing time and necessary memory.
        }
        \item{\code{loessSampleSize}}{
            The number of positions used as a sample to estimate the effect
            of distance on proportion of interactions when normalizing with
            \code{\link{normalizeDistanceEffect}}. Defaults to
            \code{defaultHiCDOCParameters$loessSampleSize} = 20000.
        }
        \item{\code{kMeansDelta}}{
            The convergence stop criterion for the clustering when detecting
            compartments with \code{\link{detectCompartments}}. When the
            centroids' distances between two iterations is lower than this
            value, the clustering stops. Defaults to
            \code{defaultHiCDOCParameters$kMeansDelta} = 0.0001.
        }
        \item{\code{kMeansIterations}}{
            The maximum number of iterations during clustering when detecting
            compartments with \code{\link{detectCompartments}}. Defaults to
            \code{defaultHiCDOCParameters$kMeansIterations} = 50.
        }
        \item{\code{kMeansRestarts}}{
            The amount of times the clustering is restarted when detecting
            compartments with \code{\link{detectCompartments}}. For each
            restart, the clustering iterates until convergence or reaching
            the maximum number of iterations. The clustering that minimizes
            inner-cluster variance is selected. Defaults to
            \code{defaultHiCDOCParameters$kMeansRestarts} = 20.
        }
        \item{\code{PC1CheckThreshold}}{
            The minimum percentage of variance that should be explained by
            the first principal component of centroids to pass sanity check.
            Defaults to 
            \code{defaultHiCDOCParameters$PC1CheckThreshold} = 0.75
        }
    }
}
}
\examples{
data(exampleHiCDOCDataSet)

# Retrieve parameters
parameters(exampleHiCDOCDataSet)

# Set parameters
parameters(exampleHiCDOCDataSet) <- list("smallChromosomeThreshold" = 50)
parameters(exampleHiCDOCDataSet) <- list(
    "weakPositionThreshold" = 10,
    "kMeansRestarts" = 30
)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filterWeakPositions.R
\name{filterWeakPositions}
\alias{filterWeakPositions}
\title{Filter weak positions.}
\usage{
filterWeakPositions(object, threshold = NULL)
}
\arguments{
\item{object}{A \code{\link{HiCDOCDataSet}}.}

\item{threshold}{The minimum average interaction for a position to be kept. If a position's
average interaction with the entire chromosome is lower than this value in
any of the replicates, it is removed from all replicates and conditions.
Defaults to \code{object$weakPositionThreshold} which is originally set to
\code{defaultHiCDOCParameters$weakPositionThreshold} = 1.}
}
\value{
A filtered \code{\link{HiCDOCDataSet}}.
}
\description{
Removes weak genomic positions whose interactions average is lower than the
threshold.
}
\details{
Detects weak genomic positions in each replicate, and removes them from all
replicates to guarantee comparability across conditions when calling
\code{\link{detectCompartments}}.
}
\examples{
data(exampleHiCDOCDataSet)
object <- exampleHiCDOCDataSet

object <- filterWeakPositions(object)

}
\seealso{
\code{\link{filterSmallChromosomes}},
\code{\link{filterSparseReplicates}},
\code{\link{HiCDOC}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotConcordanceDifferences.R
\name{plotConcordanceDifferences}
\alias{plotConcordanceDifferences}
\title{Plot the distribution of concordance differences.}
\usage{
plotConcordanceDifferences(object)
}
\arguments{
\item{object}{A \code{\link{HiCDOCDataSet}}.}
}
\value{
A \code{ggplot}.
}
\description{
Plots the distribution of concordance differences, which are the differences
between concordances of each pair of replicates from different conditions. A
concordance can be understood as a confidence in a genomic position's
assigned compartment. Mathematically, it is the log ratio of a genomic
position's distance to each compartment's centroid, normalized by the
distance between both centroids, and min-maxed to a [-1,1] interval.
}
\examples{
data(exampleHiCDOCDataSetProcessed)
plotConcordanceDifferences(exampleHiCDOCDataSetProcessed)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reduceDataSet.R
\name{reduceHiCDOCDataSet}
\alias{reduceHiCDOCDataSet}
\title{Reduce a \code{\link{HiCDOCDataSet}}.}
\usage{
reduceHiCDOCDataSet(
  object,
  chromosomes = NULL,
  conditions = NULL,
  replicates = NULL,
  dropLevels = TRUE
)
}
\arguments{
\item{object}{A \code{\link{HiCDOCDataSet}}.}

\item{chromosomes}{The chromosome names or indices in \code{chromosomes(object)} to keep.
Defaults to NULL.}

\item{conditions}{The condition names in \code{sampleConditions(object)} to keep. 
Defaults to NULL.}

\item{replicates}{The replicate names in \code{sampleReplicates(object)} to keep. 
Defaults to NULL.}

\item{dropLevels}{Whether or not to also remove unused factor levels after filtering. Should
be set to FALSE if the reduced objects are meant to be re-combined later.
Defaults to TRUE.}
}
\value{
A reduced \code{\link{HiCDOCDataSet}}.
}
\description{
Reduces a \code{\link{HiCDOCDataSet}} by keeping only given chromosomes,
conditions, or replicates.
}
\examples{
data(exampleHiCDOCDataSet)
reduced <- reduceHiCDOCDataSet(exampleHiCDOCDataSet, chromosomes = c(1, 2))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R
\name{HiCDOCDataSetFromHiCPro}
\alias{HiCDOCDataSetFromHiCPro}
\title{\code{\link{HiCDOCDataSet}} constructor from HiC-Pro files.}
\usage{
HiCDOCDataSetFromHiCPro(matrixPaths, bedPaths, replicates, conditions)
}
\arguments{
\item{matrixPaths}{A vector of paths to HiC-Pro matrix files.}

\item{bedPaths}{A vector of paths to HiC-Pro bed files.}

\item{replicates}{A vector of replicate names repeated along the conditions.}

\item{conditions}{A vector of condition names repeated along the replicates.}
}
\value{
A \code{\link{HiCDOCDataSet}}.
}
\description{
Constructs a \code{\link{HiCDOCDataSet}} from a set of HiC-Pro generated
files.
}
\examples{
\dontrun{
    # Path to each matrix file
    matrixPaths = c(
      'path/to/condition-1.replicate-1.matrix',
      'path/to/condition-1.replicate-2.matrix',
      'path/to/condition-2.replicate-1.matrix',
      'path/to/condition-2.replicate-2.matrix',
      'path/to/condition-3.replicate-1.matrix'
    )

    # Path to each bed file
    bedPaths = c(
      'path/to/condition-1.replicate-1.bed',
      'path/to/condition-1.replicate-2.bed',
      'path/to/condition-2.replicate-1.bed',
      'path/to/condition-2.replicate-2.bed',
      'path/to/condition-3.replicate-1.bed'
    )

    # Replicate and condition of each file. Can be names instead of numbers.
    replicates <- c(1, 2, 1, 2, 1)
    conditions <- c(1, 1, 2, 2, 3)

    # Instantiation of data set
    hic.experiment <- HiCDOCDataSetFromHiCPro(
      matrixPaths = matrixPaths,
      bedPaths = bedPaths,
      replicates = replicates,
      conditions = conditions
    )
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/normalizeDistanceEffect.R
\name{normalizeDistanceEffect}
\alias{normalizeDistanceEffect}
\title{Normalize distance effect.}
\usage{
normalizeDistanceEffect(object, loessSampleSize = NULL, parallel = FALSE)
}
\arguments{
\item{object}{A \code{\link{HiCDOCDataSet}}.}

\item{loessSampleSize}{The number of positions used as a sample to estimate the effect of distance
on proportion of interactions. Defaults to
\code{object$loessSampleSize} which is originally set to
\code{defaultHiCDOCParameters$loessSampleSize} = 20000.}

\item{parallel}{Should the normalization be run in parallel mode? Default to FALSE.}
}
\value{
A \code{\link{HiCDOCDataSet}} with normalized interactions.
}
\description{
Normalizes interactions by their "expected" value relative to the distance
that separates their positions. The "expected" values are estimated with a
loess regression on the proportion of interactions for each distance.
}
\examples{
data(exampleHiCDOCDataSet)
object <- normalizeDistanceEffect(exampleHiCDOCDataSet)

}
\seealso{
\code{\link{normalizeTechnicalBiases}},
\code{\link{normalizeBiologicalBiases}},
\code{\link{HiCDOC}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R
\docType{class}
\name{HiCDOCDataSet-class}
\alias{HiCDOCDataSet-class}
\alias{HiCDOCDataSet}
\title{\code{HiCDOCDataSet} S4 class.}
\description{
Data structure for a Hi-C experiment.
}
\details{
An instance of \code{HiCDOCDataSet} describes a Hi-C experiment with slots
for path(s) to input file(s), interactions, pipeline parameters defaulting to
\code{defaultHiCDOCParameters}, and computation results. It can be
constructed from 4 different types of data:
\itemize{
\item Tabular files: see \code{\link{HiCDOCDataSetFromTabular}}
\item (m)Cool files: see \code{\link{HiCDOCDataSetFromCool}}
\item HiC files: see \code{\link{HiCDOCDataSetFromHiC}}
\item HiC-Pro matrices and bed files: see \code{\link{HiCDOCDataSetFromHiCPro}}
An example \code{HiCDOCDataSet} is also available, see
\code{\link{exampleHiCDOCDataSet}}.
The \code{HiCDOCDataSet} object can be explored using the appropriate
accessors.
}
}
\section{Accessors}{
The accessors for a HiCDOCDataset object are the following:
\itemize{
\item \code{\link{chromosomes}} to retrieve the vector of chromosome names.
\item \code{\link{sampleConditions}} to  retrieve the vector of condition names,
one for each sample.
\item \code{\link{sampleReplicates}} to retrieve the vector of replicate names,
one for each sample.
}

After the detection of compartments you can use this accessors:
\itemize{
\item \code{\link{compartments}} returns a GenomicRange of the compartment
of every position in every condition.
\item \code{\link{concordances}} returns a GenomicRange of the significant
compartment differences between conditions, and their p-values.
\item \code{\link{differences}} returns a GenomicRange of the concordance
(confidence in assigned compartment) of every position in every replicate.
}

See the \link{HiCDOCDataSet-methods} man page for more details on methods
and accessors.
}

\seealso{
\code{\link{HiCDOC}},
\code{\link{exampleHiCDOCDataSet}},
\code{\link{HiCDOCDataSetFromTabular}},
\code{\link{HiCDOCDataSetFromCool}},
\code{\link{HiCDOCDataSetFromHiC}},
\code{\link{HiCDOCDataSetFromHiCPro}}
}
